#include "mvTabButton.h"
#include "mvContext.h"
#include "mvLog.h"
#include "mvItemRegistry.h"
#include "mvPythonExceptions.h"
#include "AppItems/fonts/mvFont.h"
#include "AppItems/themes/mvTheme.h"
#include "AppItems/containers/mvDragPayload.h"
#include "AppItems/widget_handlers/mvItemHandlerRegistry.h"

namespace Marvel {

	mvTabButton::mvTabButton(mvUUID uuid)
		:
		mvAppItem(uuid)
	{
	}

	void mvTabButton::applySpecificTemplate(mvAppItem* item)
	{
		auto titem = static_cast<mvTabButton*>(item);
		_flags = titem->_flags;
	}

	void mvTabButton::draw(ImDrawList* drawlist, float x, float y)
	{

		//-----------------------------------------------------------------------------
		// pre draw
		//-----------------------------------------------------------------------------

		// show/hide
		if (!config.show)
			return;

		// focusing
		if (info.focusNextFrame)
		{
			ImGui::SetKeyboardFocusHere();
			info.focusNextFrame = false;
		}

		// cache old cursor position
		ImVec2 previousCursorPos = ImGui::GetCursorPos();

		// set cursor position if user set
		if (info.dirtyPos)
			ImGui::SetCursorPos(state.pos);

		// update widget's position state
		state.pos = { ImGui::GetCursorPosX(), ImGui::GetCursorPosY() };

		// set item width
		if (config.width != 0)
			ImGui::SetNextItemWidth((float)config.width);

		// set indent
		if (config.indent > 0.0f)
			ImGui::Indent(config.indent);

		// push font if a font object is attached
		if (font)
		{
			ImFont* fontptr = static_cast<mvFont*>(font.get())->getFontPtr();
			ImGui::PushFont(fontptr);
		}

		// themes
		apply_local_theming(this);

		//-----------------------------------------------------------------------------
		// draw
		//-----------------------------------------------------------------------------
		{
			ScopedID id(uuid);

			if (ImGui::TabItemButton(info.internalLabel.c_str(), _flags))
			{
				if(config.alias.empty())
					mvAddCallback(getCallback(false), uuid, nullptr, config.user_data);
				else
					mvAddCallback(getCallback(false), config.alias, nullptr, config.user_data);
			}
		}

		//-----------------------------------------------------------------------------
		// update state
		//-----------------------------------------------------------------------------
		UpdateAppItemState(state);

		//-----------------------------------------------------------------------------
		// post draw
		//-----------------------------------------------------------------------------

		// set cursor position to cached position
		if (info.dirtyPos)
			ImGui::SetCursorPos(previousCursorPos);

		if (config.indent > 0.0f)
			ImGui::Unindent(config.indent);

		// pop font off stack
		if (font)
			ImGui::PopFont();

		// handle popping themes
		cleanup_local_theming(this);

		if (handlerRegistry)
			handlerRegistry->checkEvents(&state);

		// handle drag & drop if used
		apply_drag_drop(this);
	}

	void mvTabButton::handleSpecificKeywordArgs(PyObject* dict)
	{
		if (dict == nullptr)
			return;
		 

		// helper for bit flipping
		auto flagop = [dict](const char* keyword, int flag, int& flags)
		{
			if (PyObject* item = PyDict_GetItemString(dict, keyword)) ToBool(item) ? flags |= flag : flags &= ~flag;
		};

		// window flags
		flagop("no_reorder", ImGuiTabItemFlags_NoReorder, _flags);
		flagop("leading", ImGuiTabItemFlags_Leading, _flags);
		flagop("trailing", ImGuiTabItemFlags_Trailing, _flags);
		flagop("no_tooltip", ImGuiTabItemFlags_NoTooltip, _flags);

		if (_flags & ImGuiTabItemFlags_Leading && _flags & ImGuiTabItemFlags_Trailing)
			_flags &= ~ImGuiTabItemFlags_Leading;

	}

	void mvTabButton::getSpecificConfiguration(PyObject* dict)
	{
		if (dict == nullptr)
			return;
		 

		// helper to check and set bit
		auto checkbitset = [dict](const char* keyword, int flag, const int& flags)
		{
			PyDict_SetItemString(dict, keyword, mvPyObject(ToPyBool(flags & flag)));
		};

		// window flags
		checkbitset("no_reorder", ImGuiTabBarFlags_Reorderable, _flags);
		checkbitset("leading", ImGuiTabItemFlags_Leading, _flags);
		checkbitset("trailing", ImGuiTabItemFlags_Trailing, _flags);
		checkbitset("no_tooltip", ImGuiTabItemFlags_NoTooltip, _flags);

	}

}