
##########################################################
# Dear PyGui User Interface (MODIFIED FOR READTHEDOCS)
#   ~ Version: master
#
#   Notes:
#     * This file is automatically generated.
#
#   Resources:
#     * FAQ:         https://github.com/hoffstadt/DearPyGui/discussions/categories/frequently-asked-questions-faq 
#     * Homepage:    https://github.com/hoffstadt/DearPyGui 
#     * Wiki:        https://github.com/hoffstadt/DearPyGui/wiki 
#     * Issues:      https://github.com/hoffstadt/DearPyGui/issues
#     * Discussions: https://github.com/hoffstadt/DearPyGui/discussions
##########################################################

from typing import List, Any, Callable, Union, Tuple
from contextlib import contextmanager
import warnings
import functools
import inspect
import dearpygui._dearpygui as internal_dpg
from dearpygui._dearpygui import mvBuffer
from dearpygui._dearpygui import mvVec4
from dearpygui._dearpygui import mvMat4

########################################################################################################################
# User API Index
#
#    * Sections
#      - Helper Commands
#      - Tool Commands
#      - Information Commands
#      - Configuration Getter Commands
#      - Configuration Setter Commands
#      - State Commands
#      - Viewport Setter Commands
#      - Viewport Getter Commands
#      - Deprecated Commands
#      - Container Context Managers
#      - Public _dearpygui Wrappings
#      - Constants
#
########################################################################################################################

########################################################################################################################
# Helper Commands
########################################################################################################################

def run_callbacks(jobs):
    """ New in 1.2. Runs callbacks from the callback queue and checks arguments. """

    if jobs is None:
        pass
    else:
        for job in jobs:
            if job[0] is None:
                pass
            else:
                sig = inspect.signature(job[0])
                args = []
                for arg in range(len(sig.parameters)):
                    args.append(job[arg+1])
                job[0](*args)

def get_major_version():
    """ return Dear PyGui Major Version """
    return internal_dpg.get_app_configuration()["major_version"]

def get_minor_version():
    """ return Dear PyGui Minor Version """
    return internal_dpg.get_app_configuration()["minor_version"]

def get_dearpygui_version():
    """ return Dear PyGui Version """
    return internal_dpg.get_app_configuration()["version"]

def configure_item(item : Union[int, str], **kwargs) -> None:
	"""Configures an item after creation."""
	internal_dpg.configure_item(item, **kwargs)

def configure_app(**kwargs) -> None:
	"""Configures an item after creation."""
	internal_dpg.configure_app(**kwargs)

def configure_viewport(item : Union[int, str], **kwargs) -> None:
	"""Configures a viewport after creation."""
	internal_dpg.configure_viewport(item, **kwargs)

def start_dearpygui():
    """Prepares viewport (if not done already). sets up, cleans up, and runs main event loop.

    Returns:
        None
    """

    if not internal_dpg.is_viewport_ok():
        raise RuntimeError("Viewport was not created and shown.")
        return

    while(internal_dpg.is_dearpygui_running()):
        internal_dpg.render_dearpygui_frame()   


@contextmanager
def mutex():
    """ Handles locking/unlocking render thread mutex. """
    try:
        yield internal_dpg.lock_mutex()
    finally:
        internal_dpg.unlock_mutex()


@contextmanager
def popup(parent: Union[int, str], mousebutton: int = internal_dpg.mvMouseButton_Right, modal: bool = False, tag:Union[int, str] = 0) -> int:
    """ Popup widget. """    
    try:
        if tag == 0:
            _internal_popup_id = internal_dpg.generate_uuid()
        else:
            _internal_popup_id = tag
        _handler_reg_id = internal_dpg.add_item_handler_registry()
        internal_dpg.add_item_clicked_handler(mousebutton, parent=internal_dpg.last_item(), callback=lambda: internal_dpg.configure_item(_internal_popup_id, show=True))
        internal_dpg.bind_item_handler_registry(parent, _handler_reg_id)
        if modal:
            internal_dpg.add_window(modal=True, show=False, tag=_internal_popup_id, autosize=True)
        else:
            internal_dpg.add_window(popup=True, show=False, tag=_internal_popup_id, autosize=True)
        internal_dpg.push_container_stack(internal_dpg.last_container())
        yield _internal_popup_id

    finally:
        internal_dpg.pop_container_stack()


########################################################################################################################
# Tool Commands
########################################################################################################################

def show_style_editor() -> None:
    """Shows the standard style editor window

    Returns:
        None
    """
    internal_dpg.show_tool(internal_dpg.mvTool_Style)


def show_metrics() -> None:
    """Shows the standard metrics window

    Returns:
        None
    """
    internal_dpg.show_tool(internal_dpg.mvTool_Metrics)


def show_about() -> None:
    """Shows the standard about window

    Returns:
        None
    """
    internal_dpg.show_tool(internal_dpg.mvTool_About)


def show_debug() -> None:
    """Shows the standard debug window

    Returns:
        None
    """
    internal_dpg.show_tool(internal_dpg.mvTool_Debug)


def show_documentation() -> None:
    """Shows the standard documentation window

    Returns:
        None
    """
    internal_dpg.show_tool(internal_dpg.mvTool_Doc)


def show_font_manager() -> None:
    """Shows the standard documentation window

    Returns:
        None
    """
    internal_dpg.show_tool(internal_dpg.mvTool_Font)


def show_item_registry() -> None:
    """Shows the standard documentation window

    Returns:
        None
    """
    internal_dpg.show_tool(internal_dpg.mvTool_ItemRegistry)


########################################################################################################################
# Information Commands
########################################################################################################################

def get_item_slot(item: Union[int, str]) -> Union[int, None]:
    """Returns an item's target slot.

    Returns:
        slot as a int
    """
    return internal_dpg.get_item_info(item)["target"]


def is_item_container(item: Union[int, str]) -> Union[bool, None]:
    """Checks if item is a container.

    Returns:
        status as a bool
    """
    return internal_dpg.get_item_info(item)["container"]


def get_item_parent(item: Union[int, str]) -> Union[int, None]:
    """Gets the item's parent.

    Returns:
        parent as a int or None
    """
    return internal_dpg.get_item_info(item)["parent"]


def get_item_children(item: Union[int, str] , slot: int = -1) -> Union[dict, List[int], None]:
    """Provides access to the item's children slots.

    Returns:
        A 2-D tuple of children slots ex. ((child_slot_1),(child_slot_2),(child_slot_3),...) or a single slot if slot is used.
    """
    if slot < 0 or slot > 4:
        return internal_dpg.get_item_info(item)["children"]
    return internal_dpg.get_item_info(item)["children"][slot]


def get_item_type(item: Union[int, str]) -> Union[str]:
    """Gets the item's type.

    Returns:
        type as a string or None
    """
    return internal_dpg.get_item_info(item)["type"]


def get_item_theme(item: Union[int, str]) -> int:
    """Gets the item's theme.

    Returns:
        theme's uuid
    """
    return internal_dpg.get_item_info(item)["theme"]


def get_item_font(item: Union[int, str]) -> int:
    """Gets the item's font.

    Returns:
        font's uuid
    """
    return internal_dpg.get_item_info(item)["font"]


def get_item_disabled_theme(item: Union[int, str]) -> int:
    """Gets the item's disabled theme.

    Returns:
        theme's uuid
    """
    return internal_dpg.get_item_info(item)["disabled_theme"]


########################################################################################################################
# Configuration Setter Commands
########################################################################################################################

def enable_item(item: Union[int, str]):
    """Enables the item.

    Args:
        **item: Item to enable.

    Returns:
        None
    """
    internal_dpg.configure_item(item, enabled=True)


def disable_item(item: Union[int, str]):
    """Disables the item.

    Args:
        **item: Item to disable.

    Returns:
        None
    """
    internal_dpg.configure_item(item, enabled=False)


def set_item_label(item: Union[int, str], label: str):
    """Sets the item's displayed label, anything after the characters "##" in the name will not be shown.

    Args:
        item: Item label will be applied to.
        label: Displayed name to be applied.

    Returns:
        None
    """
    internal_dpg.configure_item(item, label=label)


def set_item_source(item: Union[int, str], source: Union[int, str]):
    """Sets the item's value, to the source's value. Widget's value will now be "linked" to source's value.

    Args:
        item: Item to me linked.
        source: Source to link to.

    Returns:
        None
    """
    internal_dpg.configure_item(item, source=source)


def set_item_pos(item: Union[int, str], pos: List[float]):
    """Sets the item's position.

    Args:
        item: Item the absolute position will be applied to.
        pos: X and Y positions relative to parent of the item.

    Returns:
        None
    """
    internal_dpg.configure_item(item, pos=pos)


def set_item_width(item: Union[int, str], width: int):
    """Sets the item's width.

    Args:
        item: Item the Width will be applied to.
        width: Width to be applied.

    Returns:
        None
    """
    internal_dpg.configure_item(item, width=width)


def set_item_height(item: Union[int, str], height: int):
    """Sets the item's height.

    Args:
        item: Item the Height will be applied to.
        height: Height to be applied.

    Returns:
        None
    """
    internal_dpg.configure_item(item, height=height)


def set_item_indent(item: Union[int, str], indent: int):
    """Sets the item's indent.

    Args:
        item: Item the Height will be applied to.
        height: Height to be applied.

    Returns:
        None
    """
    internal_dpg.configure_item(item, indent=indent)


def set_item_track_offset(item: Union[int, str], offset: float):
    """Sets the item's track offset.

    Args:
        item: Item the Height will be applied to.
        height: Height to be applied.

    Returns:
        None
    """
    internal_dpg.configure_item(item, track_offset=offset)


def set_item_payload_type(item: Union[int, str], payload_type: str):
    """Sets the item's payload type.

    Args:
        item: Item the Height will be applied to.
        height: Height to be applied.

    Returns:
        None
    """
    internal_dpg.configure_item(item, payload_type=str)


def set_item_callback(item: Union[int, str], callback: Callable):
    """Sets the item's callack.

    Args:
        item: Item the callback will be applied to.
        callback: Callback to be applied.

    Returns:
        None
    """
    internal_dpg.configure_item(item, callback=callback)


def set_item_drag_callback(item: Union[int, str], callback: Callable):
    """Sets the item's drag callack.

    Args:
        item: Item the callback will be applied to.
        callback: Callback to be applied.

    Returns:
        None
    """
    internal_dpg.configure_item(item, drag_callback=callback)


def set_item_drop_callback(item: Union[int, str], callback: Callable):
    """Sets the item's drop callack.

    Args:
        item: Item the callback will be applied to.
        callback: Callback to be applied.

    Returns:
        None
    """
    internal_dpg.configure_item(item, drop_callback=callback)


def track_item(item: Union[int, str]):
    """Track item in scroll region.

    Args:
        item: Item the callback will be applied to.
        callback: Callback to be applied.

    Returns:
        None
    """
    internal_dpg.configure_item(item, tracked=True)


def untrack_item(item: Union[int, str]):
    """Track item in scroll region.

    Args:
        item: Item the callback will be applied to.
        callback: Callback to be applied.

    Returns:
        None
    """
    internal_dpg.configure_item(item, tracked=False)


def set_item_user_data(item: Union[int, str], user_data: Any):
    """Sets the item's callack_data to any python object.

    Args:
        item: Item the callback will be applied to.
        user_data: Callback_data to be applied.

    Returns:
        None
    """
    internal_dpg.configure_item(item, user_data=user_data)


def show_item(item: Union[int, str]):
    """Shows the item.

    Args:
        item: Item to show.

    Returns:
        None
    """
    internal_dpg.configure_item(item, show=True)


def hide_item(item: Union[int, str], *, children_only: bool = False):
    """Hides the item.

    Args:
        **item: Item to hide.

    Returns:
        None
    """
    if children_only:
        children = get_item_children(item)
        for child in children:
            internal_dpg.configure_item(child, show=False)
    else:
        internal_dpg.configure_item(item, show=False)


########################################################################################################################
# Configuration Getter Commands
########################################################################################################################

def get_item_label(item: Union[int, str]) -> Union[str, None]:
    """Gets the item's label.

    Returns:
        label as a string or None
    """
    return internal_dpg.get_item_configuration(item)["label"]


def get_item_filter_key(item: Union[int, str]) -> Union[str, None]:
    """Gets the item's filter key.

    Returns:
        filter key as a string or None
    """
    return internal_dpg.get_item_configuration(item)["filter_key"]


def is_item_tracked(item: Union[int, str]) -> Union[bool, None]:
    """Checks if item is tracked.

    Returns:
        tracked as a bool or None
    """
    return internal_dpg.get_item_configuration(item)["tracked"]


def is_item_search_delayed(item: Union[int, str]) -> Union[bool, None]:
    """Checks if item is search delayed.

    Returns:
        tracked as a bool or None
    """
    return internal_dpg.get_item_configuration(item)["delay_search"]


def get_item_indent(item: Union[int, str]) -> Union[int, None]:
    """Gets the item's indent.

    Returns:
        indent as a int or None
    """
    return internal_dpg.get_item_configuration(item)["indent"]


def get_item_track_offset(item: Union[int, str]) -> Union[float, None]:
    """Gets the item's track offset.

    Returns:
        track offset as a int or None
    """
    return internal_dpg.get_item_configuration(item)["track_offset"]


def get_item_width(item: Union[int, str]) -> Union[int, None]:
    """Gets the item's width.

    Returns:
        width as a int or None
    """
    return internal_dpg.get_item_configuration(item)["width"]


def get_item_height(item: Union[int, str]) -> Union[int, None]:
    """Gets the item's height.

    Returns:
        height as a int or None
    """
    return internal_dpg.get_item_configuration(item)["height"]


def get_item_callback(item: Union[int, str]) -> Union[Callable, None]:
    """Gets the item's callback.

    Returns:
        callback as a callable or None
    """
    return internal_dpg.get_item_configuration(item)["callback"]


def get_item_drag_callback(item: Union[int, str]) -> Union[Callable, None]:
    """Gets the item's drag callback.

    Returns:
        callback as a callable or None
    """
    return internal_dpg.get_item_configuration(item)["drag_callback"]


def get_item_drop_callback(item: Union[int, str]) -> Union[Callable, None]:
    """Gets the item's drop callback.

    Returns:
        callback as a callable or None
    """
    return internal_dpg.get_item_configuration(item)["drop_callback"]


def get_item_user_data(item: Union[int, str]) -> Union[Any, None]:
    """Gets the item's callback data.

    Returns:
        callback data as a python object or None
    """
    return internal_dpg.get_item_configuration(item)["user_data"]


def get_item_source(item: Union[int, str]) -> Union[str, None]:
    """Gets the item's source.

    Returns:
        source as a string or None
    """
    return internal_dpg.get_item_configuration(item)["source"]


########################################################################################################################
# State Commands
########################################################################################################################

def is_item_hovered(item: Union[int, str]) -> Union[bool, None]:
    """Checks if item is hovered.

    Returns:
        status as a bool
    """
    return internal_dpg.get_item_state(item)["hovered"]


def is_item_active(item: Union[int, str]) -> Union[bool, None]:
    """Checks if item is active.

    Returns:
        status as a bool
    """
    return internal_dpg.get_item_state(item)["active"]


def is_item_focused(item: Union[int, str]) -> Union[bool, None]:
    """Checks if item is focused.

    Returns:
        status as a bool
    """
    return internal_dpg.get_item_state(item)["focused"]


def is_item_clicked(item: Union[int, str]) -> Union[bool, None]:
    """Checks if item is clicked.

    Returns:
        status as a bool
    """
    return internal_dpg.get_item_state(item)["clicked"]


def is_item_left_clicked(item: Union[int, str]) -> Union[bool, None]:
    """Checks if item is left clicked.

    Returns:
        status as a bool
    """
    return internal_dpg.get_item_state(item)["left_clicked"]


def is_item_right_clicked(item: Union[int, str]) -> Union[bool, None]:
    """Checks if item is right clicked.

    Returns:
        status as a bool
    """
    return internal_dpg.get_item_state(item)["right_clicked"]


def is_item_middle_clicked(item: Union[int, str]) -> Union[bool, None]:
    """Checks if item is middle clicked.

    Returns:
        status as a bool
    """
    return internal_dpg.get_item_state(item)["middle_clicked"]


def is_item_visible(item: Union[int, str]) -> Union[bool, None]:
    """Checks if item is visible.

    Returns:
        status as a bool
    """
    return internal_dpg.get_item_state(item)["visible"]


def is_item_edited(item: Union[int, str]) -> Union[bool, None]:
    """Checks if item is edited.

    Returns:
        status as a bool
    """
    return internal_dpg.get_item_state(item)["edited"]


def is_item_activated(item: Union[int, str]) -> Union[bool, None]:
    """Checks if item is activated.

    Returns:
        status as a bool
    """
    return internal_dpg.get_item_state(item)["activated"]


def is_item_deactivated(item: Union[int, str]) -> Union[bool, None]:
    """Checks if item is deactivated.

    Returns:
        status as a bool
    """
    return internal_dpg.get_item_state(item)["deactivated"]


def is_item_deactivated_after_edit(item: Union[int, str]) -> Union[bool, None]:
    """Checks if item is deactivated_after_edit.

    Returns:
        status as a bool
    """
    return internal_dpg.get_item_state(item)["deactivated_after_edit"]


def is_item_toggled_open(item: Union[int, str]) -> Union[bool, None]:
    """Checks if item is toggled_open.

    Returns:
        status as a bool
    """
    return internal_dpg.get_item_state(item)["toggled_open"]


def is_item_ok(item: Union[int, str]) -> Union[bool, None]:
    """Checks if item is ok and can be used.

    Returns:
        status as a bool
    """
    return internal_dpg.get_item_state(item)["ok"]


def is_item_shown(item: Union[int, str]) -> Union[bool, None]:
    """Checks if item is shown.

    Returns:
        status as a bool
    """
    return internal_dpg.get_item_configuration(item)["show"]


def is_item_enabled(item: Union[int, str]) -> Union[bool, None]:
    """Checks if item is enabled.

    Returns:
        status as a bool
    """
    return internal_dpg.get_item_configuration(item)["enabled"]


def get_item_pos(item: Union[int, str]) -> List[int]:
    """Returns item's position.

    Returns:
        position
    """
    return internal_dpg.get_item_state(item)["pos"]


def get_available_content_region(item: Union[int, str]) -> List[int]:
    """Returns item's available content region.

    Returns:
        position
    """
    return internal_dpg.get_item_state(item)["content_region_avail"]


def get_item_rect_size(item: Union[int, str]) -> List[int]:
    """Returns item's available content region.

    Returns:
        position
    """
    return internal_dpg.get_item_state(item)["rect_size"]


def get_item_rect_min(item: Union[int, str]) -> List[int]:
    """Returns item's minimum content region.

    Returns:
        position
    """
    return internal_dpg.get_item_state(item)["rect_min"]


def get_item_rect_max(item: Union[int, str]) -> List[int]:
    """Returns item's maximum content region.

    Returns:
        position
    """
    return internal_dpg.get_item_state(item)["rect_max"]


########################################################################################################################
# Viewport Setter Commands
########################################################################################################################

def set_viewport_clear_color(color: List[int]):
    """Sets the viewport's clear color.

    Returns:
        None
    """
    internal_dpg.configure_viewport(0, clear_color=color)


def set_viewport_small_icon(icon: str):
    """Sets the viewport's small icon. Must be ico for windows.

    Returns:
        None
    """
    internal_dpg.configure_viewport(0, small_icon=icon)


def set_viewport_large_icon(icon: str):
    """Sets the viewport's small icon. Must be ico for windows.

    Returns:
        None
    """
    internal_dpg.configure_viewport(0, large_icon=icon)


def set_viewport_pos(pos: List[float]):
    """Sets the viewport's position.

    Returns:
        None
    """
    internal_dpg.configure_viewport(0, x_pos=pos[0], y_pos=pos[1])


def set_viewport_width(width: int):
    """Sets the viewport's width.

    Returns:
        None
    """
    internal_dpg.configure_viewport(0, width=width)


def set_viewport_height(height: int):
    """Sets the viewport's height.

    Returns:
        None
    """
    internal_dpg.configure_viewport(0, height=height)


def set_viewport_min_width(width: int):
    """Sets the viewport's minimum width.

    Returns:
        None
    """
    internal_dpg.configure_viewport(0, min_width=width)


def set_viewport_max_width(width: int):
    """Sets the viewport's max width.

    Returns:
        None
    """
    internal_dpg.configure_viewport(0, max_width=width)


def set_viewport_min_height(height: int):
    """Sets the viewport's minimum height.

    Returns:
        None
    """
    internal_dpg.configure_viewport(0, min_height=height)


def set_viewport_max_height(height: int):
    """Sets the viewport's max width.

    Returns:
        None
    """
    internal_dpg.configure_viewport(0, max_height=height)


def set_viewport_title(title: str):
    """Sets the viewport's title.

    Returns:
        None
    """
    internal_dpg.configure_viewport(0, title=title)


def set_viewport_always_top(value: bool):
    """Sets the viewport always on top.

    Returns:
        None
    """
    internal_dpg.configure_viewport(0, always_on_top=value)


def set_viewport_resizable(value: bool):
    """Sets the viewport resizable.

    Returns:
        None
    """
    internal_dpg.configure_viewport(0, resizable=value)

def set_viewport_vsync(value: bool):
    """Sets the viewport vsync.

    Returns:
        None
    """
    internal_dpg.configure_viewport(0, vsync=value)


def set_viewport_decorated(value: bool):
    """Sets the viewport to be decorated.

    Returns:
        None
    """
    internal_dpg.configure_viewport(0, decorated=value)

########################################################################################################################
# Viewport Getter Commands
########################################################################################################################

def get_viewport_clear_color() ->List[int]:
    """Gets the viewport's clear color.

    Returns:
        List[int]
    """
    return internal_dpg.get_viewport_configuration()["clear_color"]


def get_viewport_pos() ->List[float]:
    """Gets the viewport's position.

    Returns:
        viewport position.
    """
    config = internal_dpg.get_viewport_configuration()
    x_pos = config["x_pos"]
    y_pos = config["y_pos"]
    return [x_pos, y_pos]


def get_viewport_width() -> int:
    """Gets the viewport's width.

    Returns:
        viewport width
    """
    return internal_dpg.get_viewport_configuration()["width"]


def get_viewport_client_width() -> int:
    """Gets the viewport's client width.

    Returns:
        viewport width
    """
    return internal_dpg.get_viewport_configuration()["client_width"]


def get_viewport_client_height() -> int:
    """Gets the viewport's client height.

    Returns:
        viewport width
    """
    return internal_dpg.get_viewport_configuration()["client_height"]


def get_viewport_height() -> int:
    """Gets the viewport's height.

    Returns:
        int
    """
    return internal_dpg.get_viewport_configuration()["height"]


def get_viewport_min_width() -> int:
    """Gets the viewport's minimum width.

    Returns:
        int
    """
    return internal_dpg.get_viewport_configuration()["min_width"]


def get_viewport_max_width() -> int:
    """Gets the viewport's max width.

    Returns:
        int
    """
    return internal_dpg.get_viewport_configuration()["max_width"]


def get_viewport_min_height() -> int:
    """Gets the viewport's minimum height.

    Returns:
        int
    """
    return internal_dpg.get_viewport_configuration()["min_height"]


def get_viewport_max_height() -> int:
    """Gets the viewport's max width.

    Returns:
        int
    """
    return internal_dpg.get_viewport_configuration()["max_height"]


def get_viewport_title() -> str:
    """Gets the viewport's title.

    Returns:
        str
    """
    return internal_dpg.get_viewport_configuration()["title"]


def is_viewport_always_top() -> bool:
    """Checks the viewport always on top flag.

    Returns:
        bool
    """
    return internal_dpg.get_viewport_configuration()["always_on_top"]


def is_viewport_resizable() -> bool:
    """Checks the viewport resizable flag.

    Returns:
        bool
    """
    return internal_dpg.get_viewport_configuration()["resizable"]


def is_viewport_vsync_on() -> bool:
    """Checks the viewport vsync flag.

    Returns:
        bool
    """
    return internal_dpg.get_viewport_configuration()["vsync"]


def is_viewport_decorated() -> bool:
    """Checks if the viewport is docorated.

    Returns:
        bool
    """
    return internal_dpg.get_viewport_configuration()["decorated"]

##########################################################
# Deprecated Commands
##########################################################

def deprecated(reason):

	string_types = (type(b''), type(u''))

	if isinstance(reason, string_types):

		def decorator(func1):

			fmt1 = "Call to deprecated function {name} ({reason})."

			@functools.wraps(func1)
			def new_func1(*args, **kwargs):
				warnings.simplefilter('always', DeprecationWarning)
				warnings.warn(
					fmt1.format(name=func1.__name__, reason=reason),
					category=DeprecationWarning,
					stacklevel=2
				)
				warnings.simplefilter('default', DeprecationWarning)
				return func1(*args, **kwargs)

			return new_func1

		return decorator

	elif inspect.isfunction(reason):

		func2 = reason
		fmt2 = "Call to deprecated function {name}."

		@functools.wraps(func2)
		def new_func2(*args, **kwargs):
			warnings.simplefilter('always', DeprecationWarning)
			warnings.warn(
				fmt2.format(name=func2.__name__),
				category=DeprecationWarning,
				stacklevel=2
			)
			warnings.simplefilter('default', DeprecationWarning)
			return func2(*args, **kwargs)

		return new_func2

@deprecated("Use 'configure_app(docking=True, docking_space=dock_space)'.")
def enable_docking(dock_space=False):
    """ deprecated function """
    internal_dpg.configure_app(docking=True, docking_space=dock_space)

@deprecated("Use 'configure_app(init_file=file)'.")
def set_init_file(file="dpg.ini"):
    """ deprecated function """
    internal_dpg.configure_app(init_file=file)

@deprecated("Use 'configure_app(init_file=file, load_init_file=True)'.")
def load_init_file(file):
    """ deprecated function """
    internal_dpg.configure_app(init_file=file, load_init_file=True)

@deprecated("Use: `is_viewport_ok(...)`")
def is_viewport_created():
    """ deprecated function """
    return internal_dpg.is_viewport_ok()

@deprecated("Use: \ncreate_viewport()\nsetup_dearpygui()\nshow_viewport()")
def setup_viewport():
    """ deprecated function """
    internal_dpg.create_viewport()
    internal_dpg.setup_dearpygui()
    internal_dpg.show_viewport()

@deprecated("Use: `bind_item_theme(...)`")
def set_item_theme(item, theme):
    """ deprecated function """
    return internal_dpg.bind_item_theme(item, theme)

@deprecated("Use: `bind_item_type_disabled_theme(...)`")
def set_item_type_disabled_theme(item, theme):
    """ deprecated function """
    return internal_dpg.bind_item_type_disabled_theme(item, theme)

@deprecated("Use: `bind_item_type_theme(...)`")
def set_item_type_theme(item, theme):
    """ deprecated function """
    return internal_dpg.bind_item_type_theme(item, theme)

@deprecated("Use: `bind_item_font(...)`")
def set_item_font(item, font):
    """ deprecated function """
    return internal_dpg.bind_item_font(item, font)

@deprecated("Use: `add_item_activated_handler(...)`")
def add_activated_handler(parent, **kwargs):
    """ deprecated function """
    return internal_dpg.add_item_activated_handler(parent, **kwargs)

@deprecated("Use: `add_item_active_handler(...)`")
def add_active_handler(parent, **kwargs):
    """ deprecated function """
    return internal_dpg.add_item_active_handler(parent, **kwargs)

@deprecated("Use: `add_item_clicked_handler(...)`")
def add_clicked_handler(parent, button=-1, **kwargs):
    """ deprecated function """
    return internal_dpg.add_item_clicked_handler(parent, button, **kwargs)

@deprecated("Use: `add_item_deactived_after_edit_handler(...)`")
def add_deactivated_after_edit_handler(parent, **kwargs):
    """ deprecated function """
    return internal_dpg.add_item_deactivated_after_edit_handler(parent, **kwargs)

@deprecated("Use: `add_item_deactivated_handler(...)`")
def add_deactivated_handler(parent, **kwargs):
    """ deprecated function """
    return internal_dpg.add_item_deactivated_handler(parent, **kwargs)

@deprecated("Use: `add_item_edited_handler(...)`")
def add_edited_handler(parent, **kwargs):
    """ deprecated function """
    return internal_dpg.add_item_edited_handler(parent, **kwargs)

@deprecated("Use: `add_item_focus_handler(...)`")
def add_focus_handler(parent, **kwargs):
    """ deprecated function """
    return internal_dpg.add_item_focus_handler(parent, **kwargs)

@deprecated("Use: `add_item_hover_handler(...)`")
def add_hover_handler(parent, **kwargs):
    """ deprecated function """
    return internal_dpg.add_item_hover_handler(parent, **kwargs)

@deprecated("Use: `add_item_resize_handler(...)`")
def add_resize_handler(parent, **kwargs):
    """ deprecated function """
    return internal_dpg.add_item_resize_handler(parent, **kwargs)

@deprecated("Use: `add_item_toggled_open_handler(...)`")
def add_toggled_open_handler(parent, **kwargs):
    """ deprecated function """
    return internal_dpg.add_item_toggled_open_handler(parent, **kwargs)

@deprecated("Use: `add_item_visible_handler(...)`")
def add_visible_handler(parent, **kwargs):
    """ deprecated function """
    return internal_dpg.add_item_visible_handler(parent, **kwargs)

@deprecated("Use: `bind_colormap(...)`")
def set_colormap(item, source):
    """ deprecated function """
    return internal_dpg.bind_colormap(item, source)

@deprecated("Use: `bind_theme(0)`")
def reset_default_theme(item, source):
    """ deprecated function """
    return internal_dpg.bind_theme(item, source)

@deprecated
def set_staging_mode(mode):
    """ deprecated function """
    pass

@deprecated
def add_table_next_column(**kwargs):
    """ deprecated function """
    pass

@deprecated("Use: add_stage")
def add_staging_container(**kwargs):
    """ deprecated function """
    return internal_dpg.add_stage(**kwargs)

@deprecated("Use: stage")
@contextmanager
def staging_container(**kwargs):
	"""
	deprecated function
	Args:
		**label (str): Overrides 'name' as label.
		**user_data (Any): User data for callbacks.
		**use_internal_label (bool): Use generated internal label instead of user specified (appends ### uuid).
		**id (Union[int, str]): Unique id used to programmatically refer to the item.If label is unused this will be the label.
	Yields:
		Union[int, str]
	"""
	try:
		warnings.warn("'staging_container' is deprecated and was changed to 'stage'", DeprecationWarning, 2)
		widget = internal_dpg.add_stage_container(**kwargs)
		internal_dpg.push_container_stack(widget)
		yield widget
	finally:
		internal_dpg.pop_container_stack()

@deprecated("Use: add_spacer(...)")
def add_spacing(**kwargs):
	"""	(deprecated function) Adds vertical spacing. 

	Args:
		label (str, optional): Overrides 'name' as label.
		user_data (Any, optional): User data for callbacks.
		use_internal_label (bool, optional): Use generated internal label instead of user specified (appends ### uuid).
		tag (Union[int, str], optional): Unique id used to programmatically refer to the item.If label is unused this will be the label.
		indent (int, optional): Offsets the widget to the right the specified number multiplied by the indent style.
		parent (Union[int, str], optional): Parent to add this item to. (runtime adding)
		before (Union[int, str], optional): This item will be displayed before the specified item in the parent.
		show (bool, optional): Attempt to render widget.
		pos (Union[List[int], Tuple[int]], optional): Places the item relative to window coordinates, [0,0] is top left.
		count (int, optional): Number of spacings to add the size is dependant on the curret style.
	Returns:
		Union[int, str]
	"""

	if 'count' in kwargs.keys():
		count = kwargs["count"]
		kwargs.pop("count", None)
		internal_dpg.add_group(**kwargs)
		internal_dpg.push_container_stack(internal_dpg.last_container())
		for i in range(count):
			internal_dpg.add_spacer()
		result_id = internal_dpg.pop_container_stack()
	else:
		result_id = internal_dpg.add_spacer(**kwargs)
	return result_id

@deprecated("Use: add_spacer(...)")
def add_dummy(**kwargs):
	"""	(deprecated function) Adds a spacer or 'dummy' object.

	Args:
		label (str, optional): Overrides 'name' as label.
		user_data (Any, optional): User data for callbacks.
		use_internal_label (bool, optional): Use generated internal label instead of user specified (appends ### uuid).
		tag (Union[int, str], optional): Unique id used to programmatically refer to the item.If label is unused this will be the label.
		width (int, optional): Width of the item.
		height (int, optional): Height of the item.
		indent (int, optional): Offsets the widget to the right the specified number multiplied by the indent style.
		parent (Union[int, str], optional): Parent to add this item to. (runtime adding)
		before (Union[int, str], optional): This item will be displayed before the specified item in the parent.
		show (bool, optional): Attempt to render widget.
		pos (Union[List[int], Tuple[int]], optional): Places the item relative to window coordinates, [0,0] is top left.
	Returns:
		Union[int, str]
	"""

	return internal_dpg.add_spacer(**kwargs)

@deprecated("Use: `destroy_context()`")
def cleanup_dearpygui():
    """ deprecated function """
    return internal_dpg.destroy_context()

@deprecated("Use: group(horizontal=True)")
def add_same_line(**kwargs):
    """ deprecated function """

    last_item = internal_dpg.last_item()
    group = internal_dpg.add_group(horizontal=True, **kwargs)
    internal_dpg.move_item(last_item, parent=group)
    internal_dpg.capture_next_item(lambda s: internal_dpg.move_item(s, parent=group))
    return group


@deprecated("Use: `add_child_window()`")
def add_child(**kwargs):
	"""	(deprecated function) Adds an embedded child window. Will show scrollbars when items do not fit.

	Args:
		label (str, optional): Overrides 'name' as label.
		user_data (Any, optional): User data for callbacks
		use_internal_label (bool, optional): Use generated internal label instead of user specified (appends ### uuid).
		tag (Union[int, str], optional): Unique id used to programmatically refer to the item.If label is unused this will be the label.
		width (int, optional): Width of the item.
		height (int, optional): Height of the item.
		indent (int, optional): Offsets the widget to the right the specified number multiplied by the indent style.
		parent (Union[int, str], optional): Parent to add this item to. (runtime adding)
		before (Union[int, str], optional): This item will be displayed before the specified item in the parent.
		payload_type (str, optional): Sender string type must be the same as the target for the target to run the payload_callback.
		drop_callback (Callable, optional): Registers a drop callback for drag and drop.
		show (bool, optional): Attempt to render widget.
		pos (Union[List[int], Tuple[int]], optional): Places the item relative to window coordinates, [0,0] is top left.
		filter_key (str, optional): Used by filter widget.
		delay_search (bool, optional): Delays searching container for specified items until the end of the app. Possible optimization when a container has many children that are not accessed often.
		tracked (bool, optional): Scroll tracking
		track_offset (float, optional): 0.0f:top, 0.5f:center, 1.0f:bottom
		border (bool, optional): Shows/Hides the border around the sides.
		autosize_x (bool, optional): Autosize the window to its parents size in x.
		autosize_y (bool, optional): Autosize the window to its parents size in y.
		no_scrollbar (bool, optional):  Disable scrollbars (window can still scroll with mouse or programmatically).
		horizontal_scrollbar (bool, optional): Allow horizontal scrollbar to appear (off by default).
		menubar (bool, optional): Shows/Hides the menubar at the top.
	Returns:
		Union[int, str]
	"""

	return internal_dpg.add_child_window(**kwargs)


@deprecated("Use: `child_window()`")
@contextmanager
def child(**kwargs):
	"""	(deprecated function) Adds an embedded child window. Will show scrollbars when items do not fit.

	Args:
		label (str, optional): Overrides 'name' as label.
		user_data (Any, optional): User data for callbacks
		use_internal_label (bool, optional): Use generated internal label instead of user specified (appends ### uuid).
		tag (Union[int, str], optional): Unique id used to programmatically refer to the item.If label is unused this will be the label.
		width (int, optional): Width of the item.
		height (int, optional): Height of the item.
		indent (int, optional): Offsets the widget to the right the specified number multiplied by the indent style.
		parent (Union[int, str], optional): Parent to add this item to. (runtime adding)
		before (Union[int, str], optional): This item will be displayed before the specified item in the parent.
		payload_type (str, optional): Sender string type must be the same as the target for the target to run the payload_callback.
		drop_callback (Callable, optional): Registers a drop callback for drag and drop.
		show (bool, optional): Attempt to render widget.
		pos (Union[List[int], Tuple[int]], optional): Places the item relative to window coordinates, [0,0] is top left.
		filter_key (str, optional): Used by filter widget.
		delay_search (bool, optional): Delays searching container for specified items until the end of the app. Possible optimization when a container has many children that are not accessed often.
		tracked (bool, optional): Scroll tracking
		track_offset (float, optional): 0.0f:top, 0.5f:center, 1.0f:bottom
		border (bool, optional): Shows/Hides the border around the sides.
		autosize_x (bool, optional): Autosize the window to its parents size in x.
		autosize_y (bool, optional): Autosize the window to its parents size in y.
		no_scrollbar (bool, optional):  Disable scrollbars (window can still scroll with mouse or programmatically).
		horizontal_scrollbar (bool, optional): Allow horizontal scrollbar to appear (off by default).
		menubar (bool, optional): Shows/Hides the menubar at the top.
	Yields:
		Union[int, str]
	"""
	try:
		widget = internal_dpg.add_child_window(**kwargs)
		internal_dpg.push_container_stack(widget)
		yield widget
	finally:
		internal_dpg.pop_container_stack()

@deprecated("Use: Just not recommended")
def setup_registries() -> None:
    """Adds default registries for fonts, handlers, textures, colormaps, and values."""
    internal_dpg.add_font_registry(tag=internal_dpg.mvReservedUUID_0)
    internal_dpg.add_handler_registry(tag=internal_dpg.mvReservedUUID_1)
    internal_dpg.add_texture_registry(tag=internal_dpg.mvReservedUUID_2)
    internal_dpg.add_value_registry(tag=internal_dpg.mvReservedUUID_3)
    internal_dpg.add_colormap_registry(tag=internal_dpg.mvReservedUUID_4)

@deprecated("Use: `set_frame_callback()`")
def set_start_callback(callback):
    """ deprecated function """
    return internal_dpg.set_frame_callback(3, callback)

##########################################################
# Container Context Managers
##########################################################


@contextmanager
def child_window(**kwargs):
	"""	 Adds an embedded child window. Will show scrollbars when items do not fit.

	Args:
		label (str, optional): Overrides 'name' as label.
		user_data (Any, optional): User data for callbacks
		use_internal_label (bool, optional): Use generated internal label instead of user specified (appends ### uuid).
		tag (Union[int, str], optional): Unique id used to programmatically refer to the item.If label is unused this will be the label.
		width (int, optional): Width of the item.
		height (int, optional): Height of the item.
		indent (int, optional): Offsets the widget to the right the specified number multiplied by the indent style.
		parent (Union[int, str], optional): Parent to add this item to. (runtime adding)
		before (Union[int, str], optional): This item will be displayed before the specified item in the parent.
		payload_type (str, optional): Sender string type must be the same as the target for the target to run the payload_callback.
		drop_callback (Callable, optional): Registers a drop callback for drag and drop.
		show (bool, optional): Attempt to render widget.
		pos (Union[List[int], Tuple[int, ...]], optional): Places the item relative to window coordinates, [0,0] is top left.
		filter_key (str, optional): Used by filter widget.
		delay_search (bool, optional): Delays searching container for specified items until the end of the app. Possible optimization when a container has many children that are not accessed often.
		tracked (bool, optional): Scroll tracking
		track_offset (float, optional): 0.0f:top, 0.5f:center, 1.0f:bottom
		border (bool, optional): Shows/Hides the border around the sides.
		autosize_x (bool, optional): Autosize the window to its parents size in x.
		autosize_y (bool, optional): Autosize the window to its parents size in y.
		no_scrollbar (bool, optional):  Disable scrollbars (window can still scroll with mouse or programmatically).
		horizontal_scrollbar (bool, optional): Allow horizontal scrollbar to appear (off by default).
		menubar (bool, optional): Shows/Hides the menubar at the top.
		id (Union[int, str], optional): (deprecated)
	Yields:
		Union[int, str]
	"""
	try:
		widget = internal_dpg.add_child_window(**kwargs)
		internal_dpg.push_container_stack(widget)
		yield widget
	finally:
		internal_dpg.pop_container_stack()

@contextmanager
def clipper(**kwargs):
	"""	 Helper to manually clip large list of items. Increases performance by not searching or drawing widgets outside of the clipped region.

	Args:
		label (str, optional): Overrides 'name' as label.
		user_data (Any, optional): User data for callbacks
		use_internal_label (bool, optional): Use generated internal label instead of user specified (appends ### uuid).
		tag (Union[int, str], optional): Unique id used to programmatically refer to the item.If label is unused this will be the label.
		width (int, optional): Width of the item.
		indent (int, optional): Offsets the widget to the right the specified number multiplied by the indent style.
		parent (Union[int, str], optional): Parent to add this item to. (runtime adding)
		before (Union[int, str], optional): This item will be displayed before the specified item in the parent.
		show (bool, optional): Attempt to render widget.
		delay_search (bool, optional): Delays searching container for specified items until the end of the app. Possible optimization when a container has many children that are not accessed often.
		id (Union[int, str], optional): (deprecated)
	Yields:
		Union[int, str]
	"""
	try:
		widget = internal_dpg.add_clipper(**kwargs)
		internal_dpg.push_container_stack(widget)
		yield widget
	finally:
		internal_dpg.pop_container_stack()

@contextmanager
def collapsing_header(**kwargs):
	"""	 Adds a collapsing header to add items to. Must be closed with the end command.

	Args:
		label (str, optional): Overrides 'name' as label.
		user_data (Any, optional): User data for callbacks
		use_internal_label (bool, optional): Use generated internal label instead of user specified (appends ### uuid).
		tag (Union[int, str], optional): Unique id used to programmatically refer to the item.If label is unused this will be the label.
		indent (int, optional): Offsets the widget to the right the specified number multiplied by the indent style.
		parent (Union[int, str], optional): Parent to add this item to. (runtime adding)
		before (Union[int, str], optional): This item will be displayed before the specified item in the parent.
		payload_type (str, optional): Sender string type must be the same as the target for the target to run the payload_callback.
		drag_callback (Callable, optional): Registers a drag callback for drag and drop.
		drop_callback (Callable, optional): Registers a drop callback for drag and drop.
		show (bool, optional): Attempt to render widget.
		pos (Union[List[int], Tuple[int, ...]], optional): Places the item relative to window coordinates, [0,0] is top left.
		filter_key (str, optional): Used by filter widget.
		delay_search (bool, optional): Delays searching container for specified items until the end of the app. Possible optimization when a container has many children that are not accessed often.
		tracked (bool, optional): Scroll tracking
		track_offset (float, optional): 0.0f:top, 0.5f:center, 1.0f:bottom
		closable (bool, optional): Adds the ability to hide this widget by pressing the (x) in the top right of widget.
		default_open (bool, optional): Sets the collapseable header open by default.
		open_on_double_click (bool, optional): Need double-click to open node.
		open_on_arrow (bool, optional): Only open when clicking on the arrow part.
		leaf (bool, optional): No collapsing, no arrow (use as a convenience for leaf nodes).
		bullet (bool, optional): Display a bullet instead of arrow.
		id (Union[int, str], optional): (deprecated)
	Yields:
		Union[int, str]
	"""
	try:
		widget = internal_dpg.add_collapsing_header(**kwargs)
		internal_dpg.push_container_stack(widget)
		yield widget
	finally:
		internal_dpg.pop_container_stack()

@contextmanager
def colormap_registry(**kwargs):
	"""	 Adds a colormap registry.

	Args:
		label (str, optional): Overrides 'name' as label.
		user_data (Any, optional): User data for callbacks
		use_internal_label (bool, optional): Use generated internal label instead of user specified (appends ### uuid).
		tag (Union[int, str], optional): Unique id used to programmatically refer to the item.If label is unused this will be the label.
		show (bool, optional): Attempt to render widget.
		id (Union[int, str], optional): (deprecated)
	Yields:
		Union[int, str]
	"""
	try:
		widget = internal_dpg.add_colormap_registry(**kwargs)
		internal_dpg.push_container_stack(widget)
		yield widget
	finally:
		internal_dpg.pop_container_stack()

@contextmanager
def drag_payload(**kwargs):
	"""	 User data payload for drag and drop operations.

	Args:
		label (str, optional): Overrides 'name' as label.
		user_data (Any, optional): User data for callbacks
		use_internal_label (bool, optional): Use generated internal label instead of user specified (appends ### uuid).
		tag (Union[int, str], optional): Unique id used to programmatically refer to the item.If label is unused this will be the label.
		parent (Union[int, str], optional): Parent to add this item to. (runtime adding)
		show (bool, optional): Attempt to render widget.
		drag_data (Any, optional): Drag data
		drop_data (Any, optional): Drop data
		payload_type (str, optional): 
		id (Union[int, str], optional): (deprecated)
	Yields:
		Union[int, str]
	"""
	try:
		widget = internal_dpg.add_drag_payload(**kwargs)
		internal_dpg.push_container_stack(widget)
		yield widget
	finally:
		internal_dpg.pop_container_stack()

@contextmanager
def draw_layer(**kwargs):
	"""	 New in 1.1. Creates a layer useful for grouping drawlist items.

	Args:
		label (str, optional): Overrides 'name' as label.
		user_data (Any, optional): User data for callbacks
		use_internal_label (bool, optional): Use generated internal label instead of user specified (appends ### uuid).
		tag (Union[int, str], optional): Unique id used to programmatically refer to the item.If label is unused this will be the label.
		parent (Union[int, str], optional): Parent to add this item to. (runtime adding)
		before (Union[int, str], optional): This item will be displayed before the specified item in the parent.
		show (bool, optional): Attempt to render widget.
		perspective_divide (bool, optional): New in 1.1. apply perspective divide
		depth_clipping (bool, optional): New in 1.1. apply depth clipping
		cull_mode (int, optional): New in 1.1. culling mode, mvCullMode_* constants. Only works with triangles currently.
		id (Union[int, str], optional): (deprecated)
	Yields:
		Union[int, str]
	"""
	try:
		widget = internal_dpg.add_draw_layer(**kwargs)
		internal_dpg.push_container_stack(widget)
		yield widget
	finally:
		internal_dpg.pop_container_stack()

@contextmanager
def draw_node(**kwargs):
	"""	 New in 1.1. Creates a drawing node to associate a transformation matrix. Child node matricies will concatenate.

	Args:
		label (str, optional): Overrides 'name' as label.
		user_data (Any, optional): User data for callbacks
		use_internal_label (bool, optional): Use generated internal label instead of user specified (appends ### uuid).
		tag (Union[int, str], optional): Unique id used to programmatically refer to the item.If label is unused this will be the label.
		parent (Union[int, str], optional): Parent to add this item to. (runtime adding)
		before (Union[int, str], optional): This item will be displayed before the specified item in the parent.
		show (bool, optional): Attempt to render widget.
		id (Union[int, str], optional): (deprecated)
	Yields:
		Union[int, str]
	"""
	try:
		widget = internal_dpg.add_draw_node(**kwargs)
		internal_dpg.push_container_stack(widget)
		yield widget
	finally:
		internal_dpg.pop_container_stack()

@contextmanager
def drawlist(width, height, **kwargs):
	"""	 Adds a drawing canvas.

	Args:
		width (int): 
		height (int): 
		label (str, optional): Overrides 'name' as label.
		user_data (Any, optional): User data for callbacks
		use_internal_label (bool, optional): Use generated internal label instead of user specified (appends ### uuid).
		tag (Union[int, str], optional): Unique id used to programmatically refer to the item.If label is unused this will be the label.
		parent (Union[int, str], optional): Parent to add this item to. (runtime adding)
		before (Union[int, str], optional): This item will be displayed before the specified item in the parent.
		callback (Callable, optional): Registers a callback.
		show (bool, optional): Attempt to render widget.
		pos (Union[List[int], Tuple[int, ...]], optional): Places the item relative to window coordinates, [0,0] is top left.
		filter_key (str, optional): Used by filter widget.
		delay_search (bool, optional): Delays searching container for specified items until the end of the app. Possible optimization when a container has many children that are not accessed often.
		tracked (bool, optional): Scroll tracking
		track_offset (float, optional): 0.0f:top, 0.5f:center, 1.0f:bottom
		id (Union[int, str], optional): (deprecated)
	Yields:
		Union[int, str]
	"""
	try:
		widget = internal_dpg.add_drawlist(width, height, **kwargs)
		internal_dpg.push_container_stack(widget)
		yield widget
	finally:
		internal_dpg.pop_container_stack()

@contextmanager
def file_dialog(**kwargs):
	"""	 Displays a file or directory selector depending on keywords. Displays a file dialog by default. Callback will be ran when the file or directory picker is closed. The app_data arguemnt will be populated with information related to the file and directory as a dictionary.

	Args:
		label (str, optional): Overrides 'name' as label.
		user_data (Any, optional): User data for callbacks
		use_internal_label (bool, optional): Use generated internal label instead of user specified (appends ### uuid).
		tag (Union[int, str], optional): Unique id used to programmatically refer to the item.If label is unused this will be the label.
		width (int, optional): Width of the item.
		height (int, optional): Height of the item.
		callback (Callable, optional): Registers a callback.
		show (bool, optional): Attempt to render widget.
		default_path (str, optional): Path that the file dialog will default to when opened.
		default_filename (str, optional): Default name that will show in the file name input.
		file_count (int, optional): Number of visible files in the dialog.
		modal (bool, optional): Forces user interaction with the file selector.
		directory_selector (bool, optional): Shows only directory/paths as options. Allows selection of directory/paths only.
		min_size (Union[List[int], Tuple[int, ...]], optional): Minimum window size.
		max_size (Union[List[int], Tuple[int, ...]], optional): Maximum window size.
		id (Union[int, str], optional): (deprecated)
	Yields:
		Union[int, str]
	"""
	try:
		widget = internal_dpg.add_file_dialog(**kwargs)
		internal_dpg.push_container_stack(widget)
		yield widget
	finally:
		internal_dpg.pop_container_stack()

@contextmanager
def filter_set(**kwargs):
	"""	 Helper to parse and apply text filters (e.g. aaaaa[, bbbbb][, ccccc])

	Args:
		label (str, optional): Overrides 'name' as label.
		user_data (Any, optional): User data for callbacks
		use_internal_label (bool, optional): Use generated internal label instead of user specified (appends ### uuid).
		tag (Union[int, str], optional): Unique id used to programmatically refer to the item.If label is unused this will be the label.
		width (int, optional): Width of the item.
		indent (int, optional): Offsets the widget to the right the specified number multiplied by the indent style.
		parent (Union[int, str], optional): Parent to add this item to. (runtime adding)
		before (Union[int, str], optional): This item will be displayed before the specified item in the parent.
		show (bool, optional): Attempt to render widget.
		delay_search (bool, optional): Delays searching container for specified items until the end of the app. Possible optimization when a container has many children that are not accessed often.
		id (Union[int, str], optional): (deprecated)
	Yields:
		Union[int, str]
	"""
	try:
		widget = internal_dpg.add_filter_set(**kwargs)
		internal_dpg.push_container_stack(widget)
		yield widget
	finally:
		internal_dpg.pop_container_stack()

@contextmanager
def font(file, size, **kwargs):
	"""	 Adds font to a font registry.

	Args:
		file (str): 
		size (int): 
		label (str, optional): Overrides 'name' as label.
		user_data (Any, optional): User data for callbacks
		use_internal_label (bool, optional): Use generated internal label instead of user specified (appends ### uuid).
		tag (Union[int, str], optional): Unique id used to programmatically refer to the item.If label is unused this will be the label.
		parent (Union[int, str], optional): Parent to add this item to. (runtime adding)
		id (Union[int, str], optional): (deprecated)
		default_font (bool, optional): (deprecated)
	Yields:
		Union[int, str]
	"""
	try:
		widget = internal_dpg.add_font(file, size, **kwargs)
		internal_dpg.push_container_stack(widget)
		yield widget
	finally:
		internal_dpg.pop_container_stack()

@contextmanager
def font_registry(**kwargs):
	"""	 Adds a font registry.

	Args:
		label (str, optional): Overrides 'name' as label.
		user_data (Any, optional): User data for callbacks
		use_internal_label (bool, optional): Use generated internal label instead of user specified (appends ### uuid).
		tag (Union[int, str], optional): Unique id used to programmatically refer to the item.If label is unused this will be the label.
		show (bool, optional): Attempt to render widget.
		id (Union[int, str], optional): (deprecated)
	Yields:
		Union[int, str]
	"""
	try:
		widget = internal_dpg.add_font_registry(**kwargs)
		internal_dpg.push_container_stack(widget)
		yield widget
	finally:
		internal_dpg.pop_container_stack()

@contextmanager
def group(**kwargs):
	"""	 Creates a group that other widgets can belong to. The group allows item commands to be issued for all of its members.

	Args:
		label (str, optional): Overrides 'name' as label.
		user_data (Any, optional): User data for callbacks
		use_internal_label (bool, optional): Use generated internal label instead of user specified (appends ### uuid).
		tag (Union[int, str], optional): Unique id used to programmatically refer to the item.If label is unused this will be the label.
		width (int, optional): Width of the item.
		indent (int, optional): Offsets the widget to the right the specified number multiplied by the indent style.
		parent (Union[int, str], optional): Parent to add this item to. (runtime adding)
		before (Union[int, str], optional): This item will be displayed before the specified item in the parent.
		payload_type (str, optional): Sender string type must be the same as the target for the target to run the payload_callback.
		drag_callback (Callable, optional): Registers a drag callback for drag and drop.
		drop_callback (Callable, optional): Registers a drop callback for drag and drop.
		show (bool, optional): Attempt to render widget.
		pos (Union[List[int], Tuple[int, ...]], optional): Places the item relative to window coordinates, [0,0] is top left.
		filter_key (str, optional): Used by filter widget.
		delay_search (bool, optional): Delays searching container for specified items until the end of the app. Possible optimization when a container has many children that are not accessed often.
		tracked (bool, optional): Scroll tracking
		track_offset (float, optional): 0.0f:top, 0.5f:center, 1.0f:bottom
		horizontal (bool, optional): Forces child widgets to be added in a horizontal layout.
		horizontal_spacing (float, optional): Spacing for the horizontal layout.
		xoffset (float, optional): Offset from containing window x item location within group.
		id (Union[int, str], optional): (deprecated)
	Yields:
		Union[int, str]
	"""
	try:
		widget = internal_dpg.add_group(**kwargs)
		internal_dpg.push_container_stack(widget)
		yield widget
	finally:
		internal_dpg.pop_container_stack()

@contextmanager
def handler_registry(**kwargs):
	"""	 Adds a handler registry.

	Args:
		label (str, optional): Overrides 'name' as label.
		user_data (Any, optional): User data for callbacks
		use_internal_label (bool, optional): Use generated internal label instead of user specified (appends ### uuid).
		tag (Union[int, str], optional): Unique id used to programmatically refer to the item.If label is unused this will be the label.
		show (bool, optional): Attempt to render widget.
		id (Union[int, str], optional): (deprecated)
	Yields:
		Union[int, str]
	"""
	try:
		widget = internal_dpg.add_handler_registry(**kwargs)
		internal_dpg.push_container_stack(widget)
		yield widget
	finally:
		internal_dpg.pop_container_stack()

@contextmanager
def item_handler_registry(**kwargs):
	"""	 Adds an item handler registry.

	Args:
		label (str, optional): Overrides 'name' as label.
		user_data (Any, optional): User data for callbacks
		use_internal_label (bool, optional): Use generated internal label instead of user specified (appends ### uuid).
		tag (Union[int, str], optional): Unique id used to programmatically refer to the item.If label is unused this will be the label.
		show (bool, optional): Attempt to render widget.
		id (Union[int, str], optional): (deprecated)
	Yields:
		Union[int, str]
	"""
	try:
		widget = internal_dpg.add_item_handler_registry(**kwargs)
		internal_dpg.push_container_stack(widget)
		yield widget
	finally:
		internal_dpg.pop_container_stack()

@contextmanager
def menu(**kwargs):
	"""	 Adds a menu to an existing menu bar.

	Args:
		label (str, optional): Overrides 'name' as label.
		user_data (Any, optional): User data for callbacks
		use_internal_label (bool, optional): Use generated internal label instead of user specified (appends ### uuid).
		tag (Union[int, str], optional): Unique id used to programmatically refer to the item.If label is unused this will be the label.
		indent (int, optional): Offsets the widget to the right the specified number multiplied by the indent style.
		parent (Union[int, str], optional): Parent to add this item to. (runtime adding)
		before (Union[int, str], optional): This item will be displayed before the specified item in the parent.
		payload_type (str, optional): Sender string type must be the same as the target for the target to run the payload_callback.
		drop_callback (Callable, optional): Registers a drop callback for drag and drop.
		show (bool, optional): Attempt to render widget.
		enabled (bool, optional): Turns off functionality of widget and applies the disabled theme.
		filter_key (str, optional): Used by filter widget.
		delay_search (bool, optional): Delays searching container for specified items until the end of the app. Possible optimization when a container has many children that are not accessed often.
		tracked (bool, optional): Scroll tracking
		track_offset (float, optional): 0.0f:top, 0.5f:center, 1.0f:bottom
		id (Union[int, str], optional): (deprecated)
	Yields:
		Union[int, str]
	"""
	try:
		widget = internal_dpg.add_menu(**kwargs)
		internal_dpg.push_container_stack(widget)
		yield widget
	finally:
		internal_dpg.pop_container_stack()

@contextmanager
def menu_bar(**kwargs):
	"""	 Adds a menu bar to a window.

	Args:
		label (str, optional): Overrides 'name' as label.
		user_data (Any, optional): User data for callbacks
		use_internal_label (bool, optional): Use generated internal label instead of user specified (appends ### uuid).
		tag (Union[int, str], optional): Unique id used to programmatically refer to the item.If label is unused this will be the label.
		indent (int, optional): Offsets the widget to the right the specified number multiplied by the indent style.
		parent (Union[int, str], optional): Parent to add this item to. (runtime adding)
		show (bool, optional): Attempt to render widget.
		delay_search (bool, optional): Delays searching container for specified items until the end of the app. Possible optimization when a container has many children that are not accessed often.
		id (Union[int, str], optional): (deprecated)
	Yields:
		Union[int, str]
	"""
	try:
		widget = internal_dpg.add_menu_bar(**kwargs)
		internal_dpg.push_container_stack(widget)
		yield widget
	finally:
		internal_dpg.pop_container_stack()

@contextmanager
def node(**kwargs):
	"""	 Adds a node to a node editor.

	Args:
		label (str, optional): Overrides 'name' as label.
		user_data (Any, optional): User data for callbacks
		use_internal_label (bool, optional): Use generated internal label instead of user specified (appends ### uuid).
		tag (Union[int, str], optional): Unique id used to programmatically refer to the item.If label is unused this will be the label.
		parent (Union[int, str], optional): Parent to add this item to. (runtime adding)
		before (Union[int, str], optional): This item will be displayed before the specified item in the parent.
		payload_type (str, optional): Sender string type must be the same as the target for the target to run the payload_callback.
		drag_callback (Callable, optional): Registers a drag callback for drag and drop.
		drop_callback (Callable, optional): Registers a drop callback for drag and drop.
		show (bool, optional): Attempt to render widget.
		pos (Union[List[int], Tuple[int, ...]], optional): Places the item relative to window coordinates, [0,0] is top left.
		filter_key (str, optional): Used by filter widget.
		delay_search (bool, optional): Delays searching container for specified items until the end of the app. Possible optimization when a container has many children that are not accessed often.
		tracked (bool, optional): Scroll tracking
		track_offset (float, optional): 0.0f:top, 0.5f:center, 1.0f:bottom
		draggable (bool, optional): Allow node to be draggable.
		id (Union[int, str], optional): (deprecated)
	Yields:
		Union[int, str]
	"""
	try:
		widget = internal_dpg.add_node(**kwargs)
		internal_dpg.push_container_stack(widget)
		yield widget
	finally:
		internal_dpg.pop_container_stack()

@contextmanager
def node_attribute(**kwargs):
	"""	 Adds a node attribute to a node.

	Args:
		label (str, optional): Overrides 'name' as label.
		user_data (Any, optional): User data for callbacks
		use_internal_label (bool, optional): Use generated internal label instead of user specified (appends ### uuid).
		tag (Union[int, str], optional): Unique id used to programmatically refer to the item.If label is unused this will be the label.
		indent (int, optional): Offsets the widget to the right the specified number multiplied by the indent style.
		parent (Union[int, str], optional): Parent to add this item to. (runtime adding)
		before (Union[int, str], optional): This item will be displayed before the specified item in the parent.
		show (bool, optional): Attempt to render widget.
		filter_key (str, optional): Used by filter widget.
		tracked (bool, optional): Scroll tracking
		track_offset (float, optional): 0.0f:top, 0.5f:center, 1.0f:bottom
		attribute_type (int, optional): mvNode_Attr_Input, mvNode_Attr_Output, or mvNode_Attr_Static.
		shape (int, optional): Pin shape.
		category (str, optional): Category
		id (Union[int, str], optional): (deprecated)
	Yields:
		Union[int, str]
	"""
	try:
		widget = internal_dpg.add_node_attribute(**kwargs)
		internal_dpg.push_container_stack(widget)
		yield widget
	finally:
		internal_dpg.pop_container_stack()

@contextmanager
def node_editor(**kwargs):
	"""	 Adds a node editor.

	Args:
		label (str, optional): Overrides 'name' as label.
		user_data (Any, optional): User data for callbacks
		use_internal_label (bool, optional): Use generated internal label instead of user specified (appends ### uuid).
		tag (Union[int, str], optional): Unique id used to programmatically refer to the item.If label is unused this will be the label.
		width (int, optional): Width of the item.
		height (int, optional): Height of the item.
		parent (Union[int, str], optional): Parent to add this item to. (runtime adding)
		before (Union[int, str], optional): This item will be displayed before the specified item in the parent.
		callback (Callable, optional): Registers a callback.
		show (bool, optional): Attempt to render widget.
		filter_key (str, optional): Used by filter widget.
		delay_search (bool, optional): Delays searching container for specified items until the end of the app. Possible optimization when a container has many children that are not accessed often.
		tracked (bool, optional): Scroll tracking
		track_offset (float, optional): 0.0f:top, 0.5f:center, 1.0f:bottom
		delink_callback (Callable, optional): Callback ran when a link is detached.
		menubar (bool, optional): Shows or hides the menubar.
		id (Union[int, str], optional): (deprecated)
	Yields:
		Union[int, str]
	"""
	try:
		widget = internal_dpg.add_node_editor(**kwargs)
		internal_dpg.push_container_stack(widget)
		yield widget
	finally:
		internal_dpg.pop_container_stack()

@contextmanager
def plot(**kwargs):
	"""	 Adds a plot which is used to hold series, and can be drawn to with draw commands.

	Args:
		label (str, optional): Overrides 'name' as label.
		user_data (Any, optional): User data for callbacks
		use_internal_label (bool, optional): Use generated internal label instead of user specified (appends ### uuid).
		tag (Union[int, str], optional): Unique id used to programmatically refer to the item.If label is unused this will be the label.
		width (int, optional): Width of the item.
		height (int, optional): Height of the item.
		indent (int, optional): Offsets the widget to the right the specified number multiplied by the indent style.
		parent (Union[int, str], optional): Parent to add this item to. (runtime adding)
		before (Union[int, str], optional): This item will be displayed before the specified item in the parent.
		payload_type (str, optional): Sender string type must be the same as the target for the target to run the payload_callback.
		callback (Callable, optional): Registers a callback.
		drag_callback (Callable, optional): Registers a drag callback for drag and drop.
		drop_callback (Callable, optional): Registers a drop callback for drag and drop.
		show (bool, optional): Attempt to render widget.
		pos (Union[List[int], Tuple[int, ...]], optional): Places the item relative to window coordinates, [0,0] is top left.
		filter_key (str, optional): Used by filter widget.
		delay_search (bool, optional): Delays searching container for specified items until the end of the app. Possible optimization when a container has many children that are not accessed often.
		tracked (bool, optional): Scroll tracking
		track_offset (float, optional): 0.0f:top, 0.5f:center, 1.0f:bottom
		no_title (bool, optional): 
		no_menus (bool, optional): 
		no_box_select (bool, optional): 
		no_mouse_pos (bool, optional): 
		no_highlight (bool, optional): 
		no_child (bool, optional): 
		query (bool, optional): 
		crosshairs (bool, optional): 
		anti_aliased (bool, optional): 
		equal_aspects (bool, optional): 
		pan_button (int, optional): enables panning when held
		pan_mod (int, optional): optional modifier that must be held for panning
		fit_button (int, optional): fits visible data when double clicked
		context_menu_button (int, optional): opens plot context menu (if enabled) when clicked
		box_select_button (int, optional): begins box selection when pressed and confirms selection when released
		box_select_mod (int, optional): begins box selection when pressed and confirms selection when released
		box_select_cancel_button (int, optional): cancels active box selection when pressed
		query_button (int, optional): begins query selection when pressed and end query selection when released
		query_mod (int, optional): optional modifier that must be held for query selection
		query_toggle_mod (int, optional): when held, active box selections turn into queries
		horizontal_mod (int, optional): expands active box selection/query horizontally to plot edge when held
		vertical_mod (int, optional): expands active box selection/query vertically to plot edge when held
		id (Union[int, str], optional): (deprecated)
	Yields:
		Union[int, str]
	"""
	try:
		widget = internal_dpg.add_plot(**kwargs)
		internal_dpg.push_container_stack(widget)
		yield widget
	finally:
		internal_dpg.pop_container_stack()

@contextmanager
def plot_axis(axis, **kwargs):
	"""	 Adds an axis to a plot.

	Args:
		axis (int): 
		label (str, optional): Overrides 'name' as label.
		user_data (Any, optional): User data for callbacks
		use_internal_label (bool, optional): Use generated internal label instead of user specified (appends ### uuid).
		tag (Union[int, str], optional): Unique id used to programmatically refer to the item.If label is unused this will be the label.
		parent (Union[int, str], optional): Parent to add this item to. (runtime adding)
		payload_type (str, optional): Sender string type must be the same as the target for the target to run the payload_callback.
		drop_callback (Callable, optional): Registers a drop callback for drag and drop.
		show (bool, optional): Attempt to render widget.
		no_gridlines (bool, optional): 
		no_tick_marks (bool, optional): 
		no_tick_labels (bool, optional): 
		log_scale (bool, optional): 
		invert (bool, optional): 
		lock_min (bool, optional): 
		lock_max (bool, optional): 
		time (bool, optional): 
		id (Union[int, str], optional): (deprecated)
	Yields:
		Union[int, str]
	"""
	try:
		widget = internal_dpg.add_plot_axis(axis, **kwargs)
		internal_dpg.push_container_stack(widget)
		yield widget
	finally:
		internal_dpg.pop_container_stack()

@contextmanager
def stage(**kwargs):
	"""	 Adds a stage.

	Args:
		label (str, optional): Overrides 'name' as label.
		user_data (Any, optional): User data for callbacks
		use_internal_label (bool, optional): Use generated internal label instead of user specified (appends ### uuid).
		tag (Union[int, str], optional): Unique id used to programmatically refer to the item.If label is unused this will be the label.
		id (Union[int, str], optional): (deprecated)
	Yields:
		Union[int, str]
	"""
	try:
		widget = internal_dpg.add_stage(**kwargs)
		internal_dpg.push_container_stack(widget)
		yield widget
	finally:
		internal_dpg.pop_container_stack()

@contextmanager
def subplots(rows, columns, **kwargs):
	"""	 Adds a collection of plots.

	Args:
		rows (int): 
		columns (int): 
		label (str, optional): Overrides 'name' as label.
		user_data (Any, optional): User data for callbacks
		use_internal_label (bool, optional): Use generated internal label instead of user specified (appends ### uuid).
		tag (Union[int, str], optional): Unique id used to programmatically refer to the item.If label is unused this will be the label.
		width (int, optional): Width of the item.
		height (int, optional): Height of the item.
		indent (int, optional): Offsets the widget to the right the specified number multiplied by the indent style.
		parent (Union[int, str], optional): Parent to add this item to. (runtime adding)
		before (Union[int, str], optional): This item will be displayed before the specified item in the parent.
		callback (Callable, optional): Registers a callback.
		show (bool, optional): Attempt to render widget.
		pos (Union[List[int], Tuple[int, ...]], optional): Places the item relative to window coordinates, [0,0] is top left.
		filter_key (str, optional): Used by filter widget.
		delay_search (bool, optional): Delays searching container for specified items until the end of the app. Possible optimization when a container has many children that are not accessed often.
		tracked (bool, optional): Scroll tracking
		track_offset (float, optional): 0.0f:top, 0.5f:center, 1.0f:bottom
		row_ratios (Union[List[float], Tuple[float, ...]], optional): 
		column_ratios (Union[List[float], Tuple[float, ...]], optional): 
		no_title (bool, optional): 
		no_menus (bool, optional): the user will not be able to open context menus with right-click
		no_resize (bool, optional): resize splitters between subplot cells will be not be provided
		no_align (bool, optional): subplot edges will not be aligned vertically or horizontally
		link_rows (bool, optional): link the y-axis limits of all plots in each row (does not apply auxiliary y-axes)
		link_columns (bool, optional): link the x-axis limits of all plots in each column
		link_all_x (bool, optional): link the x-axis limits in every plot in the subplot
		link_all_y (bool, optional): link the y-axis limits in every plot in the subplot (does not apply to auxiliary y-axes)
		column_major (bool, optional): subplots are added in column major order instead of the default row major order
		id (Union[int, str], optional): (deprecated)
	Yields:
		Union[int, str]
	"""
	try:
		widget = internal_dpg.add_subplots(rows, columns, **kwargs)
		internal_dpg.push_container_stack(widget)
		yield widget
	finally:
		internal_dpg.pop_container_stack()

@contextmanager
def tab(**kwargs):
	"""	 Adds a tab to a tab bar.

	Args:
		label (str, optional): Overrides 'name' as label.
		user_data (Any, optional): User data for callbacks
		use_internal_label (bool, optional): Use generated internal label instead of user specified (appends ### uuid).
		tag (Union[int, str], optional): Unique id used to programmatically refer to the item.If label is unused this will be the label.
		indent (int, optional): Offsets the widget to the right the specified number multiplied by the indent style.
		parent (Union[int, str], optional): Parent to add this item to. (runtime adding)
		before (Union[int, str], optional): This item will be displayed before the specified item in the parent.
		payload_type (str, optional): Sender string type must be the same as the target for the target to run the payload_callback.
		drop_callback (Callable, optional): Registers a drop callback for drag and drop.
		show (bool, optional): Attempt to render widget.
		filter_key (str, optional): Used by filter widget.
		delay_search (bool, optional): Delays searching container for specified items until the end of the app. Possible optimization when a container has many children that are not accessed often.
		tracked (bool, optional): Scroll tracking
		track_offset (float, optional): 0.0f:top, 0.5f:center, 1.0f:bottom
		closable (bool, optional): Creates a button on the tab that can hide the tab.
		no_tooltip (bool, optional): Disable tooltip for the given tab.
		order_mode (bool, optional): set using a constant: mvTabOrder_Reorderable: allows reordering, mvTabOrder_Fixed: fixed ordering, mvTabOrder_Leading: adds tab to front, mvTabOrder_Trailing: adds tab to back
		id (Union[int, str], optional): (deprecated)
	Yields:
		Union[int, str]
	"""
	try:
		widget = internal_dpg.add_tab(**kwargs)
		internal_dpg.push_container_stack(widget)
		yield widget
	finally:
		internal_dpg.pop_container_stack()

@contextmanager
def tab_bar(**kwargs):
	"""	 Adds a tab bar.

	Args:
		label (str, optional): Overrides 'name' as label.
		user_data (Any, optional): User data for callbacks
		use_internal_label (bool, optional): Use generated internal label instead of user specified (appends ### uuid).
		tag (Union[int, str], optional): Unique id used to programmatically refer to the item.If label is unused this will be the label.
		indent (int, optional): Offsets the widget to the right the specified number multiplied by the indent style.
		parent (Union[int, str], optional): Parent to add this item to. (runtime adding)
		before (Union[int, str], optional): This item will be displayed before the specified item in the parent.
		callback (Callable, optional): Registers a callback.
		show (bool, optional): Attempt to render widget.
		pos (Union[List[int], Tuple[int, ...]], optional): Places the item relative to window coordinates, [0,0] is top left.
		filter_key (str, optional): Used by filter widget.
		delay_search (bool, optional): Delays searching container for specified items until the end of the app. Possible optimization when a container has many children that are not accessed often.
		tracked (bool, optional): Scroll tracking
		track_offset (float, optional): 0.0f:top, 0.5f:center, 1.0f:bottom
		reorderable (bool, optional): Allows for the user to change the order of the tabs.
		id (Union[int, str], optional): (deprecated)
	Yields:
		Union[int, str]
	"""
	try:
		widget = internal_dpg.add_tab_bar(**kwargs)
		internal_dpg.push_container_stack(widget)
		yield widget
	finally:
		internal_dpg.pop_container_stack()

@contextmanager
def table(**kwargs):
	"""	 Adds a table.

	Args:
		label (str, optional): Overrides 'name' as label.
		user_data (Any, optional): User data for callbacks
		use_internal_label (bool, optional): Use generated internal label instead of user specified (appends ### uuid).
		tag (Union[int, str], optional): Unique id used to programmatically refer to the item.If label is unused this will be the label.
		width (int, optional): Width of the item.
		height (int, optional): Height of the item.
		indent (int, optional): Offsets the widget to the right the specified number multiplied by the indent style.
		parent (Union[int, str], optional): Parent to add this item to. (runtime adding)
		before (Union[int, str], optional): This item will be displayed before the specified item in the parent.
		source (Union[int, str], optional): Overrides 'id' as value storage key.
		callback (Callable, optional): Registers a callback.
		show (bool, optional): Attempt to render widget.
		pos (Union[List[int], Tuple[int, ...]], optional): Places the item relative to window coordinates, [0,0] is top left.
		filter_key (str, optional): Used by filter widget.
		delay_search (bool, optional): Delays searching container for specified items until the end of the app. Possible optimization when a container has many children that are not accessed often.
		header_row (bool, optional): show headers at the top of the columns
		clipper (bool, optional): Use clipper (rows must be same height).
		inner_width (int, optional): 
		policy (int, optional): 
		freeze_rows (int, optional): 
		freeze_columns (int, optional): 
		sort_multi (bool, optional): Hold shift when clicking headers to sort on multiple column.
		sort_tristate (bool, optional): Allow no sorting, disable default sorting.
		resizable (bool, optional): Enable resizing columns
		reorderable (bool, optional): Enable reordering columns in header row (need calling TableSetupColumn() + TableHeadersRow() to display headers)
		hideable (bool, optional): Enable hiding/disabling columns in context menu.
		sortable (bool, optional): Enable sorting. Call TableGetSortSpecs() to obtain sort specs. Also see ImGuiTableFlags_SortMulti and ImGuiTableFlags_SortTristate.
		context_menu_in_body (bool, optional): Right-click on columns body/contents will display table context menu. By default it is available in TableHeadersRow().
		row_background (bool, optional): Set each RowBg color with ImGuiCol_TableRowBg or ImGuiCol_TableRowBgAlt (equivalent of calling TableSetBgColor with ImGuiTableBgFlags_RowBg0 on each row manually)
		borders_innerH (bool, optional): Draw horizontal borders between rows.
		borders_outerH (bool, optional): Draw horizontal borders at the top and bottom.
		borders_innerV (bool, optional): Draw vertical borders between columns.
		borders_outerV (bool, optional): Draw vertical borders on the left and right sides.
		no_host_extendX (bool, optional): Make outer width auto-fit to columns, overriding outer_size.x value. Only available when ScrollX/ScrollY are disabled and Stretch columns are not used.
		no_host_extendY (bool, optional): Make outer height stop exactly at outer_size.y (prevent auto-extending table past the limit). Only available when ScrollX/ScrollY are disabled. Data below the limit will be clipped and not visible.
		no_keep_columns_visible (bool, optional): Disable keeping column always minimally visible when ScrollX is off and table gets too small. Not recommended if columns are resizable.
		precise_widths (bool, optional): Disable distributing remainder width to stretched columns (width allocation on a 100-wide table with 3 columns: Without this flag: 33,33,34. With this flag: 33,33,33). With larger number of columns, resizing will appear to be less smooth.
		no_clip (bool, optional): Disable clipping rectangle for every individual columns.
		pad_outerX (bool, optional): Default if BordersOuterV is on. Enable outer-most padding. Generally desirable if you have headers.
		no_pad_outerX (bool, optional): Default if BordersOuterV is off. Disable outer-most padding.
		no_pad_innerX (bool, optional): Disable inner padding between columns (double inner padding if BordersOuterV is on, single inner padding if BordersOuterV is off).
		scrollX (bool, optional): Enable horizontal scrolling. Require 'outer_size' parameter of BeginTable() to specify the container size. Changes default sizing policy. Because this create a child window, ScrollY is currently generally recommended when using ScrollX.
		scrollY (bool, optional): Enable vertical scrolling.
		no_saved_settings (bool, optional): Never load/save settings in .ini file.
		id (Union[int, str], optional): (deprecated)
	Yields:
		Union[int, str]
	"""
	try:
		widget = internal_dpg.add_table(**kwargs)
		internal_dpg.push_container_stack(widget)
		yield widget
	finally:
		internal_dpg.pop_container_stack()

@contextmanager
def table_cell(**kwargs):
	"""	 Adds a table.

	Args:
		label (str, optional): Overrides 'name' as label.
		user_data (Any, optional): User data for callbacks
		use_internal_label (bool, optional): Use generated internal label instead of user specified (appends ### uuid).
		tag (Union[int, str], optional): Unique id used to programmatically refer to the item.If label is unused this will be the label.
		height (int, optional): Height of the item.
		parent (Union[int, str], optional): Parent to add this item to. (runtime adding)
		before (Union[int, str], optional): This item will be displayed before the specified item in the parent.
		show (bool, optional): Attempt to render widget.
		filter_key (str, optional): Used by filter widget.
		id (Union[int, str], optional): (deprecated)
	Yields:
		Union[int, str]
	"""
	try:
		widget = internal_dpg.add_table_cell(**kwargs)
		internal_dpg.push_container_stack(widget)
		yield widget
	finally:
		internal_dpg.pop_container_stack()

@contextmanager
def table_row(**kwargs):
	"""	 Adds a table row.

	Args:
		label (str, optional): Overrides 'name' as label.
		user_data (Any, optional): User data for callbacks
		use_internal_label (bool, optional): Use generated internal label instead of user specified (appends ### uuid).
		tag (Union[int, str], optional): Unique id used to programmatically refer to the item.If label is unused this will be the label.
		height (int, optional): Height of the item.
		parent (Union[int, str], optional): Parent to add this item to. (runtime adding)
		before (Union[int, str], optional): This item will be displayed before the specified item in the parent.
		show (bool, optional): Attempt to render widget.
		filter_key (str, optional): Used by filter widget.
		id (Union[int, str], optional): (deprecated)
	Yields:
		Union[int, str]
	"""
	try:
		widget = internal_dpg.add_table_row(**kwargs)
		internal_dpg.push_container_stack(widget)
		yield widget
	finally:
		internal_dpg.pop_container_stack()

@contextmanager
def template_registry(**kwargs):
	"""	 Adds a template registry.

	Args:
		label (str, optional): Overrides 'name' as label.
		user_data (Any, optional): User data for callbacks
		use_internal_label (bool, optional): Use generated internal label instead of user specified (appends ### uuid).
		tag (Union[int, str], optional): Unique id used to programmatically refer to the item.If label is unused this will be the label.
		id (Union[int, str], optional): (deprecated)
	Yields:
		Union[int, str]
	"""
	try:
		widget = internal_dpg.add_template_registry(**kwargs)
		internal_dpg.push_container_stack(widget)
		yield widget
	finally:
		internal_dpg.pop_container_stack()

@contextmanager
def texture_registry(**kwargs):
	"""	 Adds a dynamic texture.

	Args:
		label (str, optional): Overrides 'name' as label.
		user_data (Any, optional): User data for callbacks
		use_internal_label (bool, optional): Use generated internal label instead of user specified (appends ### uuid).
		tag (Union[int, str], optional): Unique id used to programmatically refer to the item.If label is unused this will be the label.
		show (bool, optional): Attempt to render widget.
		id (Union[int, str], optional): (deprecated)
	Yields:
		Union[int, str]
	"""
	try:
		widget = internal_dpg.add_texture_registry(**kwargs)
		internal_dpg.push_container_stack(widget)
		yield widget
	finally:
		internal_dpg.pop_container_stack()

@contextmanager
def theme(**kwargs):
	"""	 Adds a theme.

	Args:
		label (str, optional): Overrides 'name' as label.
		user_data (Any, optional): User data for callbacks
		use_internal_label (bool, optional): Use generated internal label instead of user specified (appends ### uuid).
		tag (Union[int, str], optional): Unique id used to programmatically refer to the item.If label is unused this will be the label.
		id (Union[int, str], optional): (deprecated)
		default_theme (bool, optional): (deprecated)
	Yields:
		Union[int, str]
	"""
	try:
		widget = internal_dpg.add_theme(**kwargs)
		internal_dpg.push_container_stack(widget)
		yield widget
	finally:
		internal_dpg.pop_container_stack()

@contextmanager
def theme_component(item_type=0, **kwargs):
	"""	 Adds a theme component.

	Args:
		item_type (int, optional): 
		label (str, optional): Overrides 'name' as label.
		user_data (Any, optional): User data for callbacks
		use_internal_label (bool, optional): Use generated internal label instead of user specified (appends ### uuid).
		tag (Union[int, str], optional): Unique id used to programmatically refer to the item.If label is unused this will be the label.
		parent (Union[int, str], optional): Parent to add this item to. (runtime adding)
		before (Union[int, str], optional): This item will be displayed before the specified item in the parent.
		enabled_state (bool, optional): 
		id (Union[int, str], optional): (deprecated)
	Yields:
		Union[int, str]
	"""
	try:
		widget = internal_dpg.add_theme_component(item_type, **kwargs)
		internal_dpg.push_container_stack(widget)
		yield widget
	finally:
		internal_dpg.pop_container_stack()

@contextmanager
def tooltip(parent, **kwargs):
	"""	 Adds a tooltip window.

	Args:
		parent (Union[int, str]): 
		label (str, optional): Overrides 'name' as label.
		user_data (Any, optional): User data for callbacks
		use_internal_label (bool, optional): Use generated internal label instead of user specified (appends ### uuid).
		tag (Union[int, str], optional): Unique id used to programmatically refer to the item.If label is unused this will be the label.
		show (bool, optional): Attempt to render widget.
		id (Union[int, str], optional): (deprecated)
	Yields:
		Union[int, str]
	"""
	try:
		widget = internal_dpg.add_tooltip(parent, **kwargs)
		internal_dpg.push_container_stack(widget)
		yield widget
	finally:
		internal_dpg.pop_container_stack()

@contextmanager
def tree_node(**kwargs):
	"""	 Adds a tree node to add items to.

	Args:
		label (str, optional): Overrides 'name' as label.
		user_data (Any, optional): User data for callbacks
		use_internal_label (bool, optional): Use generated internal label instead of user specified (appends ### uuid).
		tag (Union[int, str], optional): Unique id used to programmatically refer to the item.If label is unused this will be the label.
		indent (int, optional): Offsets the widget to the right the specified number multiplied by the indent style.
		parent (Union[int, str], optional): Parent to add this item to. (runtime adding)
		before (Union[int, str], optional): This item will be displayed before the specified item in the parent.
		payload_type (str, optional): Sender string type must be the same as the target for the target to run the payload_callback.
		drag_callback (Callable, optional): Registers a drag callback for drag and drop.
		drop_callback (Callable, optional): Registers a drop callback for drag and drop.
		show (bool, optional): Attempt to render widget.
		pos (Union[List[int], Tuple[int, ...]], optional): Places the item relative to window coordinates, [0,0] is top left.
		filter_key (str, optional): Used by filter widget.
		delay_search (bool, optional): Delays searching container for specified items until the end of the app. Possible optimization when a container has many children that are not accessed often.
		tracked (bool, optional): Scroll tracking
		track_offset (float, optional): 0.0f:top, 0.5f:center, 1.0f:bottom
		default_open (bool, optional): Sets the tree node open by default.
		open_on_double_click (bool, optional): Need double-click to open node.
		open_on_arrow (bool, optional): Only open when clicking on the arrow part.
		leaf (bool, optional): No collapsing, no arrow (use as a convenience for leaf nodes).
		bullet (bool, optional): Display a bullet instead of arrow.
		selectable (bool, optional): Makes the tree selectable.
		id (Union[int, str], optional): (deprecated)
	Yields:
		Union[int, str]
	"""
	try:
		widget = internal_dpg.add_tree_node(**kwargs)
		internal_dpg.push_container_stack(widget)
		yield widget
	finally:
		internal_dpg.pop_container_stack()

@contextmanager
def value_registry(**kwargs):
	"""	 Adds a value registry.

	Args:
		label (str, optional): Overrides 'name' as label.
		user_data (Any, optional): User data for callbacks
		use_internal_label (bool, optional): Use generated internal label instead of user specified (appends ### uuid).
		tag (Union[int, str], optional): Unique id used to programmatically refer to the item.If label is unused this will be the label.
		id (Union[int, str], optional): (deprecated)
	Yields:
		Union[int, str]
	"""
	try:
		widget = internal_dpg.add_value_registry(**kwargs)
		internal_dpg.push_container_stack(widget)
		yield widget
	finally:
		internal_dpg.pop_container_stack()

@contextmanager
def viewport_drawlist(**kwargs):
	"""	 A container that is used to present draw items or layers directly to the viewport. By default this will draw to the back of the viewport. Layers and draw items should be added to this widget as children.

	Args:
		label (str, optional): Overrides 'name' as label.
		user_data (Any, optional): User data for callbacks
		use_internal_label (bool, optional): Use generated internal label instead of user specified (appends ### uuid).
		tag (Union[int, str], optional): Unique id used to programmatically refer to the item.If label is unused this will be the label.
		show (bool, optional): Attempt to render widget.
		filter_key (str, optional): Used by filter widget.
		delay_search (bool, optional): Delays searching container for specified items until the end of the app. Possible optimization when a container has many children that are not accessed often.
		front (bool, optional): Draws to the front of the view port instead of the back.
		id (Union[int, str], optional): (deprecated)
	Yields:
		Union[int, str]
	"""
	try:
		widget = internal_dpg.add_viewport_drawlist(**kwargs)
		internal_dpg.push_container_stack(widget)
		yield widget
	finally:
		internal_dpg.pop_container_stack()

@contextmanager
def viewport_menu_bar(**kwargs):
	"""	 Adds a menubar to the viewport.

	Args:
		label (str, optional): Overrides 'name' as label.
		user_data (Any, optional): User data for callbacks
		use_internal_label (bool, optional): Use generated internal label instead of user specified (appends ### uuid).
		tag (Union[int, str], optional): Unique id used to programmatically refer to the item.If label is unused this will be the label.
		indent (int, optional): Offsets the widget to the right the specified number multiplied by the indent style.
		parent (Union[int, str], optional): Parent to add this item to. (runtime adding)
		show (bool, optional): Attempt to render widget.
		delay_search (bool, optional): Delays searching container for specified items until the end of the app. Possible optimization when a container has many children that are not accessed often.
		id (Union[int, str], optional): (deprecated)
	Yields:
		Union[int, str]
	"""
	try:
		widget = internal_dpg.add_viewport_menu_bar(**kwargs)
		internal_dpg.push_container_stack(widget)
		yield widget
	finally:
		internal_dpg.pop_container_stack()

@contextmanager
def window(**kwargs):
	"""	 Creates a new window for following items to be added to.

	Args:
		label (str, optional): Overrides 'name' as label.
		user_data (Any, optional): User data for callbacks
		use_internal_label (bool, optional): Use generated internal label instead of user specified (appends ### uuid).
		tag (Union[int, str], optional): Unique id used to programmatically refer to the item.If label is unused this will be the label.
		width (int, optional): Width of the item.
		height (int, optional): Height of the item.
		indent (int, optional): Offsets the widget to the right the specified number multiplied by the indent style.
		show (bool, optional): Attempt to render widget.
		pos (Union[List[int], Tuple[int, ...]], optional): Places the item relative to window coordinates, [0,0] is top left.
		delay_search (bool, optional): Delays searching container for specified items until the end of the app. Possible optimization when a container has many children that are not accessed often.
		min_size (Union[List[int], Tuple[int, ...]], optional): Minimum window size.
		max_size (Union[List[int], Tuple[int, ...]], optional): Maximum window size.
		menubar (bool, optional): Shows or hides the menubar.
		collapsed (bool, optional): Collapse the window.
		autosize (bool, optional): Autosized the window to fit it's items.
		no_resize (bool, optional): Allows for the window size to be changed or fixed.
		no_title_bar (bool, optional): Title name for the title bar of the window.
		no_move (bool, optional): Allows for the window's position to be changed or fixed.
		no_scrollbar (bool, optional):  Disable scrollbars. (window can still scroll with mouse or programmatically)
		no_collapse (bool, optional): Disable user collapsing window by double-clicking on it.
		horizontal_scrollbar (bool, optional): Allow horizontal scrollbar to appear. (off by default)
		no_focus_on_appearing (bool, optional): Disable taking focus when transitioning from hidden to visible state.
		no_bring_to_front_on_focus (bool, optional): Disable bringing window to front when taking focus. (e.g. clicking on it or programmatically giving it focus)
		no_close (bool, optional): Disable user closing the window by removing the close button.
		no_background (bool, optional): Sets Background and border alpha to transparent.
		modal (bool, optional): Fills area behind window according to the theme and disables user ability to interact with anything except the window.
		popup (bool, optional): Fills area behind window according to the theme, removes title bar, collapse and close. Window can be closed by selecting area in the background behind the window.
		no_saved_settings (bool, optional): Never load/save settings in .ini file.
		no_open_over_existing_popup (bool, optional): Don't open if there's already a popup
		on_close (Callable, optional): Callback ran when window is closed.
		id (Union[int, str], optional): (deprecated)
	Yields:
		Union[int, str]
	"""
	try:
		widget = internal_dpg.add_window(**kwargs)
		internal_dpg.push_container_stack(widget)
		yield widget
	finally:
		internal_dpg.pop_container_stack()

##########################################################
# Core Wrappings
##########################################################

def add_2d_histogram_series(x, y, **kwargs):
	"""	 Adds a 2d histogram series.

	Args:
		x (Any): 
		y (Any): 
		label (str, optional): Overrides 'name' as label.
		user_data (Any, optional): User data for callbacks
		use_internal_label (bool, optional): Use generated internal label instead of user specified (appends ### uuid).
		tag (Union[int, str], optional): Unique id used to programmatically refer to the item.If label is unused this will be the label.
		parent (Union[int, str], optional): Parent to add this item to. (runtime adding)
		before (Union[int, str], optional): This item will be displayed before the specified item in the parent.
		source (Union[int, str], optional): Overrides 'id' as value storage key.
		show (bool, optional): Attempt to render widget.
		xbins (int, optional): 
		ybins (int, optional): 
		xmin_range (float, optional): 
		xmax_range (float, optional): 
		ymin_range (float, optional): 
		ymax_range (float, optional): 
		density (bool, optional): 
		outliers (bool, optional): 
		id (Union[int, str], optional): (deprecated)
	Returns:
		Union[int, str]
	"""

	return internal_dpg.add_2d_histogram_series(x, y, **kwargs)

def add_3d_slider(**kwargs):
	"""	 Adds a 3D box slider.

	Args:
		label (str, optional): Overrides 'name' as label.
		user_data (Any, optional): User data for callbacks
		use_internal_label (bool, optional): Use generated internal label instead of user specified (appends ### uuid).
		tag (Union[int, str], optional): Unique id used to programmatically refer to the item.If label is unused this will be the label.
		width (int, optional): Width of the item.
		height (int, optional): Height of the item.
		indent (int, optional): Offsets the widget to the right the specified number multiplied by the indent style.
		parent (Union[int, str], optional): Parent to add this item to. (runtime adding)
		before (Union[int, str], optional): This item will be displayed before the specified item in the parent.
		source (Union[int, str], optional): Overrides 'id' as value storage key.
		payload_type (str, optional): Sender string type must be the same as the target for the target to run the payload_callback.
		callback (Callable, optional): Registers a callback.
		drag_callback (Callable, optional): Registers a drag callback for drag and drop.
		drop_callback (Callable, optional): Registers a drop callback for drag and drop.
		show (bool, optional): Attempt to render widget.
		pos (Union[List[int], Tuple[int, ...]], optional): Places the item relative to window coordinates, [0,0] is top left.
		filter_key (str, optional): Used by filter widget.
		tracked (bool, optional): Scroll tracking
		track_offset (float, optional): 0.0f:top, 0.5f:center, 1.0f:bottom
		default_value (Union[List[float], Tuple[float, ...]], optional): 
		max_x (float, optional): Applies upper limit to slider.
		max_y (float, optional): Applies upper limit to slider.
		max_z (float, optional): Applies upper limit to slider.
		min_x (float, optional): Applies lower limit to slider.
		min_y (float, optional): Applies lower limit to slider.
		min_z (float, optional): Applies lower limit to slider.
		scale (float, optional): Size of the widget.
		id (Union[int, str], optional): (deprecated)
	Returns:
		Union[int, str]
	"""

	return internal_dpg.add_3d_slider(**kwargs)

def add_alias(alias, item):
	"""	 Adds an alias.

	Args:
		alias (str): 
		item (Union[int, str]): 
	Returns:
		None
	"""

	return internal_dpg.add_alias(alias, item)

def add_area_series(x, y, **kwargs):
	"""	 Adds an area series to a plot.

	Args:
		x (Any): 
		y (Any): 
		label (str, optional): Overrides 'name' as label.
		user_data (Any, optional): User data for callbacks
		use_internal_label (bool, optional): Use generated internal label instead of user specified (appends ### uuid).
		tag (Union[int, str], optional): Unique id used to programmatically refer to the item.If label is unused this will be the label.
		parent (Union[int, str], optional): Parent to add this item to. (runtime adding)
		before (Union[int, str], optional): This item will be displayed before the specified item in the parent.
		source (Union[int, str], optional): Overrides 'id' as value storage key.
		show (bool, optional): Attempt to render widget.
		fill (Union[List[int], Tuple[int, ...]], optional): 
		contribute_to_bounds (bool, optional): 
		id (Union[int, str], optional): (deprecated)
	Returns:
		Union[int, str]
	"""

	return internal_dpg.add_area_series(x, y, **kwargs)

def add_bar_series(x, y, **kwargs):
	"""	 Adds a bar series to a plot.

	Args:
		x (Any): 
		y (Any): 
		label (str, optional): Overrides 'name' as label.
		user_data (Any, optional): User data for callbacks
		use_internal_label (bool, optional): Use generated internal label instead of user specified (appends ### uuid).
		tag (Union[int, str], optional): Unique id used to programmatically refer to the item.If label is unused this will be the label.
		parent (Union[int, str], optional): Parent to add this item to. (runtime adding)
		before (Union[int, str], optional): This item will be displayed before the specified item in the parent.
		source (Union[int, str], optional): Overrides 'id' as value storage key.
		show (bool, optional): Attempt to render widget.
		weight (float, optional): 
		horizontal (bool, optional): 
		id (Union[int, str], optional): (deprecated)
	Returns:
		Union[int, str]
	"""

	return internal_dpg.add_bar_series(x, y, **kwargs)

def add_bool_value(**kwargs):
	"""	 Adds a bool value.

	Args:
		label (str, optional): Overrides 'name' as label.
		user_data (Any, optional): User data for callbacks
		use_internal_label (bool, optional): Use generated internal label instead of user specified (appends ### uuid).
		tag (Union[int, str], optional): Unique id used to programmatically refer to the item.If label is unused this will be the label.
		source (Union[int, str], optional): Overrides 'id' as value storage key.
		default_value (bool, optional): 
		parent (Union[int, str], optional): Parent to add this item to. (runtime adding)
		id (Union[int, str], optional): (deprecated)
	Returns:
		Union[int, str]
	"""

	return internal_dpg.add_bool_value(**kwargs)

def add_button(**kwargs):
	"""	 Adds a button.

	Args:
		label (str, optional): Overrides 'name' as label.
		user_data (Any, optional): User data for callbacks
		use_internal_label (bool, optional): Use generated internal label instead of user specified (appends ### uuid).
		tag (Union[int, str], optional): Unique id used to programmatically refer to the item.If label is unused this will be the label.
		width (int, optional): Width of the item.
		height (int, optional): Height of the item.
		indent (int, optional): Offsets the widget to the right the specified number multiplied by the indent style.
		parent (Union[int, str], optional): Parent to add this item to. (runtime adding)
		before (Union[int, str], optional): This item will be displayed before the specified item in the parent.
		payload_type (str, optional): Sender string type must be the same as the target for the target to run the payload_callback.
		callback (Callable, optional): Registers a callback.
		drag_callback (Callable, optional): Registers a drag callback for drag and drop.
		drop_callback (Callable, optional): Registers a drop callback for drag and drop.
		show (bool, optional): Attempt to render widget.
		enabled (bool, optional): Turns off functionality of widget and applies the disabled theme.
		pos (Union[List[int], Tuple[int, ...]], optional): Places the item relative to window coordinates, [0,0] is top left.
		filter_key (str, optional): Used by filter widget.
		tracked (bool, optional): Scroll tracking
		track_offset (float, optional): 0.0f:top, 0.5f:center, 1.0f:bottom
		small (bool, optional): Shrinks the size of the button to the text of the label it contains. Useful for embedding in text.
		arrow (bool, optional): Displays an arrow in place of the text string. This requires the direction keyword.
		direction (int, optional): Sets the cardinal direction for the arrow buy using constants mvDir_Left, mvDir_Up, mvDir_Down, mvDir_Right, mvDir_None. Arrow keyword must be set to True.
		id (Union[int, str], optional): (deprecated)
	Returns:
		Union[int, str]
	"""

	return internal_dpg.add_button(**kwargs)

def add_candle_series(dates, opens, closes, lows, highs, **kwargs):
	"""	 Adds a candle series to a plot.

	Args:
		dates (Any): 
		opens (Any): 
		closes (Any): 
		lows (Any): 
		highs (Any): 
		label (str, optional): Overrides 'name' as label.
		user_data (Any, optional): User data for callbacks
		use_internal_label (bool, optional): Use generated internal label instead of user specified (appends ### uuid).
		tag (Union[int, str], optional): Unique id used to programmatically refer to the item.If label is unused this will be the label.
		parent (Union[int, str], optional): Parent to add this item to. (runtime adding)
		before (Union[int, str], optional): This item will be displayed before the specified item in the parent.
		source (Union[int, str], optional): Overrides 'id' as value storage key.
		show (bool, optional): Attempt to render widget.
		bull_color (Union[List[int], Tuple[int, ...]], optional): 
		bear_color (Union[List[int], Tuple[int, ...]], optional): 
		weight (int, optional): 
		tooltip (bool, optional): 
		id (Union[int, str], optional): (deprecated)
	Returns:
		Union[int, str]
	"""

	return internal_dpg.add_candle_series(dates, opens, closes, lows, highs, **kwargs)

def add_char_remap(source, target, **kwargs):
	"""	 Remaps a character.

	Args:
		source (int): 
		target (int): 
		label (str, optional): Overrides 'name' as label.
		user_data (Any, optional): User data for callbacks
		use_internal_label (bool, optional): Use generated internal label instead of user specified (appends ### uuid).
		tag (Union[int, str], optional): Unique id used to programmatically refer to the item.If label is unused this will be the label.
		parent (Union[int, str], optional): Parent to add this item to. (runtime adding)
		id (Union[int, str], optional): (deprecated)
	Returns:
		Union[int, str]
	"""

	return internal_dpg.add_char_remap(source, target, **kwargs)

def add_checkbox(**kwargs):
	"""	 Adds a checkbox.

	Args:
		label (str, optional): Overrides 'name' as label.
		user_data (Any, optional): User data for callbacks
		use_internal_label (bool, optional): Use generated internal label instead of user specified (appends ### uuid).
		tag (Union[int, str], optional): Unique id used to programmatically refer to the item.If label is unused this will be the label.
		indent (int, optional): Offsets the widget to the right the specified number multiplied by the indent style.
		parent (Union[int, str], optional): Parent to add this item to. (runtime adding)
		before (Union[int, str], optional): This item will be displayed before the specified item in the parent.
		source (Union[int, str], optional): Overrides 'id' as value storage key.
		payload_type (str, optional): Sender string type must be the same as the target for the target to run the payload_callback.
		callback (Callable, optional): Registers a callback.
		drag_callback (Callable, optional): Registers a drag callback for drag and drop.
		drop_callback (Callable, optional): Registers a drop callback for drag and drop.
		show (bool, optional): Attempt to render widget.
		enabled (bool, optional): Turns off functionality of widget and applies the disabled theme.
		pos (Union[List[int], Tuple[int, ...]], optional): Places the item relative to window coordinates, [0,0] is top left.
		filter_key (str, optional): Used by filter widget.
		tracked (bool, optional): Scroll tracking
		track_offset (float, optional): 0.0f:top, 0.5f:center, 1.0f:bottom
		default_value (bool, optional): Sets the default value of the checkmark
		id (Union[int, str], optional): (deprecated)
	Returns:
		Union[int, str]
	"""

	return internal_dpg.add_checkbox(**kwargs)

def add_child_window(**kwargs):
	"""	 Adds an embedded child window. Will show scrollbars when items do not fit.

	Args:
		label (str, optional): Overrides 'name' as label.
		user_data (Any, optional): User data for callbacks
		use_internal_label (bool, optional): Use generated internal label instead of user specified (appends ### uuid).
		tag (Union[int, str], optional): Unique id used to programmatically refer to the item.If label is unused this will be the label.
		width (int, optional): Width of the item.
		height (int, optional): Height of the item.
		indent (int, optional): Offsets the widget to the right the specified number multiplied by the indent style.
		parent (Union[int, str], optional): Parent to add this item to. (runtime adding)
		before (Union[int, str], optional): This item will be displayed before the specified item in the parent.
		payload_type (str, optional): Sender string type must be the same as the target for the target to run the payload_callback.
		drop_callback (Callable, optional): Registers a drop callback for drag and drop.
		show (bool, optional): Attempt to render widget.
		pos (Union[List[int], Tuple[int, ...]], optional): Places the item relative to window coordinates, [0,0] is top left.
		filter_key (str, optional): Used by filter widget.
		delay_search (bool, optional): Delays searching container for specified items until the end of the app. Possible optimization when a container has many children that are not accessed often.
		tracked (bool, optional): Scroll tracking
		track_offset (float, optional): 0.0f:top, 0.5f:center, 1.0f:bottom
		border (bool, optional): Shows/Hides the border around the sides.
		autosize_x (bool, optional): Autosize the window to its parents size in x.
		autosize_y (bool, optional): Autosize the window to its parents size in y.
		no_scrollbar (bool, optional):  Disable scrollbars (window can still scroll with mouse or programmatically).
		horizontal_scrollbar (bool, optional): Allow horizontal scrollbar to appear (off by default).
		menubar (bool, optional): Shows/Hides the menubar at the top.
		id (Union[int, str], optional): (deprecated)
	Returns:
		Union[int, str]
	"""

	return internal_dpg.add_child_window(**kwargs)

def add_clipper(**kwargs):
	"""	 Helper to manually clip large list of items. Increases performance by not searching or drawing widgets outside of the clipped region.

	Args:
		label (str, optional): Overrides 'name' as label.
		user_data (Any, optional): User data for callbacks
		use_internal_label (bool, optional): Use generated internal label instead of user specified (appends ### uuid).
		tag (Union[int, str], optional): Unique id used to programmatically refer to the item.If label is unused this will be the label.
		width (int, optional): Width of the item.
		indent (int, optional): Offsets the widget to the right the specified number multiplied by the indent style.
		parent (Union[int, str], optional): Parent to add this item to. (runtime adding)
		before (Union[int, str], optional): This item will be displayed before the specified item in the parent.
		show (bool, optional): Attempt to render widget.
		delay_search (bool, optional): Delays searching container for specified items until the end of the app. Possible optimization when a container has many children that are not accessed often.
		id (Union[int, str], optional): (deprecated)
	Returns:
		Union[int, str]
	"""

	return internal_dpg.add_clipper(**kwargs)

def add_collapsing_header(**kwargs):
	"""	 Adds a collapsing header to add items to. Must be closed with the end command.

	Args:
		label (str, optional): Overrides 'name' as label.
		user_data (Any, optional): User data for callbacks
		use_internal_label (bool, optional): Use generated internal label instead of user specified (appends ### uuid).
		tag (Union[int, str], optional): Unique id used to programmatically refer to the item.If label is unused this will be the label.
		indent (int, optional): Offsets the widget to the right the specified number multiplied by the indent style.
		parent (Union[int, str], optional): Parent to add this item to. (runtime adding)
		before (Union[int, str], optional): This item will be displayed before the specified item in the parent.
		payload_type (str, optional): Sender string type must be the same as the target for the target to run the payload_callback.
		drag_callback (Callable, optional): Registers a drag callback for drag and drop.
		drop_callback (Callable, optional): Registers a drop callback for drag and drop.
		show (bool, optional): Attempt to render widget.
		pos (Union[List[int], Tuple[int, ...]], optional): Places the item relative to window coordinates, [0,0] is top left.
		filter_key (str, optional): Used by filter widget.
		delay_search (bool, optional): Delays searching container for specified items until the end of the app. Possible optimization when a container has many children that are not accessed often.
		tracked (bool, optional): Scroll tracking
		track_offset (float, optional): 0.0f:top, 0.5f:center, 1.0f:bottom
		closable (bool, optional): Adds the ability to hide this widget by pressing the (x) in the top right of widget.
		default_open (bool, optional): Sets the collapseable header open by default.
		open_on_double_click (bool, optional): Need double-click to open node.
		open_on_arrow (bool, optional): Only open when clicking on the arrow part.
		leaf (bool, optional): No collapsing, no arrow (use as a convenience for leaf nodes).
		bullet (bool, optional): Display a bullet instead of arrow.
		id (Union[int, str], optional): (deprecated)
	Returns:
		Union[int, str]
	"""

	return internal_dpg.add_collapsing_header(**kwargs)

def add_color_button(default_value=(0, 0, 0, 255), **kwargs):
	"""	 Adds a color button.

	Args:
		default_value (Union[List[int], Tuple[int, ...]], optional): 
		label (str, optional): Overrides 'name' as label.
		user_data (Any, optional): User data for callbacks
		use_internal_label (bool, optional): Use generated internal label instead of user specified (appends ### uuid).
		tag (Union[int, str], optional): Unique id used to programmatically refer to the item.If label is unused this will be the label.
		width (int, optional): Width of the item.
		height (int, optional): Height of the item.
		indent (int, optional): Offsets the widget to the right the specified number multiplied by the indent style.
		parent (Union[int, str], optional): Parent to add this item to. (runtime adding)
		before (Union[int, str], optional): This item will be displayed before the specified item in the parent.
		payload_type (str, optional): Sender string type must be the same as the target for the target to run the payload_callback.
		callback (Callable, optional): Registers a callback.
		drag_callback (Callable, optional): Registers a drag callback for drag and drop.
		drop_callback (Callable, optional): Registers a drop callback for drag and drop.
		show (bool, optional): Attempt to render widget.
		enabled (bool, optional): Turns off functionality of widget and applies the disabled theme.
		pos (Union[List[int], Tuple[int, ...]], optional): Places the item relative to window coordinates, [0,0] is top left.
		filter_key (str, optional): Used by filter widget.
		tracked (bool, optional): Scroll tracking
		track_offset (float, optional): 0.0f:top, 0.5f:center, 1.0f:bottom
		no_alpha (bool, optional): Removes the displayed slider that can change alpha channel.
		no_border (bool, optional): Disable border around the image.
		no_drag_drop (bool, optional): Disable ability to drag and drop small preview (color square) to apply colors to other items.
		id (Union[int, str], optional): (deprecated)
	Returns:
		Union[int, str]
	"""

	return internal_dpg.add_color_button(default_value, **kwargs)

def add_color_edit(default_value=(0, 0, 0, 255), **kwargs):
	"""	 Adds an RGBA color editor. Left clicking the small color preview will provide a color picker. Click and draging the small color preview will copy the color to be applied on any other color widget.

	Args:
		default_value (Union[List[int], Tuple[int, ...]], optional): 
		label (str, optional): Overrides 'name' as label.
		user_data (Any, optional): User data for callbacks
		use_internal_label (bool, optional): Use generated internal label instead of user specified (appends ### uuid).
		tag (Union[int, str], optional): Unique id used to programmatically refer to the item.If label is unused this will be the label.
		width (int, optional): Width of the item.
		height (int, optional): Height of the item.
		indent (int, optional): Offsets the widget to the right the specified number multiplied by the indent style.
		parent (Union[int, str], optional): Parent to add this item to. (runtime adding)
		before (Union[int, str], optional): This item will be displayed before the specified item in the parent.
		source (Union[int, str], optional): Overrides 'id' as value storage key.
		payload_type (str, optional): Sender string type must be the same as the target for the target to run the payload_callback.
		callback (Callable, optional): Registers a callback.
		drag_callback (Callable, optional): Registers a drag callback for drag and drop.
		drop_callback (Callable, optional): Registers a drop callback for drag and drop.
		show (bool, optional): Attempt to render widget.
		enabled (bool, optional): Turns off functionality of widget and applies the disabled theme.
		pos (Union[List[int], Tuple[int, ...]], optional): Places the item relative to window coordinates, [0,0] is top left.
		filter_key (str, optional): Used by filter widget.
		tracked (bool, optional): Scroll tracking
		track_offset (float, optional): 0.0f:top, 0.5f:center, 1.0f:bottom
		no_alpha (bool, optional): Removes the displayed slider that can change alpha channel.
		no_picker (bool, optional): Disable picker popup when color square is clicked.
		no_options (bool, optional): Disable toggling options menu when right-clicking on inputs/small preview.
		no_small_preview (bool, optional): Disable colored square preview next to the inputs. (e.g. to show only the inputs). This only displays if the side preview is not shown.
		no_inputs (bool, optional): Disable inputs sliders/text widgets. (e.g. to show only the small preview colored square)
		no_tooltip (bool, optional): Disable tooltip when hovering the preview.
		no_label (bool, optional): Disable display of inline text label.
		no_drag_drop (bool, optional): Disable ability to drag and drop small preview (color square) to apply colors to other items.
		alpha_bar (bool, optional): Show vertical alpha bar/gradient in picker.
		alpha_preview (int, optional): mvColorEdit_AlphaPreviewNone, mvColorEdit_AlphaPreview, or mvColorEdit_AlphaPreviewHalf
		display_mode (int, optional): mvColorEdit_rgb, mvColorEdit_hsv, or mvColorEdit_hex
		display_type (int, optional): mvColorEdit_uint8 or mvColorEdit_float
		input_mode (int, optional): mvColorEdit_input_rgb or mvColorEdit_input_hsv
		id (Union[int, str], optional): (deprecated)
	Returns:
		Union[int, str]
	"""

	return internal_dpg.add_color_edit(default_value, **kwargs)

def add_color_picker(default_value=(0, 0, 0, 255), **kwargs):
	"""	 Adds an RGB color picker. Right click the color picker for options. Click and drag the color preview to copy the color and drop on any other color widget to apply. Right Click allows the style of the color picker to be changed.

	Args:
		default_value (Union[List[int], Tuple[int, ...]], optional): 
		label (str, optional): Overrides 'name' as label.
		user_data (Any, optional): User data for callbacks
		use_internal_label (bool, optional): Use generated internal label instead of user specified (appends ### uuid).
		tag (Union[int, str], optional): Unique id used to programmatically refer to the item.If label is unused this will be the label.
		width (int, optional): Width of the item.
		height (int, optional): Height of the item.
		indent (int, optional): Offsets the widget to the right the specified number multiplied by the indent style.
		parent (Union[int, str], optional): Parent to add this item to. (runtime adding)
		before (Union[int, str], optional): This item will be displayed before the specified item in the parent.
		source (Union[int, str], optional): Overrides 'id' as value storage key.
		payload_type (str, optional): Sender string type must be the same as the target for the target to run the payload_callback.
		callback (Callable, optional): Registers a callback.
		drag_callback (Callable, optional): Registers a drag callback for drag and drop.
		drop_callback (Callable, optional): Registers a drop callback for drag and drop.
		show (bool, optional): Attempt to render widget.
		enabled (bool, optional): Turns off functionality of widget and applies the disabled theme.
		pos (Union[List[int], Tuple[int, ...]], optional): Places the item relative to window coordinates, [0,0] is top left.
		filter_key (str, optional): Used by filter widget.
		tracked (bool, optional): Scroll tracking
		track_offset (float, optional): 0.0f:top, 0.5f:center, 1.0f:bottom
		no_alpha (bool, optional): Removes the displayed slider that can change alpha channel.
		no_side_preview (bool, optional): Disable bigger color preview on right side of the picker, use small colored square preview instead , unless small preview is also hidden.
		no_small_preview (bool, optional): Disable colored square preview next to the inputs. (e.g. to show only the inputs). This only displays if the side preview is not shown.
		no_inputs (bool, optional): Disable inputs sliders/text widgets. (e.g. to show only the small preview colored square)
		no_tooltip (bool, optional): Disable tooltip when hovering the preview.
		no_label (bool, optional): Disable display of inline text label.
		alpha_bar (bool, optional): Show vertical alpha bar/gradient in picker.
		display_rgb (bool, optional): Override _display_ type among RGB/HSV/Hex.
		display_hsv (bool, optional): Override _display_ type among RGB/HSV/Hex.
		display_hex (bool, optional): Override _display_ type among RGB/HSV/Hex.
		picker_mode (int, optional): mvColorPicker_bar or mvColorPicker_wheel
		alpha_preview (int, optional): mvColorEdit_AlphaPreviewNone, mvColorEdit_AlphaPreview, or mvColorEdit_AlphaPreviewHalf
		display_type (int, optional): mvColorEdit_uint8 or mvColorEdit_float
		input_mode (int, optional): mvColorEdit_input_rgb or mvColorEdit_input_hsv
		id (Union[int, str], optional): (deprecated)
	Returns:
		Union[int, str]
	"""

	return internal_dpg.add_color_picker(default_value, **kwargs)

def add_color_value(**kwargs):
	"""	 Adds a color value.

	Args:
		label (str, optional): Overrides 'name' as label.
		user_data (Any, optional): User data for callbacks
		use_internal_label (bool, optional): Use generated internal label instead of user specified (appends ### uuid).
		tag (Union[int, str], optional): Unique id used to programmatically refer to the item.If label is unused this will be the label.
		source (Union[int, str], optional): Overrides 'id' as value storage key.
		default_value (Union[List[float], Tuple[float, ...]], optional): 
		parent (Union[int, str], optional): Parent to add this item to. (runtime adding)
		id (Union[int, str], optional): (deprecated)
	Returns:
		Union[int, str]
	"""

	return internal_dpg.add_color_value(**kwargs)

def add_colormap(colors, qualitative, **kwargs):
	"""	 Adds a legend that pairs colors with normalized value 0.0->1.0. Each color will be  This is typically used with a heat series. (ex. [[0, 0, 0, 255], [255, 255, 255, 255]] will be mapped to a soft transition from 0.0-1.0)

	Args:
		colors (Any): colors that will be mapped to the normalized value 0.0->1.0
		qualitative (bool): Qualitative will create hard transitions for color boundries across the value range when enabled.
		label (str, optional): Overrides 'name' as label.
		user_data (Any, optional): User data for callbacks
		use_internal_label (bool, optional): Use generated internal label instead of user specified (appends ### uuid).
		tag (Union[int, str], optional): Unique id used to programmatically refer to the item.If label is unused this will be the label.
		show (bool, optional): Attempt to render widget.
		parent (Union[int, str], optional): Parent to add this item to. (runtime adding)
		id (Union[int, str], optional): (deprecated)
	Returns:
		Union[int, str]
	"""

	return internal_dpg.add_colormap(colors, qualitative, **kwargs)

def add_colormap_button(default_value=(0, 0, 0, 255), **kwargs):
	"""	 Adds a button that a color map can be bound to.

	Args:
		default_value (Union[List[int], Tuple[int, ...]], optional): 
		label (str, optional): Overrides 'name' as label.
		user_data (Any, optional): User data for callbacks
		use_internal_label (bool, optional): Use generated internal label instead of user specified (appends ### uuid).
		tag (Union[int, str], optional): Unique id used to programmatically refer to the item.If label is unused this will be the label.
		width (int, optional): Width of the item.
		height (int, optional): Height of the item.
		indent (int, optional): Offsets the widget to the right the specified number multiplied by the indent style.
		parent (Union[int, str], optional): Parent to add this item to. (runtime adding)
		before (Union[int, str], optional): This item will be displayed before the specified item in the parent.
		payload_type (str, optional): Sender string type must be the same as the target for the target to run the payload_callback.
		callback (Callable, optional): Registers a callback.
		drag_callback (Callable, optional): Registers a drag callback for drag and drop.
		drop_callback (Callable, optional): Registers a drop callback for drag and drop.
		show (bool, optional): Attempt to render widget.
		enabled (bool, optional): Turns off functionality of widget and applies the disabled theme.
		pos (Union[List[int], Tuple[int, ...]], optional): Places the item relative to window coordinates, [0,0] is top left.
		filter_key (str, optional): Used by filter widget.
		tracked (bool, optional): Scroll tracking
		track_offset (float, optional): 0.0f:top, 0.5f:center, 1.0f:bottom
		id (Union[int, str], optional): (deprecated)
	Returns:
		Union[int, str]
	"""

	return internal_dpg.add_colormap_button(default_value, **kwargs)

def add_colormap_registry(**kwargs):
	"""	 Adds a colormap registry.

	Args:
		label (str, optional): Overrides 'name' as label.
		user_data (Any, optional): User data for callbacks
		use_internal_label (bool, optional): Use generated internal label instead of user specified (appends ### uuid).
		tag (Union[int, str], optional): Unique id used to programmatically refer to the item.If label is unused this will be the label.
		show (bool, optional): Attempt to render widget.
		id (Union[int, str], optional): (deprecated)
	Returns:
		Union[int, str]
	"""

	return internal_dpg.add_colormap_registry(**kwargs)

def add_colormap_scale(**kwargs):
	"""	 Adds a legend that pairs values with colors. This is typically used with a heat series. 

	Args:
		label (str, optional): Overrides 'name' as label.
		user_data (Any, optional): User data for callbacks
		use_internal_label (bool, optional): Use generated internal label instead of user specified (appends ### uuid).
		tag (Union[int, str], optional): Unique id used to programmatically refer to the item.If label is unused this will be the label.
		width (int, optional): Width of the item.
		height (int, optional): Height of the item.
		indent (int, optional): Offsets the widget to the right the specified number multiplied by the indent style.
		parent (Union[int, str], optional): Parent to add this item to. (runtime adding)
		before (Union[int, str], optional): This item will be displayed before the specified item in the parent.
		source (Union[int, str], optional): Overrides 'id' as value storage key.
		payload_type (str, optional): Sender string type must be the same as the target for the target to run the payload_callback.
		drop_callback (Callable, optional): Registers a drop callback for drag and drop.
		show (bool, optional): Attempt to render widget.
		pos (Union[List[int], Tuple[int, ...]], optional): Places the item relative to window coordinates, [0,0] is top left.
		colormap (Union[int, str], optional): mvPlotColormap_* constants or mvColorMap uuid from a color map registry
		min_scale (float, optional): Sets the min number of the color scale. Typically is the same as the min scale from the heat series.
		max_scale (float, optional): Sets the max number of the color scale. Typically is the same as the max scale from the heat series.
		id (Union[int, str], optional): (deprecated)
		drag_callback (Callable, optional): (deprecated)
	Returns:
		Union[int, str]
	"""

	return internal_dpg.add_colormap_scale(**kwargs)

def add_colormap_slider(**kwargs):
	"""	 Adds a color slider that a color map can be bound to.

	Args:
		label (str, optional): Overrides 'name' as label.
		user_data (Any, optional): User data for callbacks
		use_internal_label (bool, optional): Use generated internal label instead of user specified (appends ### uuid).
		tag (Union[int, str], optional): Unique id used to programmatically refer to the item.If label is unused this will be the label.
		width (int, optional): Width of the item.
		height (int, optional): Height of the item.
		indent (int, optional): Offsets the widget to the right the specified number multiplied by the indent style.
		parent (Union[int, str], optional): Parent to add this item to. (runtime adding)
		before (Union[int, str], optional): This item will be displayed before the specified item in the parent.
		payload_type (str, optional): Sender string type must be the same as the target for the target to run the payload_callback.
		callback (Callable, optional): Registers a callback.
		drop_callback (Callable, optional): Registers a drop callback for drag and drop.
		show (bool, optional): Attempt to render widget.
		pos (Union[List[int], Tuple[int, ...]], optional): Places the item relative to window coordinates, [0,0] is top left.
		filter_key (str, optional): Used by filter widget.
		tracked (bool, optional): Scroll tracking
		track_offset (float, optional): 0.0f:top, 0.5f:center, 1.0f:bottom
		default_value (float, optional): 
		id (Union[int, str], optional): (deprecated)
		drag_callback (Callable, optional): (deprecated)
	Returns:
		Union[int, str]
	"""

	return internal_dpg.add_colormap_slider(**kwargs)

def add_combo(items=(), **kwargs):
	"""	 Adds a combo dropdown that allows a user to select a single option from a drop down window. All items will be shown as selectables on the dropdown.

	Args:
		items (Union[List[str], Tuple[str, ...]], optional): A tuple of items to be shown in the drop down window. Can consist of any combination of types but will convert all items to strings to be shown.
		label (str, optional): Overrides 'name' as label.
		user_data (Any, optional): User data for callbacks
		use_internal_label (bool, optional): Use generated internal label instead of user specified (appends ### uuid).
		tag (Union[int, str], optional): Unique id used to programmatically refer to the item.If label is unused this will be the label.
		width (int, optional): Width of the item.
		indent (int, optional): Offsets the widget to the right the specified number multiplied by the indent style.
		parent (Union[int, str], optional): Parent to add this item to. (runtime adding)
		before (Union[int, str], optional): This item will be displayed before the specified item in the parent.
		source (Union[int, str], optional): Overrides 'id' as value storage key.
		payload_type (str, optional): Sender string type must be the same as the target for the target to run the payload_callback.
		callback (Callable, optional): Registers a callback.
		drag_callback (Callable, optional): Registers a drag callback for drag and drop.
		drop_callback (Callable, optional): Registers a drop callback for drag and drop.
		show (bool, optional): Attempt to render widget.
		enabled (bool, optional): Turns off functionality of widget and applies the disabled theme.
		pos (Union[List[int], Tuple[int, ...]], optional): Places the item relative to window coordinates, [0,0] is top left.
		filter_key (str, optional): Used by filter widget.
		tracked (bool, optional): Scroll tracking
		track_offset (float, optional): 0.0f:top, 0.5f:center, 1.0f:bottom
		default_value (str, optional): Sets a selected item from the drop down by specifying the string value.
		popup_align_left (bool, optional): Align the contents on the popup toward the left.
		no_arrow_button (bool, optional): Display the preview box without the square arrow button indicating dropdown activity.
		no_preview (bool, optional): Display only the square arrow button and not the selected value.
		height_mode (int, optional): Controlls the number of items shown in the dropdown by the constants mvComboHeight_Small, mvComboHeight_Regular, mvComboHeight_Large, mvComboHeight_Largest
		id (Union[int, str], optional): (deprecated)
	Returns:
		Union[int, str]
	"""

	return internal_dpg.add_combo(items, **kwargs)

def add_date_picker(**kwargs):
	"""	 Adds a data picker.

	Args:
		label (str, optional): Overrides 'name' as label.
		user_data (Any, optional): User data for callbacks
		use_internal_label (bool, optional): Use generated internal label instead of user specified (appends ### uuid).
		tag (Union[int, str], optional): Unique id used to programmatically refer to the item.If label is unused this will be the label.
		indent (int, optional): Offsets the widget to the right the specified number multiplied by the indent style.
		parent (Union[int, str], optional): Parent to add this item to. (runtime adding)
		before (Union[int, str], optional): This item will be displayed before the specified item in the parent.
		payload_type (str, optional): Sender string type must be the same as the target for the target to run the payload_callback.
		callback (Callable, optional): Registers a callback.
		drag_callback (Callable, optional): Registers a drag callback for drag and drop.
		drop_callback (Callable, optional): Registers a drop callback for drag and drop.
		show (bool, optional): Attempt to render widget.
		pos (Union[List[int], Tuple[int, ...]], optional): Places the item relative to window coordinates, [0,0] is top left.
		filter_key (str, optional): Used by filter widget.
		tracked (bool, optional): Scroll tracking
		track_offset (float, optional): 0.0f:top, 0.5f:center, 1.0f:bottom
		default_value (dict, optional): 
		level (int, optional): Use avaliable constants. mvDatePickerLevel_Day, mvDatePickerLevel_Month, mvDatePickerLevel_Year
		id (Union[int, str], optional): (deprecated)
	Returns:
		Union[int, str]
	"""

	return internal_dpg.add_date_picker(**kwargs)

def add_double4_value(**kwargs):
	"""	 Adds a double value.

	Args:
		label (str, optional): Overrides 'name' as label.
		user_data (Any, optional): User data for callbacks
		use_internal_label (bool, optional): Use generated internal label instead of user specified (appends ### uuid).
		tag (Union[int, str], optional): Unique id used to programmatically refer to the item.If label is unused this will be the label.
		source (Union[int, str], optional): Overrides 'id' as value storage key.
		default_value (Any, optional): 
		parent (Union[int, str], optional): Parent to add this item to. (runtime adding)
		id (Union[int, str], optional): (deprecated)
	Returns:
		Union[int, str]
	"""

	return internal_dpg.add_double4_value(**kwargs)

def add_double_value(**kwargs):
	"""	 Adds a double value.

	Args:
		label (str, optional): Overrides 'name' as label.
		user_data (Any, optional): User data for callbacks
		use_internal_label (bool, optional): Use generated internal label instead of user specified (appends ### uuid).
		tag (Union[int, str], optional): Unique id used to programmatically refer to the item.If label is unused this will be the label.
		source (Union[int, str], optional): Overrides 'id' as value storage key.
		default_value (float, optional): 
		parent (Union[int, str], optional): Parent to add this item to. (runtime adding)
		id (Union[int, str], optional): (deprecated)
	Returns:
		Union[int, str]
	"""

	return internal_dpg.add_double_value(**kwargs)

def add_drag_float(**kwargs):
	"""	 Adds drag for a single float value. Directly entry can be done with double click or CTRL+Click. Min and Max alone are a soft limit for the drag. Use clamped keyword to also apply limits to the direct entry modes.

	Args:
		label (str, optional): Overrides 'name' as label.
		user_data (Any, optional): User data for callbacks
		use_internal_label (bool, optional): Use generated internal label instead of user specified (appends ### uuid).
		tag (Union[int, str], optional): Unique id used to programmatically refer to the item.If label is unused this will be the label.
		width (int, optional): Width of the item.
		indent (int, optional): Offsets the widget to the right the specified number multiplied by the indent style.
		parent (Union[int, str], optional): Parent to add this item to. (runtime adding)
		before (Union[int, str], optional): This item will be displayed before the specified item in the parent.
		source (Union[int, str], optional): Overrides 'id' as value storage key.
		payload_type (str, optional): Sender string type must be the same as the target for the target to run the payload_callback.
		callback (Callable, optional): Registers a callback.
		drag_callback (Callable, optional): Registers a drag callback for drag and drop.
		drop_callback (Callable, optional): Registers a drop callback for drag and drop.
		show (bool, optional): Attempt to render widget.
		enabled (bool, optional): Turns off functionality of widget and applies the disabled theme.
		pos (Union[List[int], Tuple[int, ...]], optional): Places the item relative to window coordinates, [0,0] is top left.
		filter_key (str, optional): Used by filter widget.
		tracked (bool, optional): Scroll tracking
		track_offset (float, optional): 0.0f:top, 0.5f:center, 1.0f:bottom
		default_value (float, optional): 
		format (str, optional): Determines the format the float will be displayed as use python string formatting.
		speed (float, optional): Sets the sensitivity the float will be modified while dragging.
		min_value (float, optional): Applies a limit only to draging entry only.
		max_value (float, optional): Applies a limit only to draging entry only.
		no_input (bool, optional): Disable direct entry methods or Enter key allowing to input text directly into the widget.
		clamped (bool, optional): Applies the min and max limits to direct entry methods also such as double click and CTRL+Click.
		id (Union[int, str], optional): (deprecated)
	Returns:
		Union[int, str]
	"""

	return internal_dpg.add_drag_float(**kwargs)

def add_drag_floatx(**kwargs):
	"""	 Adds drag input for a set of float values up to 4. Directly entry can be done with double click or CTRL+Click. Min and Max alone are a soft limit for the drag. Use clamped keyword to also apply limits to the direct entry modes.

	Args:
		label (str, optional): Overrides 'name' as label.
		user_data (Any, optional): User data for callbacks
		use_internal_label (bool, optional): Use generated internal label instead of user specified (appends ### uuid).
		tag (Union[int, str], optional): Unique id used to programmatically refer to the item.If label is unused this will be the label.
		width (int, optional): Width of the item.
		indent (int, optional): Offsets the widget to the right the specified number multiplied by the indent style.
		parent (Union[int, str], optional): Parent to add this item to. (runtime adding)
		before (Union[int, str], optional): This item will be displayed before the specified item in the parent.
		source (Union[int, str], optional): Overrides 'id' as value storage key.
		payload_type (str, optional): Sender string type must be the same as the target for the target to run the payload_callback.
		callback (Callable, optional): Registers a callback.
		drag_callback (Callable, optional): Registers a drag callback for drag and drop.
		drop_callback (Callable, optional): Registers a drop callback for drag and drop.
		show (bool, optional): Attempt to render widget.
		enabled (bool, optional): Turns off functionality of widget and applies the disabled theme.
		pos (Union[List[int], Tuple[int, ...]], optional): Places the item relative to window coordinates, [0,0] is top left.
		filter_key (str, optional): Used by filter widget.
		tracked (bool, optional): Scroll tracking
		track_offset (float, optional): 0.0f:top, 0.5f:center, 1.0f:bottom
		default_value (Union[List[float], Tuple[float, ...]], optional): 
		size (int, optional): Number of floats to be displayed.
		format (str, optional): Determines the format the float will be displayed as use python string formatting.
		speed (float, optional): Sets the sensitivity the float will be modified while dragging.
		min_value (float, optional): Applies a limit only to draging entry only.
		max_value (float, optional): Applies a limit only to draging entry only.
		no_input (bool, optional): Disable direct entry methods or Enter key allowing to input text directly into the widget.
		clamped (bool, optional): Applies the min and max limits to direct entry methods also such as double click and CTRL+Click.
		id (Union[int, str], optional): (deprecated)
	Returns:
		Union[int, str]
	"""

	return internal_dpg.add_drag_floatx(**kwargs)

def add_drag_int(**kwargs):
	"""	 Adds drag for a single int value. Directly entry can be done with double click or CTRL+Click. Min and Max alone are a soft limit for the drag. Use clamped keyword to also apply limits to the direct entry modes.

	Args:
		label (str, optional): Overrides 'name' as label.
		user_data (Any, optional): User data for callbacks
		use_internal_label (bool, optional): Use generated internal label instead of user specified (appends ### uuid).
		tag (Union[int, str], optional): Unique id used to programmatically refer to the item.If label is unused this will be the label.
		width (int, optional): Width of the item.
		indent (int, optional): Offsets the widget to the right the specified number multiplied by the indent style.
		parent (Union[int, str], optional): Parent to add this item to. (runtime adding)
		before (Union[int, str], optional): This item will be displayed before the specified item in the parent.
		source (Union[int, str], optional): Overrides 'id' as value storage key.
		payload_type (str, optional): Sender string type must be the same as the target for the target to run the payload_callback.
		callback (Callable, optional): Registers a callback.
		drag_callback (Callable, optional): Registers a drag callback for drag and drop.
		drop_callback (Callable, optional): Registers a drop callback for drag and drop.
		show (bool, optional): Attempt to render widget.
		enabled (bool, optional): Turns off functionality of widget and applies the disabled theme.
		pos (Union[List[int], Tuple[int, ...]], optional): Places the item relative to window coordinates, [0,0] is top left.
		filter_key (str, optional): Used by filter widget.
		tracked (bool, optional): Scroll tracking
		track_offset (float, optional): 0.0f:top, 0.5f:center, 1.0f:bottom
		default_value (int, optional): 
		format (str, optional): Determines the format the float will be displayed as use python string formatting.
		speed (float, optional): Sets the sensitivity the float will be modified while dragging.
		min_value (int, optional): Applies a limit only to draging entry only.
		max_value (int, optional): Applies a limit only to draging entry only.
		no_input (bool, optional): Disable direct entry methods or Enter key allowing to input text directly into the widget.
		clamped (bool, optional): Applies the min and max limits to direct entry methods also such as double click and CTRL+Click.
		id (Union[int, str], optional): (deprecated)
	Returns:
		Union[int, str]
	"""

	return internal_dpg.add_drag_int(**kwargs)

def add_drag_intx(**kwargs):
	"""	 Adds drag input for a set of int values up to 4. Directly entry can be done with double click or CTRL+Click. Min and Max alone are a soft limit for the drag. Use clamped keyword to also apply limits to the direct entry modes.

	Args:
		label (str, optional): Overrides 'name' as label.
		user_data (Any, optional): User data for callbacks
		use_internal_label (bool, optional): Use generated internal label instead of user specified (appends ### uuid).
		tag (Union[int, str], optional): Unique id used to programmatically refer to the item.If label is unused this will be the label.
		width (int, optional): Width of the item.
		indent (int, optional): Offsets the widget to the right the specified number multiplied by the indent style.
		parent (Union[int, str], optional): Parent to add this item to. (runtime adding)
		before (Union[int, str], optional): This item will be displayed before the specified item in the parent.
		source (Union[int, str], optional): Overrides 'id' as value storage key.
		payload_type (str, optional): Sender string type must be the same as the target for the target to run the payload_callback.
		callback (Callable, optional): Registers a callback.
		drag_callback (Callable, optional): Registers a drag callback for drag and drop.
		drop_callback (Callable, optional): Registers a drop callback for drag and drop.
		show (bool, optional): Attempt to render widget.
		enabled (bool, optional): Turns off functionality of widget and applies the disabled theme.
		pos (Union[List[int], Tuple[int, ...]], optional): Places the item relative to window coordinates, [0,0] is top left.
		filter_key (str, optional): Used by filter widget.
		tracked (bool, optional): Scroll tracking
		track_offset (float, optional): 0.0f:top, 0.5f:center, 1.0f:bottom
		default_value (Union[List[int], Tuple[int, ...]], optional): 
		size (int, optional): Number of ints to be displayed.
		format (str, optional): Determines the format the int will be displayed as use python string formatting.
		speed (float, optional): Sets the sensitivity the float will be modified while dragging.
		min_value (int, optional): Applies a limit only to draging entry only.
		max_value (int, optional): Applies a limit only to draging entry only.
		no_input (bool, optional): Disable direct entry methods or Enter key allowing to input text directly into the widget.
		clamped (bool, optional): Applies the min and max limits to direct entry methods also such as double click and CTRL+Click.
		id (Union[int, str], optional): (deprecated)
	Returns:
		Union[int, str]
	"""

	return internal_dpg.add_drag_intx(**kwargs)

def add_drag_line(**kwargs):
	"""	 Adds a drag line to a plot.

	Args:
		label (str, optional): Overrides 'name' as label.
		user_data (Any, optional): User data for callbacks
		use_internal_label (bool, optional): Use generated internal label instead of user specified (appends ### uuid).
		tag (Union[int, str], optional): Unique id used to programmatically refer to the item.If label is unused this will be the label.
		parent (Union[int, str], optional): Parent to add this item to. (runtime adding)
		before (Union[int, str], optional): This item will be displayed before the specified item in the parent.
		source (Union[int, str], optional): Overrides 'id' as value storage key.
		callback (Callable, optional): Registers a callback.
		show (bool, optional): Attempt to render widget.
		default_value (Any, optional): 
		color (Union[List[int], Tuple[int, ...]], optional): 
		thickness (float, optional): 
		show_label (bool, optional): 
		vertical (bool, optional): 
		id (Union[int, str], optional): (deprecated)
	Returns:
		Union[int, str]
	"""

	return internal_dpg.add_drag_line(**kwargs)

def add_drag_payload(**kwargs):
	"""	 User data payload for drag and drop operations.

	Args:
		label (str, optional): Overrides 'name' as label.
		user_data (Any, optional): User data for callbacks
		use_internal_label (bool, optional): Use generated internal label instead of user specified (appends ### uuid).
		tag (Union[int, str], optional): Unique id used to programmatically refer to the item.If label is unused this will be the label.
		parent (Union[int, str], optional): Parent to add this item to. (runtime adding)
		show (bool, optional): Attempt to render widget.
		drag_data (Any, optional): Drag data
		drop_data (Any, optional): Drop data
		payload_type (str, optional): 
		id (Union[int, str], optional): (deprecated)
	Returns:
		Union[int, str]
	"""

	return internal_dpg.add_drag_payload(**kwargs)

def add_drag_point(**kwargs):
	"""	 Adds a drag point to a plot.

	Args:
		label (str, optional): Overrides 'name' as label.
		user_data (Any, optional): User data for callbacks
		use_internal_label (bool, optional): Use generated internal label instead of user specified (appends ### uuid).
		tag (Union[int, str], optional): Unique id used to programmatically refer to the item.If label is unused this will be the label.
		parent (Union[int, str], optional): Parent to add this item to. (runtime adding)
		before (Union[int, str], optional): This item will be displayed before the specified item in the parent.
		source (Union[int, str], optional): Overrides 'id' as value storage key.
		callback (Callable, optional): Registers a callback.
		show (bool, optional): Attempt to render widget.
		default_value (Any, optional): 
		color (Union[List[int], Tuple[int, ...]], optional): 
		thickness (float, optional): 
		show_label (bool, optional): 
		id (Union[int, str], optional): (deprecated)
	Returns:
		Union[int, str]
	"""

	return internal_dpg.add_drag_point(**kwargs)

def add_draw_layer(**kwargs):
	"""	 New in 1.1. Creates a layer useful for grouping drawlist items.

	Args:
		label (str, optional): Overrides 'name' as label.
		user_data (Any, optional): User data for callbacks
		use_internal_label (bool, optional): Use generated internal label instead of user specified (appends ### uuid).
		tag (Union[int, str], optional): Unique id used to programmatically refer to the item.If label is unused this will be the label.
		parent (Union[int, str], optional): Parent to add this item to. (runtime adding)
		before (Union[int, str], optional): This item will be displayed before the specified item in the parent.
		show (bool, optional): Attempt to render widget.
		perspective_divide (bool, optional): New in 1.1. apply perspective divide
		depth_clipping (bool, optional): New in 1.1. apply depth clipping
		cull_mode (int, optional): New in 1.1. culling mode, mvCullMode_* constants. Only works with triangles currently.
		id (Union[int, str], optional): (deprecated)
	Returns:
		Union[int, str]
	"""

	return internal_dpg.add_draw_layer(**kwargs)

def add_draw_node(**kwargs):
	"""	 New in 1.1. Creates a drawing node to associate a transformation matrix. Child node matricies will concatenate.

	Args:
		label (str, optional): Overrides 'name' as label.
		user_data (Any, optional): User data for callbacks
		use_internal_label (bool, optional): Use generated internal label instead of user specified (appends ### uuid).
		tag (Union[int, str], optional): Unique id used to programmatically refer to the item.If label is unused this will be the label.
		parent (Union[int, str], optional): Parent to add this item to. (runtime adding)
		before (Union[int, str], optional): This item will be displayed before the specified item in the parent.
		show (bool, optional): Attempt to render widget.
		id (Union[int, str], optional): (deprecated)
	Returns:
		Union[int, str]
	"""

	return internal_dpg.add_draw_node(**kwargs)

def add_drawlist(width, height, **kwargs):
	"""	 Adds a drawing canvas.

	Args:
		width (int): 
		height (int): 
		label (str, optional): Overrides 'name' as label.
		user_data (Any, optional): User data for callbacks
		use_internal_label (bool, optional): Use generated internal label instead of user specified (appends ### uuid).
		tag (Union[int, str], optional): Unique id used to programmatically refer to the item.If label is unused this will be the label.
		parent (Union[int, str], optional): Parent to add this item to. (runtime adding)
		before (Union[int, str], optional): This item will be displayed before the specified item in the parent.
		callback (Callable, optional): Registers a callback.
		show (bool, optional): Attempt to render widget.
		pos (Union[List[int], Tuple[int, ...]], optional): Places the item relative to window coordinates, [0,0] is top left.
		filter_key (str, optional): Used by filter widget.
		delay_search (bool, optional): Delays searching container for specified items until the end of the app. Possible optimization when a container has many children that are not accessed often.
		tracked (bool, optional): Scroll tracking
		track_offset (float, optional): 0.0f:top, 0.5f:center, 1.0f:bottom
		id (Union[int, str], optional): (deprecated)
	Returns:
		Union[int, str]
	"""

	return internal_dpg.add_drawlist(width, height, **kwargs)

def add_dynamic_texture(width, height, default_value, **kwargs):
	"""	 Adds a dynamic texture.

	Args:
		width (int): 
		height (int): 
		default_value (Union[List[float], Tuple[float, ...]]): 
		label (str, optional): Overrides 'name' as label.
		user_data (Any, optional): User data for callbacks
		use_internal_label (bool, optional): Use generated internal label instead of user specified (appends ### uuid).
		tag (Union[int, str], optional): Unique id used to programmatically refer to the item.If label is unused this will be the label.
		parent (Union[int, str], optional): Parent to add this item to. (runtime adding)
		id (Union[int, str], optional): (deprecated)
	Returns:
		Union[int, str]
	"""

	return internal_dpg.add_dynamic_texture(width, height, default_value, **kwargs)

def add_error_series(x, y, negative, positive, **kwargs):
	"""	 Adds an error series to a plot.

	Args:
		x (Any): 
		y (Any): 
		negative (Any): 
		positive (Any): 
		label (str, optional): Overrides 'name' as label.
		user_data (Any, optional): User data for callbacks
		use_internal_label (bool, optional): Use generated internal label instead of user specified (appends ### uuid).
		tag (Union[int, str], optional): Unique id used to programmatically refer to the item.If label is unused this will be the label.
		parent (Union[int, str], optional): Parent to add this item to. (runtime adding)
		before (Union[int, str], optional): This item will be displayed before the specified item in the parent.
		source (Union[int, str], optional): Overrides 'id' as value storage key.
		show (bool, optional): Attempt to render widget.
		contribute_to_bounds (bool, optional): 
		horizontal (bool, optional): 
		id (Union[int, str], optional): (deprecated)
	Returns:
		Union[int, str]
	"""

	return internal_dpg.add_error_series(x, y, negative, positive, **kwargs)

def add_file_dialog(**kwargs):
	"""	 Displays a file or directory selector depending on keywords. Displays a file dialog by default. Callback will be ran when the file or directory picker is closed. The app_data arguemnt will be populated with information related to the file and directory as a dictionary.

	Args:
		label (str, optional): Overrides 'name' as label.
		user_data (Any, optional): User data for callbacks
		use_internal_label (bool, optional): Use generated internal label instead of user specified (appends ### uuid).
		tag (Union[int, str], optional): Unique id used to programmatically refer to the item.If label is unused this will be the label.
		width (int, optional): Width of the item.
		height (int, optional): Height of the item.
		callback (Callable, optional): Registers a callback.
		show (bool, optional): Attempt to render widget.
		default_path (str, optional): Path that the file dialog will default to when opened.
		default_filename (str, optional): Default name that will show in the file name input.
		file_count (int, optional): Number of visible files in the dialog.
		modal (bool, optional): Forces user interaction with the file selector.
		directory_selector (bool, optional): Shows only directory/paths as options. Allows selection of directory/paths only.
		min_size (Union[List[int], Tuple[int, ...]], optional): Minimum window size.
		max_size (Union[List[int], Tuple[int, ...]], optional): Maximum window size.
		id (Union[int, str], optional): (deprecated)
	Returns:
		Union[int, str]
	"""

	return internal_dpg.add_file_dialog(**kwargs)

def add_file_extension(extension, **kwargs):
	"""	 Creates a file extension filter option in the file dialog.

	Args:
		extension (str): Extension that will show as an when the parent is a file dialog.
		label (str, optional): Overrides 'name' as label.
		user_data (Any, optional): User data for callbacks
		use_internal_label (bool, optional): Use generated internal label instead of user specified (appends ### uuid).
		tag (Union[int, str], optional): Unique id used to programmatically refer to the item.If label is unused this will be the label.
		width (int, optional): Width of the item.
		height (int, optional): Height of the item.
		parent (Union[int, str], optional): Parent to add this item to. (runtime adding)
		before (Union[int, str], optional): This item will be displayed before the specified item in the parent.
		custom_text (str, optional): Replaces the displayed text in the drop down for this extension.
		color (Union[List[int], Tuple[int, ...]], optional): Color for the text that will be shown with specified extensions.
		id (Union[int, str], optional): (deprecated)
	Returns:
		Union[int, str]
	"""

	return internal_dpg.add_file_extension(extension, **kwargs)

def add_filter_set(**kwargs):
	"""	 Helper to parse and apply text filters (e.g. aaaaa[, bbbbb][, ccccc])

	Args:
		label (str, optional): Overrides 'name' as label.
		user_data (Any, optional): User data for callbacks
		use_internal_label (bool, optional): Use generated internal label instead of user specified (appends ### uuid).
		tag (Union[int, str], optional): Unique id used to programmatically refer to the item.If label is unused this will be the label.
		width (int, optional): Width of the item.
		indent (int, optional): Offsets the widget to the right the specified number multiplied by the indent style.
		parent (Union[int, str], optional): Parent to add this item to. (runtime adding)
		before (Union[int, str], optional): This item will be displayed before the specified item in the parent.
		show (bool, optional): Attempt to render widget.
		delay_search (bool, optional): Delays searching container for specified items until the end of the app. Possible optimization when a container has many children that are not accessed often.
		id (Union[int, str], optional): (deprecated)
	Returns:
		Union[int, str]
	"""

	return internal_dpg.add_filter_set(**kwargs)

def add_float4_value(**kwargs):
	"""	 Adds a float4 value.

	Args:
		label (str, optional): Overrides 'name' as label.
		user_data (Any, optional): User data for callbacks
		use_internal_label (bool, optional): Use generated internal label instead of user specified (appends ### uuid).
		tag (Union[int, str], optional): Unique id used to programmatically refer to the item.If label is unused this will be the label.
		source (Union[int, str], optional): Overrides 'id' as value storage key.
		default_value (Union[List[float], Tuple[float, ...]], optional): 
		parent (Union[int, str], optional): Parent to add this item to. (runtime adding)
		id (Union[int, str], optional): (deprecated)
	Returns:
		Union[int, str]
	"""

	return internal_dpg.add_float4_value(**kwargs)

def add_float_value(**kwargs):
	"""	 Adds a float value.

	Args:
		label (str, optional): Overrides 'name' as label.
		user_data (Any, optional): User data for callbacks
		use_internal_label (bool, optional): Use generated internal label instead of user specified (appends ### uuid).
		tag (Union[int, str], optional): Unique id used to programmatically refer to the item.If label is unused this will be the label.
		source (Union[int, str], optional): Overrides 'id' as value storage key.
		default_value (float, optional): 
		parent (Union[int, str], optional): Parent to add this item to. (runtime adding)
		id (Union[int, str], optional): (deprecated)
	Returns:
		Union[int, str]
	"""

	return internal_dpg.add_float_value(**kwargs)

def add_float_vect_value(**kwargs):
	"""	 Adds a float vect value.

	Args:
		label (str, optional): Overrides 'name' as label.
		user_data (Any, optional): User data for callbacks
		use_internal_label (bool, optional): Use generated internal label instead of user specified (appends ### uuid).
		tag (Union[int, str], optional): Unique id used to programmatically refer to the item.If label is unused this will be the label.
		source (Union[int, str], optional): Overrides 'id' as value storage key.
		default_value (Union[List[float], Tuple[float, ...]], optional): 
		parent (Union[int, str], optional): Parent to add this item to. (runtime adding)
		id (Union[int, str], optional): (deprecated)
	Returns:
		Union[int, str]
	"""

	return internal_dpg.add_float_vect_value(**kwargs)

def add_font(file, size, **kwargs):
	"""	 Adds font to a font registry.

	Args:
		file (str): 
		size (int): 
		label (str, optional): Overrides 'name' as label.
		user_data (Any, optional): User data for callbacks
		use_internal_label (bool, optional): Use generated internal label instead of user specified (appends ### uuid).
		tag (Union[int, str], optional): Unique id used to programmatically refer to the item.If label is unused this will be the label.
		parent (Union[int, str], optional): Parent to add this item to. (runtime adding)
		id (Union[int, str], optional): (deprecated)
		default_font (bool, optional): (deprecated)
	Returns:
		Union[int, str]
	"""

	return internal_dpg.add_font(file, size, **kwargs)

def add_font_chars(chars, **kwargs):
	"""	 Adds specific font characters to a font.

	Args:
		chars (Union[List[int], Tuple[int, ...]]): 
		label (str, optional): Overrides 'name' as label.
		user_data (Any, optional): User data for callbacks
		use_internal_label (bool, optional): Use generated internal label instead of user specified (appends ### uuid).
		tag (Union[int, str], optional): Unique id used to programmatically refer to the item.If label is unused this will be the label.
		parent (Union[int, str], optional): Parent to add this item to. (runtime adding)
		id (Union[int, str], optional): (deprecated)
	Returns:
		Union[int, str]
	"""

	return internal_dpg.add_font_chars(chars, **kwargs)

def add_font_range(first_char, last_char, **kwargs):
	"""	 Adds a range of font characters to a font.

	Args:
		first_char (int): 
		last_char (int): 
		label (str, optional): Overrides 'name' as label.
		user_data (Any, optional): User data for callbacks
		use_internal_label (bool, optional): Use generated internal label instead of user specified (appends ### uuid).
		tag (Union[int, str], optional): Unique id used to programmatically refer to the item.If label is unused this will be the label.
		parent (Union[int, str], optional): Parent to add this item to. (runtime adding)
		id (Union[int, str], optional): (deprecated)
	Returns:
		Union[int, str]
	"""

	return internal_dpg.add_font_range(first_char, last_char, **kwargs)

def add_font_range_hint(hint, **kwargs):
	"""	 Adds a range of font characters (mvFontRangeHint_ constants).

	Args:
		hint (int): 
		label (str, optional): Overrides 'name' as label.
		user_data (Any, optional): User data for callbacks
		use_internal_label (bool, optional): Use generated internal label instead of user specified (appends ### uuid).
		tag (Union[int, str], optional): Unique id used to programmatically refer to the item.If label is unused this will be the label.
		parent (Union[int, str], optional): Parent to add this item to. (runtime adding)
		id (Union[int, str], optional): (deprecated)
	Returns:
		Union[int, str]
	"""

	return internal_dpg.add_font_range_hint(hint, **kwargs)

def add_font_registry(**kwargs):
	"""	 Adds a font registry.

	Args:
		label (str, optional): Overrides 'name' as label.
		user_data (Any, optional): User data for callbacks
		use_internal_label (bool, optional): Use generated internal label instead of user specified (appends ### uuid).
		tag (Union[int, str], optional): Unique id used to programmatically refer to the item.If label is unused this will be the label.
		show (bool, optional): Attempt to render widget.
		id (Union[int, str], optional): (deprecated)
	Returns:
		Union[int, str]
	"""

	return internal_dpg.add_font_registry(**kwargs)

def add_group(**kwargs):
	"""	 Creates a group that other widgets can belong to. The group allows item commands to be issued for all of its members.

	Args:
		label (str, optional): Overrides 'name' as label.
		user_data (Any, optional): User data for callbacks
		use_internal_label (bool, optional): Use generated internal label instead of user specified (appends ### uuid).
		tag (Union[int, str], optional): Unique id used to programmatically refer to the item.If label is unused this will be the label.
		width (int, optional): Width of the item.
		indent (int, optional): Offsets the widget to the right the specified number multiplied by the indent style.
		parent (Union[int, str], optional): Parent to add this item to. (runtime adding)
		before (Union[int, str], optional): This item will be displayed before the specified item in the parent.
		payload_type (str, optional): Sender string type must be the same as the target for the target to run the payload_callback.
		drag_callback (Callable, optional): Registers a drag callback for drag and drop.
		drop_callback (Callable, optional): Registers a drop callback for drag and drop.
		show (bool, optional): Attempt to render widget.
		pos (Union[List[int], Tuple[int, ...]], optional): Places the item relative to window coordinates, [0,0] is top left.
		filter_key (str, optional): Used by filter widget.
		delay_search (bool, optional): Delays searching container for specified items until the end of the app. Possible optimization when a container has many children that are not accessed often.
		tracked (bool, optional): Scroll tracking
		track_offset (float, optional): 0.0f:top, 0.5f:center, 1.0f:bottom
		horizontal (bool, optional): Forces child widgets to be added in a horizontal layout.
		horizontal_spacing (float, optional): Spacing for the horizontal layout.
		xoffset (float, optional): Offset from containing window x item location within group.
		id (Union[int, str], optional): (deprecated)
	Returns:
		Union[int, str]
	"""

	return internal_dpg.add_group(**kwargs)

def add_handler_registry(**kwargs):
	"""	 Adds a handler registry.

	Args:
		label (str, optional): Overrides 'name' as label.
		user_data (Any, optional): User data for callbacks
		use_internal_label (bool, optional): Use generated internal label instead of user specified (appends ### uuid).
		tag (Union[int, str], optional): Unique id used to programmatically refer to the item.If label is unused this will be the label.
		show (bool, optional): Attempt to render widget.
		id (Union[int, str], optional): (deprecated)
	Returns:
		Union[int, str]
	"""

	return internal_dpg.add_handler_registry(**kwargs)

def add_heat_series(x, rows, cols, **kwargs):
	"""	 Adds a heat series to a plot.

	Args:
		x (Any): 
		rows (int): 
		cols (int): 
		label (str, optional): Overrides 'name' as label.
		user_data (Any, optional): User data for callbacks
		use_internal_label (bool, optional): Use generated internal label instead of user specified (appends ### uuid).
		tag (Union[int, str], optional): Unique id used to programmatically refer to the item.If label is unused this will be the label.
		parent (Union[int, str], optional): Parent to add this item to. (runtime adding)
		before (Union[int, str], optional): This item will be displayed before the specified item in the parent.
		source (Union[int, str], optional): Overrides 'id' as value storage key.
		show (bool, optional): Attempt to render widget.
		scale_min (float, optional): Sets the color scale min. Typically paired with the color scale widget scale_min.
		scale_max (float, optional): Sets the color scale max. Typically paired with the color scale widget scale_max.
		bounds_min (Any, optional): 
		bounds_max (Any, optional): 
		format (str, optional): 
		contribute_to_bounds (bool, optional): 
		id (Union[int, str], optional): (deprecated)
	Returns:
		Union[int, str]
	"""

	return internal_dpg.add_heat_series(x, rows, cols, **kwargs)

def add_histogram_series(x, **kwargs):
	"""	 Adds a histogram series to a plot.

	Args:
		x (Any): 
		label (str, optional): Overrides 'name' as label.
		user_data (Any, optional): User data for callbacks
		use_internal_label (bool, optional): Use generated internal label instead of user specified (appends ### uuid).
		tag (Union[int, str], optional): Unique id used to programmatically refer to the item.If label is unused this will be the label.
		parent (Union[int, str], optional): Parent to add this item to. (runtime adding)
		before (Union[int, str], optional): This item will be displayed before the specified item in the parent.
		source (Union[int, str], optional): Overrides 'id' as value storage key.
		show (bool, optional): Attempt to render widget.
		bins (int, optional): 
		bar_scale (float, optional): 
		min_range (float, optional): 
		max_range (float, optional): 
		cumlative (bool, optional): 
		density (bool, optional): 
		outliers (bool, optional): 
		contribute_to_bounds (bool, optional): 
		id (Union[int, str], optional): (deprecated)
	Returns:
		Union[int, str]
	"""

	return internal_dpg.add_histogram_series(x, **kwargs)

def add_hline_series(x, **kwargs):
	"""	 Adds an infinite horizontal line series to a plot.

	Args:
		x (Any): 
		label (str, optional): Overrides 'name' as label.
		user_data (Any, optional): User data for callbacks
		use_internal_label (bool, optional): Use generated internal label instead of user specified (appends ### uuid).
		tag (Union[int, str], optional): Unique id used to programmatically refer to the item.If label is unused this will be the label.
		parent (Union[int, str], optional): Parent to add this item to. (runtime adding)
		before (Union[int, str], optional): This item will be displayed before the specified item in the parent.
		source (Union[int, str], optional): Overrides 'id' as value storage key.
		show (bool, optional): Attempt to render widget.
		id (Union[int, str], optional): (deprecated)
	Returns:
		Union[int, str]
	"""

	return internal_dpg.add_hline_series(x, **kwargs)

def add_image(texture_tag, **kwargs):
	"""	 Adds an image from a specified texture. uv_min and uv_max represent the normalized texture coordinates of the original image that will be shown. Using range (0.0,0.0)->(1.0,1.0) for texture coordinates will generally display the entire texture.

	Args:
		texture_tag (Union[int, str]): The texture_tag should come from a texture that was added to a texture registry.
		label (str, optional): Overrides 'name' as label.
		user_data (Any, optional): User data for callbacks
		use_internal_label (bool, optional): Use generated internal label instead of user specified (appends ### uuid).
		tag (Union[int, str], optional): Unique id used to programmatically refer to the item.If label is unused this will be the label.
		width (int, optional): Width of the item.
		height (int, optional): Height of the item.
		indent (int, optional): Offsets the widget to the right the specified number multiplied by the indent style.
		parent (Union[int, str], optional): Parent to add this item to. (runtime adding)
		before (Union[int, str], optional): This item will be displayed before the specified item in the parent.
		source (Union[int, str], optional): Overrides 'id' as value storage key.
		payload_type (str, optional): Sender string type must be the same as the target for the target to run the payload_callback.
		drag_callback (Callable, optional): Registers a drag callback for drag and drop.
		drop_callback (Callable, optional): Registers a drop callback for drag and drop.
		show (bool, optional): Attempt to render widget.
		pos (Union[List[int], Tuple[int, ...]], optional): Places the item relative to window coordinates, [0,0] is top left.
		filter_key (str, optional): Used by filter widget.
		tracked (bool, optional): Scroll tracking
		track_offset (float, optional): 0.0f:top, 0.5f:center, 1.0f:bottom
		tint_color (Union[List[float], Tuple[float, ...]], optional): Applies a color tint to the entire texture.
		border_color (Union[List[float], Tuple[float, ...]], optional): Displays a border of the specified color around the texture. If the theme style has turned off the border it will not be shown.
		uv_min (Union[List[float], Tuple[float, ...]], optional): Normalized texture coordinates min point.
		uv_max (Union[List[float], Tuple[float, ...]], optional): Normalized texture coordinates max point.
		id (Union[int, str], optional): (deprecated)
	Returns:
		Union[int, str]
	"""

	return internal_dpg.add_image(texture_tag, **kwargs)

def add_image_button(texture_tag, **kwargs):
	"""	 Adds an button with a texture. uv_min and uv_max represent the normalized texture coordinates of the original image that will be shown. Using range (0.0,0.0)->(1.0,1.0) texture coordinates will generally display the entire texture

	Args:
		texture_tag (Union[int, str]): The texture_tag should come from a texture that was added to a texture registry.
		label (str, optional): Overrides 'name' as label.
		user_data (Any, optional): User data for callbacks
		use_internal_label (bool, optional): Use generated internal label instead of user specified (appends ### uuid).
		tag (Union[int, str], optional): Unique id used to programmatically refer to the item.If label is unused this will be the label.
		width (int, optional): Width of the item.
		height (int, optional): Height of the item.
		indent (int, optional): Offsets the widget to the right the specified number multiplied by the indent style.
		parent (Union[int, str], optional): Parent to add this item to. (runtime adding)
		before (Union[int, str], optional): This item will be displayed before the specified item in the parent.
		source (Union[int, str], optional): Overrides 'id' as value storage key.
		payload_type (str, optional): Sender string type must be the same as the target for the target to run the payload_callback.
		callback (Callable, optional): Registers a callback.
		drag_callback (Callable, optional): Registers a drag callback for drag and drop.
		drop_callback (Callable, optional): Registers a drop callback for drag and drop.
		show (bool, optional): Attempt to render widget.
		enabled (bool, optional): Turns off functionality of widget and applies the disabled theme.
		pos (Union[List[int], Tuple[int, ...]], optional): Places the item relative to window coordinates, [0,0] is top left.
		filter_key (str, optional): Used by filter widget.
		tracked (bool, optional): Scroll tracking
		track_offset (float, optional): 0.0f:top, 0.5f:center, 1.0f:bottom
		frame_padding (int, optional): Empty space around the outside of the texture. Button will show around the texture.
		tint_color (Union[List[float], Tuple[float, ...]], optional): Applies a color tint to the entire texture.
		background_color (Union[List[float], Tuple[float, ...]], optional): Displays a border of the specified color around the texture.
		uv_min (Union[List[float], Tuple[float, ...]], optional): Normalized texture coordinates min point.
		uv_max (Union[List[float], Tuple[float, ...]], optional): Normalized texture coordinates max point.
		id (Union[int, str], optional): (deprecated)
	Returns:
		Union[int, str]
	"""

	return internal_dpg.add_image_button(texture_tag, **kwargs)

def add_image_series(texture_tag, bounds_min, bounds_max, **kwargs):
	"""	 Adds an image series to a plot.

	Args:
		texture_tag (Union[int, str]): 
		bounds_min (Any): 
		bounds_max (Any): 
		label (str, optional): Overrides 'name' as label.
		user_data (Any, optional): User data for callbacks
		use_internal_label (bool, optional): Use generated internal label instead of user specified (appends ### uuid).
		tag (Union[int, str], optional): Unique id used to programmatically refer to the item.If label is unused this will be the label.
		parent (Union[int, str], optional): Parent to add this item to. (runtime adding)
		before (Union[int, str], optional): This item will be displayed before the specified item in the parent.
		source (Union[int, str], optional): Overrides 'id' as value storage key.
		show (bool, optional): Attempt to render widget.
		uv_min (Union[List[float], Tuple[float, ...]], optional): normalized texture coordinates
		uv_max (Union[List[float], Tuple[float, ...]], optional): normalized texture coordinates
		tint_color (Union[List[int], Tuple[int, ...]], optional): 
		id (Union[int, str], optional): (deprecated)
	Returns:
		Union[int, str]
	"""

	return internal_dpg.add_image_series(texture_tag, bounds_min, bounds_max, **kwargs)

def add_input_float(**kwargs):
	"""	 Adds input for an float. +/- buttons can be activated by setting the value of step.

	Args:
		label (str, optional): Overrides 'name' as label.
		user_data (Any, optional): User data for callbacks
		use_internal_label (bool, optional): Use generated internal label instead of user specified (appends ### uuid).
		tag (Union[int, str], optional): Unique id used to programmatically refer to the item.If label is unused this will be the label.
		width (int, optional): Width of the item.
		indent (int, optional): Offsets the widget to the right the specified number multiplied by the indent style.
		parent (Union[int, str], optional): Parent to add this item to. (runtime adding)
		before (Union[int, str], optional): This item will be displayed before the specified item in the parent.
		source (Union[int, str], optional): Overrides 'id' as value storage key.
		payload_type (str, optional): Sender string type must be the same as the target for the target to run the payload_callback.
		callback (Callable, optional): Registers a callback.
		drag_callback (Callable, optional): Registers a drag callback for drag and drop.
		drop_callback (Callable, optional): Registers a drop callback for drag and drop.
		show (bool, optional): Attempt to render widget.
		enabled (bool, optional): Turns off functionality of widget and applies the disabled theme.
		pos (Union[List[int], Tuple[int, ...]], optional): Places the item relative to window coordinates, [0,0] is top left.
		filter_key (str, optional): Used by filter widget.
		tracked (bool, optional): Scroll tracking
		track_offset (float, optional): 0.0f:top, 0.5f:center, 1.0f:bottom
		default_value (float, optional): 
		format (str, optional): Determines the format the float will be displayed as use python string formatting.
		min_value (float, optional): Value for lower limit of input. By default this limits the step buttons. Use min_clamped to limit manual input.
		max_value (float, optional): Value for upper limit of input. By default this limits the step buttons. Use max_clamped to limit manual input.
		step (float, optional): Increment to change value by when the step buttons are pressed. Setting this to a value of 0 or smaller will turn off step buttons.
		step_fast (float, optional): After holding the step buttons for extended time the increments will switch to this value.
		min_clamped (bool, optional): Activates and deactivates the enforcment of min_value.
		max_clamped (bool, optional): Activates and deactivates the enforcment of max_value.
		on_enter (bool, optional): Only runs callback on enter key press.
		readonly (bool, optional): Activates read only mode where no text can be input but text can still be highlighted.
		id (Union[int, str], optional): (deprecated)
	Returns:
		Union[int, str]
	"""

	return internal_dpg.add_input_float(**kwargs)

def add_input_floatx(**kwargs):
	"""	 Adds multi float input for up to 4 float values.

	Args:
		label (str, optional): Overrides 'name' as label.
		user_data (Any, optional): User data for callbacks
		use_internal_label (bool, optional): Use generated internal label instead of user specified (appends ### uuid).
		tag (Union[int, str], optional): Unique id used to programmatically refer to the item.If label is unused this will be the label.
		width (int, optional): Width of the item.
		indent (int, optional): Offsets the widget to the right the specified number multiplied by the indent style.
		parent (Union[int, str], optional): Parent to add this item to. (runtime adding)
		before (Union[int, str], optional): This item will be displayed before the specified item in the parent.
		source (Union[int, str], optional): Overrides 'id' as value storage key.
		payload_type (str, optional): Sender string type must be the same as the target for the target to run the payload_callback.
		callback (Callable, optional): Registers a callback.
		drag_callback (Callable, optional): Registers a drag callback for drag and drop.
		drop_callback (Callable, optional): Registers a drop callback for drag and drop.
		show (bool, optional): Attempt to render widget.
		enabled (bool, optional): Turns off functionality of widget and applies the disabled theme.
		pos (Union[List[int], Tuple[int, ...]], optional): Places the item relative to window coordinates, [0,0] is top left.
		filter_key (str, optional): Used by filter widget.
		tracked (bool, optional): Scroll tracking
		track_offset (float, optional): 0.0f:top, 0.5f:center, 1.0f:bottom
		default_value (Union[List[float], Tuple[float, ...]], optional): 
		format (str, optional): Determines the format the float will be displayed as use python string formatting.
		min_value (float, optional): Value for lower limit of input for each cell. Use min_clamped to turn on.
		max_value (float, optional): Value for upper limit of input for each cell. Use max_clamped to turn on.
		size (int, optional): Number of components displayed for input.
		min_clamped (bool, optional): Activates and deactivates the enforcment of min_value.
		max_clamped (bool, optional): Activates and deactivates the enforcment of max_value.
		on_enter (bool, optional): Only runs callback on enter key press.
		readonly (bool, optional): Activates read only mode where no text can be input but text can still be highlighted.
		id (Union[int, str], optional): (deprecated)
	Returns:
		Union[int, str]
	"""

	return internal_dpg.add_input_floatx(**kwargs)

def add_input_int(**kwargs):
	"""	 Adds input for an int. +/- buttons can be activated by setting the value of step.

	Args:
		label (str, optional): Overrides 'name' as label.
		user_data (Any, optional): User data for callbacks
		use_internal_label (bool, optional): Use generated internal label instead of user specified (appends ### uuid).
		tag (Union[int, str], optional): Unique id used to programmatically refer to the item.If label is unused this will be the label.
		width (int, optional): Width of the item.
		indent (int, optional): Offsets the widget to the right the specified number multiplied by the indent style.
		parent (Union[int, str], optional): Parent to add this item to. (runtime adding)
		before (Union[int, str], optional): This item will be displayed before the specified item in the parent.
		source (Union[int, str], optional): Overrides 'id' as value storage key.
		payload_type (str, optional): Sender string type must be the same as the target for the target to run the payload_callback.
		callback (Callable, optional): Registers a callback.
		drag_callback (Callable, optional): Registers a drag callback for drag and drop.
		drop_callback (Callable, optional): Registers a drop callback for drag and drop.
		show (bool, optional): Attempt to render widget.
		enabled (bool, optional): Turns off functionality of widget and applies the disabled theme.
		pos (Union[List[int], Tuple[int, ...]], optional): Places the item relative to window coordinates, [0,0] is top left.
		filter_key (str, optional): Used by filter widget.
		tracked (bool, optional): Scroll tracking
		track_offset (float, optional): 0.0f:top, 0.5f:center, 1.0f:bottom
		default_value (int, optional): 
		min_value (int, optional): Value for lower limit of input. By default this limits the step buttons. Use min_clamped to limit manual input.
		max_value (int, optional): Value for upper limit of input. By default this limits the step buttons. Use max_clamped to limit manual input.
		step (int, optional): Increment to change value by when the step buttons are pressed. Setting this to a value of 0 or smaller will turn off step buttons.
		step_fast (int, optional): After holding the step buttons for extended time the increments will switch to this value.
		min_clamped (bool, optional): Activates and deactivates the enforcment of min_value.
		max_clamped (bool, optional): Activates and deactivates the enforcment of max_value.
		on_enter (bool, optional): Only runs callback on enter key press.
		readonly (bool, optional): Activates read only mode where no text can be input but text can still be highlighted.
		id (Union[int, str], optional): (deprecated)
	Returns:
		Union[int, str]
	"""

	return internal_dpg.add_input_int(**kwargs)

def add_input_intx(**kwargs):
	"""	 Adds multi int input for up to 4 integer values.

	Args:
		label (str, optional): Overrides 'name' as label.
		user_data (Any, optional): User data for callbacks
		use_internal_label (bool, optional): Use generated internal label instead of user specified (appends ### uuid).
		tag (Union[int, str], optional): Unique id used to programmatically refer to the item.If label is unused this will be the label.
		width (int, optional): Width of the item.
		indent (int, optional): Offsets the widget to the right the specified number multiplied by the indent style.
		parent (Union[int, str], optional): Parent to add this item to. (runtime adding)
		before (Union[int, str], optional): This item will be displayed before the specified item in the parent.
		source (Union[int, str], optional): Overrides 'id' as value storage key.
		payload_type (str, optional): Sender string type must be the same as the target for the target to run the payload_callback.
		callback (Callable, optional): Registers a callback.
		drag_callback (Callable, optional): Registers a drag callback for drag and drop.
		drop_callback (Callable, optional): Registers a drop callback for drag and drop.
		show (bool, optional): Attempt to render widget.
		enabled (bool, optional): Turns off functionality of widget and applies the disabled theme.
		pos (Union[List[int], Tuple[int, ...]], optional): Places the item relative to window coordinates, [0,0] is top left.
		filter_key (str, optional): Used by filter widget.
		tracked (bool, optional): Scroll tracking
		track_offset (float, optional): 0.0f:top, 0.5f:center, 1.0f:bottom
		default_value (Union[List[int], Tuple[int, ...]], optional): 
		min_value (int, optional): Value for lower limit of input for each cell. Use min_clamped to turn on.
		max_value (int, optional): Value for upper limit of input for each cell. Use max_clamped to turn on.
		size (int, optional): Number of components displayed for input.
		min_clamped (bool, optional): Activates and deactivates the enforcment of min_value.
		max_clamped (bool, optional): Activates and deactivates the enforcment of max_value.
		on_enter (bool, optional): Only runs callback on enter.
		readonly (bool, optional): Activates read only mode where no text can be input but text can still be highlighted.
		id (Union[int, str], optional): (deprecated)
	Returns:
		Union[int, str]
	"""

	return internal_dpg.add_input_intx(**kwargs)

def add_input_text(**kwargs):
	"""	 Adds input for text.

	Args:
		label (str, optional): Overrides 'name' as label.
		user_data (Any, optional): User data for callbacks
		use_internal_label (bool, optional): Use generated internal label instead of user specified (appends ### uuid).
		tag (Union[int, str], optional): Unique id used to programmatically refer to the item.If label is unused this will be the label.
		width (int, optional): Width of the item.
		height (int, optional): Height of the item.
		indent (int, optional): Offsets the widget to the right the specified number multiplied by the indent style.
		parent (Union[int, str], optional): Parent to add this item to. (runtime adding)
		before (Union[int, str], optional): This item will be displayed before the specified item in the parent.
		source (Union[int, str], optional): Overrides 'id' as value storage key.
		payload_type (str, optional): Sender string type must be the same as the target for the target to run the payload_callback.
		callback (Callable, optional): Registers a callback.
		drag_callback (Callable, optional): Registers a drag callback for drag and drop.
		drop_callback (Callable, optional): Registers a drop callback for drag and drop.
		show (bool, optional): Attempt to render widget.
		enabled (bool, optional): Turns off functionality of widget and applies the disabled theme.
		pos (Union[List[int], Tuple[int, ...]], optional): Places the item relative to window coordinates, [0,0] is top left.
		filter_key (str, optional): Used by filter widget.
		tracked (bool, optional): Scroll tracking
		track_offset (float, optional): 0.0f:top, 0.5f:center, 1.0f:bottom
		default_value (str, optional): 
		hint (str, optional): Displayed only when value is an empty string. Will reappear if input value is set to empty string. Will not show if default value is anything other than default empty string.
		multiline (bool, optional): Allows for multiline text input.
		no_spaces (bool, optional): Filter out spaces and tabs.
		uppercase (bool, optional): Automatically make all inputs uppercase.
		tab_input (bool, optional): Allows tabs to be input into the string value instead of changing item focus.
		decimal (bool, optional): Only allow characters 0123456789.+-*/
		hexadecimal (bool, optional): Only allow characters 0123456789ABCDEFabcdef
		readonly (bool, optional): Activates read only mode where no text can be input but text can still be highlighted.
		password (bool, optional): Display all input characters as '*'.
		scientific (bool, optional): Only allow characters 0123456789.+-*/eE (Scientific notation input)
		on_enter (bool, optional): Only runs callback on enter key press.
		id (Union[int, str], optional): (deprecated)
	Returns:
		Union[int, str]
	"""

	return internal_dpg.add_input_text(**kwargs)

def add_int4_value(**kwargs):
	"""	 Adds a int4 value.

	Args:
		label (str, optional): Overrides 'name' as label.
		user_data (Any, optional): User data for callbacks
		use_internal_label (bool, optional): Use generated internal label instead of user specified (appends ### uuid).
		tag (Union[int, str], optional): Unique id used to programmatically refer to the item.If label is unused this will be the label.
		source (Union[int, str], optional): Overrides 'id' as value storage key.
		default_value (Union[List[int], Tuple[int, ...]], optional): 
		parent (Union[int, str], optional): Parent to add this item to. (runtime adding)
		id (Union[int, str], optional): (deprecated)
	Returns:
		Union[int, str]
	"""

	return internal_dpg.add_int4_value(**kwargs)

def add_int_value(**kwargs):
	"""	 Adds a int value.

	Args:
		label (str, optional): Overrides 'name' as label.
		user_data (Any, optional): User data for callbacks
		use_internal_label (bool, optional): Use generated internal label instead of user specified (appends ### uuid).
		tag (Union[int, str], optional): Unique id used to programmatically refer to the item.If label is unused this will be the label.
		source (Union[int, str], optional): Overrides 'id' as value storage key.
		default_value (int, optional): 
		parent (Union[int, str], optional): Parent to add this item to. (runtime adding)
		id (Union[int, str], optional): (deprecated)
	Returns:
		Union[int, str]
	"""

	return internal_dpg.add_int_value(**kwargs)

def add_item_activated_handler(**kwargs):
	"""	 Adds a activated handler.

	Args:
		label (str, optional): Overrides 'name' as label.
		user_data (Any, optional): User data for callbacks
		use_internal_label (bool, optional): Use generated internal label instead of user specified (appends ### uuid).
		tag (Union[int, str], optional): Unique id used to programmatically refer to the item.If label is unused this will be the label.
		parent (Union[int, str], optional): Parent to add this item to. (runtime adding)
		callback (Callable, optional): Registers a callback.
		show (bool, optional): Attempt to render widget.
		id (Union[int, str], optional): (deprecated)
	Returns:
		Union[int, str]
	"""

	return internal_dpg.add_item_activated_handler(**kwargs)

def add_item_active_handler(**kwargs):
	"""	 Adds a active handler.

	Args:
		label (str, optional): Overrides 'name' as label.
		user_data (Any, optional): User data for callbacks
		use_internal_label (bool, optional): Use generated internal label instead of user specified (appends ### uuid).
		tag (Union[int, str], optional): Unique id used to programmatically refer to the item.If label is unused this will be the label.
		parent (Union[int, str], optional): Parent to add this item to. (runtime adding)
		callback (Callable, optional): Registers a callback.
		show (bool, optional): Attempt to render widget.
		id (Union[int, str], optional): (deprecated)
	Returns:
		Union[int, str]
	"""

	return internal_dpg.add_item_active_handler(**kwargs)

def add_item_clicked_handler(button=-1, **kwargs):
	"""	 Adds a clicked handler.

	Args:
		button (int, optional): Submits callback for all mouse buttons
		label (str, optional): Overrides 'name' as label.
		user_data (Any, optional): User data for callbacks
		use_internal_label (bool, optional): Use generated internal label instead of user specified (appends ### uuid).
		tag (Union[int, str], optional): Unique id used to programmatically refer to the item.If label is unused this will be the label.
		parent (Union[int, str], optional): Parent to add this item to. (runtime adding)
		callback (Callable, optional): Registers a callback.
		show (bool, optional): Attempt to render widget.
		id (Union[int, str], optional): (deprecated)
	Returns:
		Union[int, str]
	"""

	return internal_dpg.add_item_clicked_handler(button, **kwargs)

def add_item_deactivated_after_edit_handler(**kwargs):
	"""	 Adds a deactivated after edit handler.

	Args:
		label (str, optional): Overrides 'name' as label.
		user_data (Any, optional): User data for callbacks
		use_internal_label (bool, optional): Use generated internal label instead of user specified (appends ### uuid).
		tag (Union[int, str], optional): Unique id used to programmatically refer to the item.If label is unused this will be the label.
		parent (Union[int, str], optional): Parent to add this item to. (runtime adding)
		callback (Callable, optional): Registers a callback.
		show (bool, optional): Attempt to render widget.
		id (Union[int, str], optional): (deprecated)
	Returns:
		Union[int, str]
	"""

	return internal_dpg.add_item_deactivated_after_edit_handler(**kwargs)

def add_item_deactivated_handler(**kwargs):
	"""	 Adds a deactivated handler.

	Args:
		label (str, optional): Overrides 'name' as label.
		user_data (Any, optional): User data for callbacks
		use_internal_label (bool, optional): Use generated internal label instead of user specified (appends ### uuid).
		tag (Union[int, str], optional): Unique id used to programmatically refer to the item.If label is unused this will be the label.
		parent (Union[int, str], optional): Parent to add this item to. (runtime adding)
		callback (Callable, optional): Registers a callback.
		show (bool, optional): Attempt to render widget.
		id (Union[int, str], optional): (deprecated)
	Returns:
		Union[int, str]
	"""

	return internal_dpg.add_item_deactivated_handler(**kwargs)

def add_item_edited_handler(**kwargs):
	"""	 Adds an edited handler.

	Args:
		label (str, optional): Overrides 'name' as label.
		user_data (Any, optional): User data for callbacks
		use_internal_label (bool, optional): Use generated internal label instead of user specified (appends ### uuid).
		tag (Union[int, str], optional): Unique id used to programmatically refer to the item.If label is unused this will be the label.
		parent (Union[int, str], optional): Parent to add this item to. (runtime adding)
		callback (Callable, optional): Registers a callback.
		show (bool, optional): Attempt to render widget.
		id (Union[int, str], optional): (deprecated)
	Returns:
		Union[int, str]
	"""

	return internal_dpg.add_item_edited_handler(**kwargs)

def add_item_focus_handler(**kwargs):
	"""	 Adds a focus handler.

	Args:
		label (str, optional): Overrides 'name' as label.
		user_data (Any, optional): User data for callbacks
		use_internal_label (bool, optional): Use generated internal label instead of user specified (appends ### uuid).
		tag (Union[int, str], optional): Unique id used to programmatically refer to the item.If label is unused this will be the label.
		parent (Union[int, str], optional): Parent to add this item to. (runtime adding)
		callback (Callable, optional): Registers a callback.
		show (bool, optional): Attempt to render widget.
		id (Union[int, str], optional): (deprecated)
	Returns:
		Union[int, str]
	"""

	return internal_dpg.add_item_focus_handler(**kwargs)

def add_item_handler_registry(**kwargs):
	"""	 Adds an item handler registry.

	Args:
		label (str, optional): Overrides 'name' as label.
		user_data (Any, optional): User data for callbacks
		use_internal_label (bool, optional): Use generated internal label instead of user specified (appends ### uuid).
		tag (Union[int, str], optional): Unique id used to programmatically refer to the item.If label is unused this will be the label.
		show (bool, optional): Attempt to render widget.
		id (Union[int, str], optional): (deprecated)
	Returns:
		Union[int, str]
	"""

	return internal_dpg.add_item_handler_registry(**kwargs)

def add_item_hover_handler(**kwargs):
	"""	 Adds a hover handler.

	Args:
		label (str, optional): Overrides 'name' as label.
		user_data (Any, optional): User data for callbacks
		use_internal_label (bool, optional): Use generated internal label instead of user specified (appends ### uuid).
		tag (Union[int, str], optional): Unique id used to programmatically refer to the item.If label is unused this will be the label.
		parent (Union[int, str], optional): Parent to add this item to. (runtime adding)
		callback (Callable, optional): Registers a callback.
		show (bool, optional): Attempt to render widget.
		id (Union[int, str], optional): (deprecated)
	Returns:
		Union[int, str]
	"""

	return internal_dpg.add_item_hover_handler(**kwargs)

def add_item_resize_handler(**kwargs):
	"""	 Adds a resize handler.

	Args:
		label (str, optional): Overrides 'name' as label.
		user_data (Any, optional): User data for callbacks
		use_internal_label (bool, optional): Use generated internal label instead of user specified (appends ### uuid).
		tag (Union[int, str], optional): Unique id used to programmatically refer to the item.If label is unused this will be the label.
		parent (Union[int, str], optional): Parent to add this item to. (runtime adding)
		callback (Callable, optional): Registers a callback.
		show (bool, optional): Attempt to render widget.
		id (Union[int, str], optional): (deprecated)
	Returns:
		Union[int, str]
	"""

	return internal_dpg.add_item_resize_handler(**kwargs)

def add_item_toggled_open_handler(**kwargs):
	"""	 Adds a togged open handler.

	Args:
		label (str, optional): Overrides 'name' as label.
		user_data (Any, optional): User data for callbacks
		use_internal_label (bool, optional): Use generated internal label instead of user specified (appends ### uuid).
		tag (Union[int, str], optional): Unique id used to programmatically refer to the item.If label is unused this will be the label.
		parent (Union[int, str], optional): Parent to add this item to. (runtime adding)
		callback (Callable, optional): Registers a callback.
		show (bool, optional): Attempt to render widget.
		id (Union[int, str], optional): (deprecated)
	Returns:
		Union[int, str]
	"""

	return internal_dpg.add_item_toggled_open_handler(**kwargs)

def add_item_visible_handler(**kwargs):
	"""	 Adds a visible handler.

	Args:
		label (str, optional): Overrides 'name' as label.
		user_data (Any, optional): User data for callbacks
		use_internal_label (bool, optional): Use generated internal label instead of user specified (appends ### uuid).
		tag (Union[int, str], optional): Unique id used to programmatically refer to the item.If label is unused this will be the label.
		parent (Union[int, str], optional): Parent to add this item to. (runtime adding)
		callback (Callable, optional): Registers a callback.
		show (bool, optional): Attempt to render widget.
		id (Union[int, str], optional): (deprecated)
	Returns:
		Union[int, str]
	"""

	return internal_dpg.add_item_visible_handler(**kwargs)

def add_key_down_handler(key=-1, **kwargs):
	"""	 Adds a key down handler.

	Args:
		key (int, optional): Submits callback for all keys
		label (str, optional): Overrides 'name' as label.
		user_data (Any, optional): User data for callbacks
		use_internal_label (bool, optional): Use generated internal label instead of user specified (appends ### uuid).
		tag (Union[int, str], optional): Unique id used to programmatically refer to the item.If label is unused this will be the label.
		callback (Callable, optional): Registers a callback.
		show (bool, optional): Attempt to render widget.
		parent (Union[int, str], optional): Parent to add this item to. (runtime adding)
		id (Union[int, str], optional): (deprecated)
	Returns:
		Union[int, str]
	"""

	return internal_dpg.add_key_down_handler(key, **kwargs)

def add_key_press_handler(key=-1, **kwargs):
	"""	 Adds a key press handler.

	Args:
		key (int, optional): Submits callback for all keys
		label (str, optional): Overrides 'name' as label.
		user_data (Any, optional): User data for callbacks
		use_internal_label (bool, optional): Use generated internal label instead of user specified (appends ### uuid).
		tag (Union[int, str], optional): Unique id used to programmatically refer to the item.If label is unused this will be the label.
		callback (Callable, optional): Registers a callback.
		show (bool, optional): Attempt to render widget.
		parent (Union[int, str], optional): Parent to add this item to. (runtime adding)
		id (Union[int, str], optional): (deprecated)
	Returns:
		Union[int, str]
	"""

	return internal_dpg.add_key_press_handler(key, **kwargs)

def add_key_release_handler(key=-1, **kwargs):
	"""	 Adds a key release handler.

	Args:
		key (int, optional): Submits callback for all keys
		label (str, optional): Overrides 'name' as label.
		user_data (Any, optional): User data for callbacks
		use_internal_label (bool, optional): Use generated internal label instead of user specified (appends ### uuid).
		tag (Union[int, str], optional): Unique id used to programmatically refer to the item.If label is unused this will be the label.
		callback (Callable, optional): Registers a callback.
		show (bool, optional): Attempt to render widget.
		parent (Union[int, str], optional): Parent to add this item to. (runtime adding)
		id (Union[int, str], optional): (deprecated)
	Returns:
		Union[int, str]
	"""

	return internal_dpg.add_key_release_handler(key, **kwargs)

def add_knob_float(**kwargs):
	"""	 Adds a knob that rotates based on change in x mouse position.

	Args:
		label (str, optional): Overrides 'name' as label.
		user_data (Any, optional): User data for callbacks
		use_internal_label (bool, optional): Use generated internal label instead of user specified (appends ### uuid).
		tag (Union[int, str], optional): Unique id used to programmatically refer to the item.If label is unused this will be the label.
		width (int, optional): Width of the item.
		height (int, optional): Height of the item.
		indent (int, optional): Offsets the widget to the right the specified number multiplied by the indent style.
		parent (Union[int, str], optional): Parent to add this item to. (runtime adding)
		before (Union[int, str], optional): This item will be displayed before the specified item in the parent.
		source (Union[int, str], optional): Overrides 'id' as value storage key.
		payload_type (str, optional): Sender string type must be the same as the target for the target to run the payload_callback.
		callback (Callable, optional): Registers a callback.
		drag_callback (Callable, optional): Registers a drag callback for drag and drop.
		drop_callback (Callable, optional): Registers a drop callback for drag and drop.
		show (bool, optional): Attempt to render widget.
		pos (Union[List[int], Tuple[int, ...]], optional): Places the item relative to window coordinates, [0,0] is top left.
		filter_key (str, optional): Used by filter widget.
		tracked (bool, optional): Scroll tracking
		track_offset (float, optional): 0.0f:top, 0.5f:center, 1.0f:bottom
		default_value (float, optional): 
		min_value (float, optional): Applies lower limit to value.
		max_value (float, optional): Applies upper limit to value.
		id (Union[int, str], optional): (deprecated)
	Returns:
		Union[int, str]
	"""

	return internal_dpg.add_knob_float(**kwargs)

def add_line_series(x, y, **kwargs):
	"""	 Adds a line series to a plot.

	Args:
		x (Any): 
		y (Any): 
		label (str, optional): Overrides 'name' as label.
		user_data (Any, optional): User data for callbacks
		use_internal_label (bool, optional): Use generated internal label instead of user specified (appends ### uuid).
		tag (Union[int, str], optional): Unique id used to programmatically refer to the item.If label is unused this will be the label.
		parent (Union[int, str], optional): Parent to add this item to. (runtime adding)
		before (Union[int, str], optional): This item will be displayed before the specified item in the parent.
		source (Union[int, str], optional): Overrides 'id' as value storage key.
		show (bool, optional): Attempt to render widget.
		id (Union[int, str], optional): (deprecated)
	Returns:
		Union[int, str]
	"""

	return internal_dpg.add_line_series(x, y, **kwargs)

def add_listbox(items=(), **kwargs):
	"""	 Adds a listbox. If height is not large enough to show all items a scroll bar will appear.

	Args:
		items (Union[List[str], Tuple[str, ...]], optional): A tuple of items to be shown in the listbox. Can consist of any combination of types. All items will be displayed as strings.
		label (str, optional): Overrides 'name' as label.
		user_data (Any, optional): User data for callbacks
		use_internal_label (bool, optional): Use generated internal label instead of user specified (appends ### uuid).
		tag (Union[int, str], optional): Unique id used to programmatically refer to the item.If label is unused this will be the label.
		width (int, optional): Width of the item.
		indent (int, optional): Offsets the widget to the right the specified number multiplied by the indent style.
		parent (Union[int, str], optional): Parent to add this item to. (runtime adding)
		before (Union[int, str], optional): This item will be displayed before the specified item in the parent.
		source (Union[int, str], optional): Overrides 'id' as value storage key.
		payload_type (str, optional): Sender string type must be the same as the target for the target to run the payload_callback.
		callback (Callable, optional): Registers a callback.
		drag_callback (Callable, optional): Registers a drag callback for drag and drop.
		drop_callback (Callable, optional): Registers a drop callback for drag and drop.
		show (bool, optional): Attempt to render widget.
		enabled (bool, optional): Turns off functionality of widget and applies the disabled theme.
		pos (Union[List[int], Tuple[int, ...]], optional): Places the item relative to window coordinates, [0,0] is top left.
		filter_key (str, optional): Used by filter widget.
		tracked (bool, optional): Scroll tracking
		track_offset (float, optional): 0.0f:top, 0.5f:center, 1.0f:bottom
		default_value (str, optional): String value fo the item that will be selected by default.
		num_items (int, optional): Expands the height of the listbox to show specified number of items.
		id (Union[int, str], optional): (deprecated)
	Returns:
		Union[int, str]
	"""

	return internal_dpg.add_listbox(items, **kwargs)

def add_loading_indicator(**kwargs):
	"""	 Adds a rotating animated loading symbol.

	Args:
		label (str, optional): Overrides 'name' as label.
		user_data (Any, optional): User data for callbacks
		use_internal_label (bool, optional): Use generated internal label instead of user specified (appends ### uuid).
		tag (Union[int, str], optional): Unique id used to programmatically refer to the item.If label is unused this will be the label.
		width (int, optional): Width of the item.
		height (int, optional): Height of the item.
		indent (int, optional): Offsets the widget to the right the specified number multiplied by the indent style.
		parent (Union[int, str], optional): Parent to add this item to. (runtime adding)
		before (Union[int, str], optional): This item will be displayed before the specified item in the parent.
		payload_type (str, optional): Sender string type must be the same as the target for the target to run the payload_callback.
		drop_callback (Callable, optional): Registers a drop callback for drag and drop.
		show (bool, optional): Attempt to render widget.
		pos (Union[List[int], Tuple[int, ...]], optional): Places the item relative to window coordinates, [0,0] is top left.
		style (int, optional): 0 is rotating dots style, 1 is rotating bar style.
		circle_count (int, optional): Number of dots show if dots or size of circle if circle.
		speed (float, optional): Speed the anamation will rotate.
		radius (float, optional): Radius size of the loading indicator.
		thickness (float, optional): Thickness of the circles or line.
		color (Union[List[int], Tuple[int, ...]], optional): Color of the growing center circle.
		secondary_color (Union[List[int], Tuple[int, ...]], optional): Background of the dots in dot mode.
		id (Union[int, str], optional): (deprecated)
	Returns:
		Union[int, str]
	"""

	return internal_dpg.add_loading_indicator(**kwargs)

def add_menu(**kwargs):
	"""	 Adds a menu to an existing menu bar.

	Args:
		label (str, optional): Overrides 'name' as label.
		user_data (Any, optional): User data for callbacks
		use_internal_label (bool, optional): Use generated internal label instead of user specified (appends ### uuid).
		tag (Union[int, str], optional): Unique id used to programmatically refer to the item.If label is unused this will be the label.
		indent (int, optional): Offsets the widget to the right the specified number multiplied by the indent style.
		parent (Union[int, str], optional): Parent to add this item to. (runtime adding)
		before (Union[int, str], optional): This item will be displayed before the specified item in the parent.
		payload_type (str, optional): Sender string type must be the same as the target for the target to run the payload_callback.
		drop_callback (Callable, optional): Registers a drop callback for drag and drop.
		show (bool, optional): Attempt to render widget.
		enabled (bool, optional): Turns off functionality of widget and applies the disabled theme.
		filter_key (str, optional): Used by filter widget.
		delay_search (bool, optional): Delays searching container for specified items until the end of the app. Possible optimization when a container has many children that are not accessed often.
		tracked (bool, optional): Scroll tracking
		track_offset (float, optional): 0.0f:top, 0.5f:center, 1.0f:bottom
		id (Union[int, str], optional): (deprecated)
	Returns:
		Union[int, str]
	"""

	return internal_dpg.add_menu(**kwargs)

def add_menu_bar(**kwargs):
	"""	 Adds a menu bar to a window.

	Args:
		label (str, optional): Overrides 'name' as label.
		user_data (Any, optional): User data for callbacks
		use_internal_label (bool, optional): Use generated internal label instead of user specified (appends ### uuid).
		tag (Union[int, str], optional): Unique id used to programmatically refer to the item.If label is unused this will be the label.
		indent (int, optional): Offsets the widget to the right the specified number multiplied by the indent style.
		parent (Union[int, str], optional): Parent to add this item to. (runtime adding)
		show (bool, optional): Attempt to render widget.
		delay_search (bool, optional): Delays searching container for specified items until the end of the app. Possible optimization when a container has many children that are not accessed often.
		id (Union[int, str], optional): (deprecated)
	Returns:
		Union[int, str]
	"""

	return internal_dpg.add_menu_bar(**kwargs)

def add_menu_item(**kwargs):
	"""	 Adds a menu item to an existing menu. Menu items act similar to selectables and has a bool value. When placed in a menu the checkmark will reflect its value.

	Args:
		label (str, optional): Overrides 'name' as label.
		user_data (Any, optional): User data for callbacks
		use_internal_label (bool, optional): Use generated internal label instead of user specified (appends ### uuid).
		tag (Union[int, str], optional): Unique id used to programmatically refer to the item.If label is unused this will be the label.
		indent (int, optional): Offsets the widget to the right the specified number multiplied by the indent style.
		parent (Union[int, str], optional): Parent to add this item to. (runtime adding)
		before (Union[int, str], optional): This item will be displayed before the specified item in the parent.
		payload_type (str, optional): Sender string type must be the same as the target for the target to run the payload_callback.
		callback (Callable, optional): Registers a callback.
		drop_callback (Callable, optional): Registers a drop callback for drag and drop.
		show (bool, optional): Attempt to render widget.
		enabled (bool, optional): Turns off functionality of widget and applies the disabled theme.
		filter_key (str, optional): Used by filter widget.
		tracked (bool, optional): Scroll tracking
		track_offset (float, optional): 0.0f:top, 0.5f:center, 1.0f:bottom
		default_value (bool, optional): This value also controls the checkmark when shown.
		shortcut (str, optional): Displays text on the menu item. Typically used to show a shortcut key command.
		check (bool, optional): Displays a checkmark on the menu item when it is selected and placed in a menu.
		id (Union[int, str], optional): (deprecated)
		drag_callback (Callable, optional): (deprecated)
	Returns:
		Union[int, str]
	"""

	return internal_dpg.add_menu_item(**kwargs)

def add_mouse_click_handler(button=-1, **kwargs):
	"""	 Adds a mouse click handler.

	Args:
		button (int, optional): Submits callback for all mouse buttons
		label (str, optional): Overrides 'name' as label.
		user_data (Any, optional): User data for callbacks
		use_internal_label (bool, optional): Use generated internal label instead of user specified (appends ### uuid).
		tag (Union[int, str], optional): Unique id used to programmatically refer to the item.If label is unused this will be the label.
		callback (Callable, optional): Registers a callback.
		show (bool, optional): Attempt to render widget.
		parent (Union[int, str], optional): Parent to add this item to. (runtime adding)
		id (Union[int, str], optional): (deprecated)
	Returns:
		Union[int, str]
	"""

	return internal_dpg.add_mouse_click_handler(button, **kwargs)

def add_mouse_double_click_handler(button=-1, **kwargs):
	"""	 Adds a mouse double click handler.

	Args:
		button (int, optional): Submits callback for all mouse buttons
		label (str, optional): Overrides 'name' as label.
		user_data (Any, optional): User data for callbacks
		use_internal_label (bool, optional): Use generated internal label instead of user specified (appends ### uuid).
		tag (Union[int, str], optional): Unique id used to programmatically refer to the item.If label is unused this will be the label.
		callback (Callable, optional): Registers a callback.
		show (bool, optional): Attempt to render widget.
		parent (Union[int, str], optional): Parent to add this item to. (runtime adding)
		id (Union[int, str], optional): (deprecated)
	Returns:
		Union[int, str]
	"""

	return internal_dpg.add_mouse_double_click_handler(button, **kwargs)

def add_mouse_down_handler(button=-1, **kwargs):
	"""	 Adds a mouse down handler.

	Args:
		button (int, optional): Submits callback for all mouse buttons
		label (str, optional): Overrides 'name' as label.
		user_data (Any, optional): User data for callbacks
		use_internal_label (bool, optional): Use generated internal label instead of user specified (appends ### uuid).
		tag (Union[int, str], optional): Unique id used to programmatically refer to the item.If label is unused this will be the label.
		callback (Callable, optional): Registers a callback.
		show (bool, optional): Attempt to render widget.
		parent (Union[int, str], optional): Parent to add this item to. (runtime adding)
		id (Union[int, str], optional): (deprecated)
	Returns:
		Union[int, str]
	"""

	return internal_dpg.add_mouse_down_handler(button, **kwargs)

def add_mouse_drag_handler(button=-1, threshold=10.0, **kwargs):
	"""	 Adds a mouse drag handler.

	Args:
		button (int, optional): Submits callback for all mouse buttons
		threshold (float, optional): The threshold the mouse must be dragged before the callback is ran
		label (str, optional): Overrides 'name' as label.
		user_data (Any, optional): User data for callbacks
		use_internal_label (bool, optional): Use generated internal label instead of user specified (appends ### uuid).
		tag (Union[int, str], optional): Unique id used to programmatically refer to the item.If label is unused this will be the label.
		callback (Callable, optional): Registers a callback.
		show (bool, optional): Attempt to render widget.
		parent (Union[int, str], optional): Parent to add this item to. (runtime adding)
		id (Union[int, str], optional): (deprecated)
	Returns:
		Union[int, str]
	"""

	return internal_dpg.add_mouse_drag_handler(button, threshold, **kwargs)

def add_mouse_move_handler(**kwargs):
	"""	 Adds a mouse move handler.

	Args:
		label (str, optional): Overrides 'name' as label.
		user_data (Any, optional): User data for callbacks
		use_internal_label (bool, optional): Use generated internal label instead of user specified (appends ### uuid).
		tag (Union[int, str], optional): Unique id used to programmatically refer to the item.If label is unused this will be the label.
		callback (Callable, optional): Registers a callback.
		show (bool, optional): Attempt to render widget.
		parent (Union[int, str], optional): Parent to add this item to. (runtime adding)
		id (Union[int, str], optional): (deprecated)
	Returns:
		Union[int, str]
	"""

	return internal_dpg.add_mouse_move_handler(**kwargs)

def add_mouse_release_handler(button=-1, **kwargs):
	"""	 Adds a mouse release handler.

	Args:
		button (int, optional): Submits callback for all mouse buttons
		label (str, optional): Overrides 'name' as label.
		user_data (Any, optional): User data for callbacks
		use_internal_label (bool, optional): Use generated internal label instead of user specified (appends ### uuid).
		tag (Union[int, str], optional): Unique id used to programmatically refer to the item.If label is unused this will be the label.
		callback (Callable, optional): Registers a callback.
		show (bool, optional): Attempt to render widget.
		parent (Union[int, str], optional): Parent to add this item to. (runtime adding)
		id (Union[int, str], optional): (deprecated)
	Returns:
		Union[int, str]
	"""

	return internal_dpg.add_mouse_release_handler(button, **kwargs)

def add_mouse_wheel_handler(**kwargs):
	"""	 Adds a mouse wheel handler.

	Args:
		label (str, optional): Overrides 'name' as label.
		user_data (Any, optional): User data for callbacks
		use_internal_label (bool, optional): Use generated internal label instead of user specified (appends ### uuid).
		tag (Union[int, str], optional): Unique id used to programmatically refer to the item.If label is unused this will be the label.
		callback (Callable, optional): Registers a callback.
		show (bool, optional): Attempt to render widget.
		parent (Union[int, str], optional): Parent to add this item to. (runtime adding)
		id (Union[int, str], optional): (deprecated)
	Returns:
		Union[int, str]
	"""

	return internal_dpg.add_mouse_wheel_handler(**kwargs)

def add_node(**kwargs):
	"""	 Adds a node to a node editor.

	Args:
		label (str, optional): Overrides 'name' as label.
		user_data (Any, optional): User data for callbacks
		use_internal_label (bool, optional): Use generated internal label instead of user specified (appends ### uuid).
		tag (Union[int, str], optional): Unique id used to programmatically refer to the item.If label is unused this will be the label.
		parent (Union[int, str], optional): Parent to add this item to. (runtime adding)
		before (Union[int, str], optional): This item will be displayed before the specified item in the parent.
		payload_type (str, optional): Sender string type must be the same as the target for the target to run the payload_callback.
		drag_callback (Callable, optional): Registers a drag callback for drag and drop.
		drop_callback (Callable, optional): Registers a drop callback for drag and drop.
		show (bool, optional): Attempt to render widget.
		pos (Union[List[int], Tuple[int, ...]], optional): Places the item relative to window coordinates, [0,0] is top left.
		filter_key (str, optional): Used by filter widget.
		delay_search (bool, optional): Delays searching container for specified items until the end of the app. Possible optimization when a container has many children that are not accessed often.
		tracked (bool, optional): Scroll tracking
		track_offset (float, optional): 0.0f:top, 0.5f:center, 1.0f:bottom
		draggable (bool, optional): Allow node to be draggable.
		id (Union[int, str], optional): (deprecated)
	Returns:
		Union[int, str]
	"""

	return internal_dpg.add_node(**kwargs)

def add_node_attribute(**kwargs):
	"""	 Adds a node attribute to a node.

	Args:
		label (str, optional): Overrides 'name' as label.
		user_data (Any, optional): User data for callbacks
		use_internal_label (bool, optional): Use generated internal label instead of user specified (appends ### uuid).
		tag (Union[int, str], optional): Unique id used to programmatically refer to the item.If label is unused this will be the label.
		indent (int, optional): Offsets the widget to the right the specified number multiplied by the indent style.
		parent (Union[int, str], optional): Parent to add this item to. (runtime adding)
		before (Union[int, str], optional): This item will be displayed before the specified item in the parent.
		show (bool, optional): Attempt to render widget.
		filter_key (str, optional): Used by filter widget.
		tracked (bool, optional): Scroll tracking
		track_offset (float, optional): 0.0f:top, 0.5f:center, 1.0f:bottom
		attribute_type (int, optional): mvNode_Attr_Input, mvNode_Attr_Output, or mvNode_Attr_Static.
		shape (int, optional): Pin shape.
		category (str, optional): Category
		id (Union[int, str], optional): (deprecated)
	Returns:
		Union[int, str]
	"""

	return internal_dpg.add_node_attribute(**kwargs)

def add_node_editor(**kwargs):
	"""	 Adds a node editor.

	Args:
		label (str, optional): Overrides 'name' as label.
		user_data (Any, optional): User data for callbacks
		use_internal_label (bool, optional): Use generated internal label instead of user specified (appends ### uuid).
		tag (Union[int, str], optional): Unique id used to programmatically refer to the item.If label is unused this will be the label.
		width (int, optional): Width of the item.
		height (int, optional): Height of the item.
		parent (Union[int, str], optional): Parent to add this item to. (runtime adding)
		before (Union[int, str], optional): This item will be displayed before the specified item in the parent.
		callback (Callable, optional): Registers a callback.
		show (bool, optional): Attempt to render widget.
		filter_key (str, optional): Used by filter widget.
		delay_search (bool, optional): Delays searching container for specified items until the end of the app. Possible optimization when a container has many children that are not accessed often.
		tracked (bool, optional): Scroll tracking
		track_offset (float, optional): 0.0f:top, 0.5f:center, 1.0f:bottom
		delink_callback (Callable, optional): Callback ran when a link is detached.
		menubar (bool, optional): Shows or hides the menubar.
		id (Union[int, str], optional): (deprecated)
	Returns:
		Union[int, str]
	"""

	return internal_dpg.add_node_editor(**kwargs)

def add_node_link(attr_1, attr_2, **kwargs):
	"""	 Adds a node link between 2 node attributes.

	Args:
		attr_1 (Union[int, str]): 
		attr_2 (Union[int, str]): 
		label (str, optional): Overrides 'name' as label.
		user_data (Any, optional): User data for callbacks
		use_internal_label (bool, optional): Use generated internal label instead of user specified (appends ### uuid).
		tag (Union[int, str], optional): Unique id used to programmatically refer to the item.If label is unused this will be the label.
		parent (Union[int, str], optional): Parent to add this item to. (runtime adding)
		show (bool, optional): Attempt to render widget.
		id (Union[int, str], optional): (deprecated)
	Returns:
		Union[int, str]
	"""

	return internal_dpg.add_node_link(attr_1, attr_2, **kwargs)

def add_pie_series(x, y, radius, values, labels, **kwargs):
	"""	 Adds an pie series to a plot.

	Args:
		x (float): 
		y (float): 
		radius (float): 
		values (Any): 
		labels (Union[List[str], Tuple[str, ...]]): 
		label (str, optional): Overrides 'name' as label.
		user_data (Any, optional): User data for callbacks
		use_internal_label (bool, optional): Use generated internal label instead of user specified (appends ### uuid).
		tag (Union[int, str], optional): Unique id used to programmatically refer to the item.If label is unused this will be the label.
		parent (Union[int, str], optional): Parent to add this item to. (runtime adding)
		before (Union[int, str], optional): This item will be displayed before the specified item in the parent.
		source (Union[int, str], optional): Overrides 'id' as value storage key.
		show (bool, optional): Attempt to render widget.
		format (str, optional): 
		angle (float, optional): 
		normalize (bool, optional): 
		id (Union[int, str], optional): (deprecated)
	Returns:
		Union[int, str]
	"""

	return internal_dpg.add_pie_series(x, y, radius, values, labels, **kwargs)

def add_plot(**kwargs):
	"""	 Adds a plot which is used to hold series, and can be drawn to with draw commands.

	Args:
		label (str, optional): Overrides 'name' as label.
		user_data (Any, optional): User data for callbacks
		use_internal_label (bool, optional): Use generated internal label instead of user specified (appends ### uuid).
		tag (Union[int, str], optional): Unique id used to programmatically refer to the item.If label is unused this will be the label.
		width (int, optional): Width of the item.
		height (int, optional): Height of the item.
		indent (int, optional): Offsets the widget to the right the specified number multiplied by the indent style.
		parent (Union[int, str], optional): Parent to add this item to. (runtime adding)
		before (Union[int, str], optional): This item will be displayed before the specified item in the parent.
		payload_type (str, optional): Sender string type must be the same as the target for the target to run the payload_callback.
		callback (Callable, optional): Registers a callback.
		drag_callback (Callable, optional): Registers a drag callback for drag and drop.
		drop_callback (Callable, optional): Registers a drop callback for drag and drop.
		show (bool, optional): Attempt to render widget.
		pos (Union[List[int], Tuple[int, ...]], optional): Places the item relative to window coordinates, [0,0] is top left.
		filter_key (str, optional): Used by filter widget.
		delay_search (bool, optional): Delays searching container for specified items until the end of the app. Possible optimization when a container has many children that are not accessed often.
		tracked (bool, optional): Scroll tracking
		track_offset (float, optional): 0.0f:top, 0.5f:center, 1.0f:bottom
		no_title (bool, optional): 
		no_menus (bool, optional): 
		no_box_select (bool, optional): 
		no_mouse_pos (bool, optional): 
		no_highlight (bool, optional): 
		no_child (bool, optional): 
		query (bool, optional): 
		crosshairs (bool, optional): 
		anti_aliased (bool, optional): 
		equal_aspects (bool, optional): 
		pan_button (int, optional): enables panning when held
		pan_mod (int, optional): optional modifier that must be held for panning
		fit_button (int, optional): fits visible data when double clicked
		context_menu_button (int, optional): opens plot context menu (if enabled) when clicked
		box_select_button (int, optional): begins box selection when pressed and confirms selection when released
		box_select_mod (int, optional): begins box selection when pressed and confirms selection when released
		box_select_cancel_button (int, optional): cancels active box selection when pressed
		query_button (int, optional): begins query selection when pressed and end query selection when released
		query_mod (int, optional): optional modifier that must be held for query selection
		query_toggle_mod (int, optional): when held, active box selections turn into queries
		horizontal_mod (int, optional): expands active box selection/query horizontally to plot edge when held
		vertical_mod (int, optional): expands active box selection/query vertically to plot edge when held
		id (Union[int, str], optional): (deprecated)
	Returns:
		Union[int, str]
	"""

	return internal_dpg.add_plot(**kwargs)

def add_plot_annotation(**kwargs):
	"""	 Adds an annotation to a plot.

	Args:
		label (str, optional): Overrides 'name' as label.
		user_data (Any, optional): User data for callbacks
		use_internal_label (bool, optional): Use generated internal label instead of user specified (appends ### uuid).
		tag (Union[int, str], optional): Unique id used to programmatically refer to the item.If label is unused this will be the label.
		parent (Union[int, str], optional): Parent to add this item to. (runtime adding)
		before (Union[int, str], optional): This item will be displayed before the specified item in the parent.
		source (Union[int, str], optional): Overrides 'id' as value storage key.
		show (bool, optional): Attempt to render widget.
		default_value (Any, optional): 
		offset (Union[List[float], Tuple[float, ...]], optional): 
		color (Union[List[int], Tuple[int, ...]], optional): 
		clamped (bool, optional): 
		id (Union[int, str], optional): (deprecated)
	Returns:
		Union[int, str]
	"""

	return internal_dpg.add_plot_annotation(**kwargs)

def add_plot_axis(axis, **kwargs):
	"""	 Adds an axis to a plot.

	Args:
		axis (int): 
		label (str, optional): Overrides 'name' as label.
		user_data (Any, optional): User data for callbacks
		use_internal_label (bool, optional): Use generated internal label instead of user specified (appends ### uuid).
		tag (Union[int, str], optional): Unique id used to programmatically refer to the item.If label is unused this will be the label.
		parent (Union[int, str], optional): Parent to add this item to. (runtime adding)
		payload_type (str, optional): Sender string type must be the same as the target for the target to run the payload_callback.
		drop_callback (Callable, optional): Registers a drop callback for drag and drop.
		show (bool, optional): Attempt to render widget.
		no_gridlines (bool, optional): 
		no_tick_marks (bool, optional): 
		no_tick_labels (bool, optional): 
		log_scale (bool, optional): 
		invert (bool, optional): 
		lock_min (bool, optional): 
		lock_max (bool, optional): 
		time (bool, optional): 
		id (Union[int, str], optional): (deprecated)
	Returns:
		Union[int, str]
	"""

	return internal_dpg.add_plot_axis(axis, **kwargs)

def add_plot_legend(**kwargs):
	"""	 Adds a plot legend to a plot.

	Args:
		label (str, optional): Overrides 'name' as label.
		user_data (Any, optional): User data for callbacks
		use_internal_label (bool, optional): Use generated internal label instead of user specified (appends ### uuid).
		tag (Union[int, str], optional): Unique id used to programmatically refer to the item.If label is unused this will be the label.
		parent (Union[int, str], optional): Parent to add this item to. (runtime adding)
		payload_type (str, optional): Sender string type must be the same as the target for the target to run the payload_callback.
		drop_callback (Callable, optional): Registers a drop callback for drag and drop.
		show (bool, optional): Attempt to render widget.
		location (int, optional): location, mvPlot_Location_*
		horizontal (bool, optional): 
		outside (bool, optional): 
		id (Union[int, str], optional): (deprecated)
	Returns:
		Union[int, str]
	"""

	return internal_dpg.add_plot_legend(**kwargs)

def add_progress_bar(**kwargs):
	"""	 Adds a progress bar.

	Args:
		label (str, optional): Overrides 'name' as label.
		user_data (Any, optional): User data for callbacks
		use_internal_label (bool, optional): Use generated internal label instead of user specified (appends ### uuid).
		tag (Union[int, str], optional): Unique id used to programmatically refer to the item.If label is unused this will be the label.
		width (int, optional): Width of the item.
		height (int, optional): Height of the item.
		indent (int, optional): Offsets the widget to the right the specified number multiplied by the indent style.
		parent (Union[int, str], optional): Parent to add this item to. (runtime adding)
		before (Union[int, str], optional): This item will be displayed before the specified item in the parent.
		source (Union[int, str], optional): Overrides 'id' as value storage key.
		payload_type (str, optional): Sender string type must be the same as the target for the target to run the payload_callback.
		drag_callback (Callable, optional): Registers a drag callback for drag and drop.
		drop_callback (Callable, optional): Registers a drop callback for drag and drop.
		show (bool, optional): Attempt to render widget.
		pos (Union[List[int], Tuple[int, ...]], optional): Places the item relative to window coordinates, [0,0] is top left.
		filter_key (str, optional): Used by filter widget.
		tracked (bool, optional): Scroll tracking
		track_offset (float, optional): 0.0f:top, 0.5f:center, 1.0f:bottom
		overlay (str, optional): Overlayed text onto the bar that typically used to display the value of the progress.
		default_value (float, optional): Normalized value to fill the bar from 0.0 to 1.0.
		id (Union[int, str], optional): (deprecated)
	Returns:
		Union[int, str]
	"""

	return internal_dpg.add_progress_bar(**kwargs)

def add_radio_button(items=(), **kwargs):
	"""	 Adds a set of radio buttons. If items keyword is empty, nothing will be shown.

	Args:
		items (Union[List[str], Tuple[str, ...]], optional): A tuple of items to be shown as radio options. Can consist of any combination of types. All types will be shown as strings.
		label (str, optional): Overrides 'name' as label.
		user_data (Any, optional): User data for callbacks
		use_internal_label (bool, optional): Use generated internal label instead of user specified (appends ### uuid).
		tag (Union[int, str], optional): Unique id used to programmatically refer to the item.If label is unused this will be the label.
		indent (int, optional): Offsets the widget to the right the specified number multiplied by the indent style.
		parent (Union[int, str], optional): Parent to add this item to. (runtime adding)
		before (Union[int, str], optional): This item will be displayed before the specified item in the parent.
		source (Union[int, str], optional): Overrides 'id' as value storage key.
		payload_type (str, optional): Sender string type must be the same as the target for the target to run the payload_callback.
		callback (Callable, optional): Registers a callback.
		drag_callback (Callable, optional): Registers a drag callback for drag and drop.
		drop_callback (Callable, optional): Registers a drop callback for drag and drop.
		show (bool, optional): Attempt to render widget.
		enabled (bool, optional): Turns off functionality of widget and applies the disabled theme.
		pos (Union[List[int], Tuple[int, ...]], optional): Places the item relative to window coordinates, [0,0] is top left.
		filter_key (str, optional): Used by filter widget.
		tracked (bool, optional): Scroll tracking
		track_offset (float, optional): 0.0f:top, 0.5f:center, 1.0f:bottom
		default_value (str, optional): Default selected radio option. Set by using the string value of the item.
		horizontal (bool, optional): Displays the radio options horizontally.
		id (Union[int, str], optional): (deprecated)
	Returns:
		Union[int, str]
	"""

	return internal_dpg.add_radio_button(items, **kwargs)

def add_raw_texture(width, height, default_value, **kwargs):
	"""	 Adds a raw texture.

	Args:
		width (int): 
		height (int): 
		default_value (Union[List[float], Tuple[float, ...]]): 
		label (str, optional): Overrides 'name' as label.
		user_data (Any, optional): User data for callbacks
		use_internal_label (bool, optional): Use generated internal label instead of user specified (appends ### uuid).
		tag (Union[int, str], optional): Unique id used to programmatically refer to the item.If label is unused this will be the label.
		format (int, optional): Data format.
		parent (Union[int, str], optional): Parent to add this item to. (runtime adding)
		id (Union[int, str], optional): (deprecated)
	Returns:
		Union[int, str]
	"""

	return internal_dpg.add_raw_texture(width, height, default_value, **kwargs)

def add_scatter_series(x, y, **kwargs):
	"""	 Adds a scatter series to a plot.

	Args:
		x (Any): 
		y (Any): 
		label (str, optional): Overrides 'name' as label.
		user_data (Any, optional): User data for callbacks
		use_internal_label (bool, optional): Use generated internal label instead of user specified (appends ### uuid).
		tag (Union[int, str], optional): Unique id used to programmatically refer to the item.If label is unused this will be the label.
		parent (Union[int, str], optional): Parent to add this item to. (runtime adding)
		before (Union[int, str], optional): This item will be displayed before the specified item in the parent.
		source (Union[int, str], optional): Overrides 'id' as value storage key.
		show (bool, optional): Attempt to render widget.
		id (Union[int, str], optional): (deprecated)
	Returns:
		Union[int, str]
	"""

	return internal_dpg.add_scatter_series(x, y, **kwargs)

def add_selectable(**kwargs):
	"""	 Adds a selectable. Similar to a button but can indicate its selected state.

	Args:
		label (str, optional): Overrides 'name' as label.
		user_data (Any, optional): User data for callbacks
		use_internal_label (bool, optional): Use generated internal label instead of user specified (appends ### uuid).
		tag (Union[int, str], optional): Unique id used to programmatically refer to the item.If label is unused this will be the label.
		width (int, optional): Width of the item.
		height (int, optional): Height of the item.
		indent (int, optional): Offsets the widget to the right the specified number multiplied by the indent style.
		parent (Union[int, str], optional): Parent to add this item to. (runtime adding)
		before (Union[int, str], optional): This item will be displayed before the specified item in the parent.
		source (Union[int, str], optional): Overrides 'id' as value storage key.
		payload_type (str, optional): Sender string type must be the same as the target for the target to run the payload_callback.
		callback (Callable, optional): Registers a callback.
		drag_callback (Callable, optional): Registers a drag callback for drag and drop.
		drop_callback (Callable, optional): Registers a drop callback for drag and drop.
		show (bool, optional): Attempt to render widget.
		enabled (bool, optional): Turns off functionality of widget and applies the disabled theme.
		pos (Union[List[int], Tuple[int, ...]], optional): Places the item relative to window coordinates, [0,0] is top left.
		filter_key (str, optional): Used by filter widget.
		tracked (bool, optional): Scroll tracking
		track_offset (float, optional): 0.0f:top, 0.5f:center, 1.0f:bottom
		default_value (bool, optional): 
		span_columns (bool, optional): Forces the selectable to span the width of all columns if placed in a table.
		id (Union[int, str], optional): (deprecated)
	Returns:
		Union[int, str]
	"""

	return internal_dpg.add_selectable(**kwargs)

def add_separator(**kwargs):
	"""	 Adds a horizontal line separator.

	Args:
		label (str, optional): Overrides 'name' as label.
		user_data (Any, optional): User data for callbacks
		use_internal_label (bool, optional): Use generated internal label instead of user specified (appends ### uuid).
		tag (Union[int, str], optional): Unique id used to programmatically refer to the item.If label is unused this will be the label.
		indent (int, optional): Offsets the widget to the right the specified number multiplied by the indent style.
		parent (Union[int, str], optional): Parent to add this item to. (runtime adding)
		before (Union[int, str], optional): This item will be displayed before the specified item in the parent.
		show (bool, optional): Attempt to render widget.
		pos (Union[List[int], Tuple[int, ...]], optional): Places the item relative to window coordinates, [0,0] is top left.
		id (Union[int, str], optional): (deprecated)
	Returns:
		Union[int, str]
	"""

	return internal_dpg.add_separator(**kwargs)

def add_series_value(**kwargs):
	"""	 Adds a plot series value.

	Args:
		label (str, optional): Overrides 'name' as label.
		user_data (Any, optional): User data for callbacks
		use_internal_label (bool, optional): Use generated internal label instead of user specified (appends ### uuid).
		tag (Union[int, str], optional): Unique id used to programmatically refer to the item.If label is unused this will be the label.
		source (Union[int, str], optional): Overrides 'id' as value storage key.
		default_value (Any, optional): 
		parent (Union[int, str], optional): Parent to add this item to. (runtime adding)
		id (Union[int, str], optional): (deprecated)
	Returns:
		Union[int, str]
	"""

	return internal_dpg.add_series_value(**kwargs)

def add_shade_series(x, y1, **kwargs):
	"""	 Adds a shade series to a plot.

	Args:
		x (Any): 
		y1 (Any): 
		label (str, optional): Overrides 'name' as label.
		user_data (Any, optional): User data for callbacks
		use_internal_label (bool, optional): Use generated internal label instead of user specified (appends ### uuid).
		tag (Union[int, str], optional): Unique id used to programmatically refer to the item.If label is unused this will be the label.
		parent (Union[int, str], optional): Parent to add this item to. (runtime adding)
		before (Union[int, str], optional): This item will be displayed before the specified item in the parent.
		source (Union[int, str], optional): Overrides 'id' as value storage key.
		show (bool, optional): Attempt to render widget.
		y2 (Any, optional): 
		id (Union[int, str], optional): (deprecated)
	Returns:
		Union[int, str]
	"""

	return internal_dpg.add_shade_series(x, y1, **kwargs)

def add_simple_plot(**kwargs):
	"""	 Adds a simple plot for visualization of a 1 dimensional set of values.

	Args:
		label (str, optional): Overrides 'name' as label.
		user_data (Any, optional): User data for callbacks
		use_internal_label (bool, optional): Use generated internal label instead of user specified (appends ### uuid).
		tag (Union[int, str], optional): Unique id used to programmatically refer to the item.If label is unused this will be the label.
		width (int, optional): Width of the item.
		height (int, optional): Height of the item.
		indent (int, optional): Offsets the widget to the right the specified number multiplied by the indent style.
		parent (Union[int, str], optional): Parent to add this item to. (runtime adding)
		before (Union[int, str], optional): This item will be displayed before the specified item in the parent.
		source (Union[int, str], optional): Overrides 'id' as value storage key.
		payload_type (str, optional): Sender string type must be the same as the target for the target to run the payload_callback.
		drag_callback (Callable, optional): Registers a drag callback for drag and drop.
		drop_callback (Callable, optional): Registers a drop callback for drag and drop.
		show (bool, optional): Attempt to render widget.
		filter_key (str, optional): Used by filter widget.
		tracked (bool, optional): Scroll tracking
		track_offset (float, optional): 0.0f:top, 0.5f:center, 1.0f:bottom
		default_value (Union[List[float], Tuple[float, ...]], optional): 
		overlay (str, optional): overlays text (similar to a plot title)
		histogram (bool, optional): 
		autosize (bool, optional): 
		min_scale (float, optional): 
		max_scale (float, optional): 
		id (Union[int, str], optional): (deprecated)
	Returns:
		Union[int, str]
	"""

	return internal_dpg.add_simple_plot(**kwargs)

def add_slider_float(**kwargs):
	"""	 Adds slider for a single float value. Directly entry can be done with double click or CTRL+Click. Min and Max alone are a soft limit for the slider. Use clamped keyword to also apply limits to the direct entry modes.

	Args:
		label (str, optional): Overrides 'name' as label.
		user_data (Any, optional): User data for callbacks
		use_internal_label (bool, optional): Use generated internal label instead of user specified (appends ### uuid).
		tag (Union[int, str], optional): Unique id used to programmatically refer to the item.If label is unused this will be the label.
		width (int, optional): Width of the item.
		height (int, optional): Height of the item.
		indent (int, optional): Offsets the widget to the right the specified number multiplied by the indent style.
		parent (Union[int, str], optional): Parent to add this item to. (runtime adding)
		before (Union[int, str], optional): This item will be displayed before the specified item in the parent.
		source (Union[int, str], optional): Overrides 'id' as value storage key.
		payload_type (str, optional): Sender string type must be the same as the target for the target to run the payload_callback.
		callback (Callable, optional): Registers a callback.
		drag_callback (Callable, optional): Registers a drag callback for drag and drop.
		drop_callback (Callable, optional): Registers a drop callback for drag and drop.
		show (bool, optional): Attempt to render widget.
		enabled (bool, optional): Turns off functionality of widget and applies the disabled theme.
		pos (Union[List[int], Tuple[int, ...]], optional): Places the item relative to window coordinates, [0,0] is top left.
		filter_key (str, optional): Used by filter widget.
		tracked (bool, optional): Scroll tracking
		track_offset (float, optional): 0.0f:top, 0.5f:center, 1.0f:bottom
		default_value (float, optional): 
		vertical (bool, optional): Sets orientation of the slidebar and slider to vertical.
		no_input (bool, optional): Disable direct entry methods double-click or ctrl+click or Enter key allowing to input text directly into the item.
		clamped (bool, optional): Applies the min and max limits to direct entry methods also such as double click and CTRL+Click.
		min_value (float, optional): Applies a limit only to sliding entry only.
		max_value (float, optional): Applies a limit only to sliding entry only.
		format (str, optional): Determines the format the float will be displayed as use python string formatting.
		id (Union[int, str], optional): (deprecated)
	Returns:
		Union[int, str]
	"""

	return internal_dpg.add_slider_float(**kwargs)

def add_slider_floatx(**kwargs):
	"""	 Adds multi slider for up to 4 float values. Directly entry can be done with double click or CTRL+Click. Min and Max alone are a soft limit for the slider. Use clamped keyword to also apply limits to the direct entry modes.

	Args:
		label (str, optional): Overrides 'name' as label.
		user_data (Any, optional): User data for callbacks
		use_internal_label (bool, optional): Use generated internal label instead of user specified (appends ### uuid).
		tag (Union[int, str], optional): Unique id used to programmatically refer to the item.If label is unused this will be the label.
		width (int, optional): Width of the item.
		indent (int, optional): Offsets the widget to the right the specified number multiplied by the indent style.
		parent (Union[int, str], optional): Parent to add this item to. (runtime adding)
		before (Union[int, str], optional): This item will be displayed before the specified item in the parent.
		source (Union[int, str], optional): Overrides 'id' as value storage key.
		payload_type (str, optional): Sender string type must be the same as the target for the target to run the payload_callback.
		callback (Callable, optional): Registers a callback.
		drag_callback (Callable, optional): Registers a drag callback for drag and drop.
		drop_callback (Callable, optional): Registers a drop callback for drag and drop.
		show (bool, optional): Attempt to render widget.
		enabled (bool, optional): Turns off functionality of widget and applies the disabled theme.
		pos (Union[List[int], Tuple[int, ...]], optional): Places the item relative to window coordinates, [0,0] is top left.
		filter_key (str, optional): Used by filter widget.
		tracked (bool, optional): Scroll tracking
		track_offset (float, optional): 0.0f:top, 0.5f:center, 1.0f:bottom
		default_value (Union[List[float], Tuple[float, ...]], optional): 
		size (int, optional): Number of floats to be displayed.
		no_input (bool, optional): Disable direct entry methods double-click or ctrl+click or Enter key allowing to input text directly into the item.
		clamped (bool, optional): Applies the min and max limits to direct entry methods also such as double click and CTRL+Click.
		min_value (float, optional): Applies a limit only to sliding entry only.
		max_value (float, optional): Applies a limit only to sliding entry only.
		format (str, optional): Determines the format the int will be displayed as use python string formatting.
		id (Union[int, str], optional): (deprecated)
	Returns:
		Union[int, str]
	"""

	return internal_dpg.add_slider_floatx(**kwargs)

def add_slider_int(**kwargs):
	"""	 Adds slider for a single int value. Directly entry can be done with double click or CTRL+Click. Min and Max alone are a soft limit for the slider. Use clamped keyword to also apply limits to the direct entry modes.

	Args:
		label (str, optional): Overrides 'name' as label.
		user_data (Any, optional): User data for callbacks
		use_internal_label (bool, optional): Use generated internal label instead of user specified (appends ### uuid).
		tag (Union[int, str], optional): Unique id used to programmatically refer to the item.If label is unused this will be the label.
		width (int, optional): Width of the item.
		height (int, optional): Height of the item.
		indent (int, optional): Offsets the widget to the right the specified number multiplied by the indent style.
		parent (Union[int, str], optional): Parent to add this item to. (runtime adding)
		before (Union[int, str], optional): This item will be displayed before the specified item in the parent.
		source (Union[int, str], optional): Overrides 'id' as value storage key.
		payload_type (str, optional): Sender string type must be the same as the target for the target to run the payload_callback.
		callback (Callable, optional): Registers a callback.
		drag_callback (Callable, optional): Registers a drag callback for drag and drop.
		drop_callback (Callable, optional): Registers a drop callback for drag and drop.
		show (bool, optional): Attempt to render widget.
		enabled (bool, optional): Turns off functionality of widget and applies the disabled theme.
		pos (Union[List[int], Tuple[int, ...]], optional): Places the item relative to window coordinates, [0,0] is top left.
		filter_key (str, optional): Used by filter widget.
		tracked (bool, optional): Scroll tracking
		track_offset (float, optional): 0.0f:top, 0.5f:center, 1.0f:bottom
		default_value (int, optional): 
		vertical (bool, optional): Sets orientation of the slidebar and slider to vertical.
		no_input (bool, optional): Disable direct entry methods double-click or ctrl+click or Enter key allowing to input text directly into the item.
		clamped (bool, optional): Applies the min and max limits to direct entry methods also such as double click and CTRL+Click.
		min_value (int, optional): Applies a limit only to sliding entry only.
		max_value (int, optional): Applies a limit only to sliding entry only.
		format (str, optional): Determines the format the int will be displayed as use python string formatting.
		id (Union[int, str], optional): (deprecated)
	Returns:
		Union[int, str]
	"""

	return internal_dpg.add_slider_int(**kwargs)

def add_slider_intx(**kwargs):
	"""	 Adds multi slider for up to 4 int values. Directly entry can be done with double click or CTRL+Click. Min and Max alone are a soft limit for the slider. Use clamped keyword to also apply limits to the direct entry modes.

	Args:
		label (str, optional): Overrides 'name' as label.
		user_data (Any, optional): User data for callbacks
		use_internal_label (bool, optional): Use generated internal label instead of user specified (appends ### uuid).
		tag (Union[int, str], optional): Unique id used to programmatically refer to the item.If label is unused this will be the label.
		width (int, optional): Width of the item.
		indent (int, optional): Offsets the widget to the right the specified number multiplied by the indent style.
		parent (Union[int, str], optional): Parent to add this item to. (runtime adding)
		before (Union[int, str], optional): This item will be displayed before the specified item in the parent.
		source (Union[int, str], optional): Overrides 'id' as value storage key.
		payload_type (str, optional): Sender string type must be the same as the target for the target to run the payload_callback.
		callback (Callable, optional): Registers a callback.
		drag_callback (Callable, optional): Registers a drag callback for drag and drop.
		drop_callback (Callable, optional): Registers a drop callback for drag and drop.
		show (bool, optional): Attempt to render widget.
		enabled (bool, optional): Turns off functionality of widget and applies the disabled theme.
		pos (Union[List[int], Tuple[int, ...]], optional): Places the item relative to window coordinates, [0,0] is top left.
		filter_key (str, optional): Used by filter widget.
		tracked (bool, optional): Scroll tracking
		track_offset (float, optional): 0.0f:top, 0.5f:center, 1.0f:bottom
		default_value (Union[List[int], Tuple[int, ...]], optional): 
		size (int, optional): Number of ints to be displayed.
		no_input (bool, optional): Disable direct entry methods double-click or ctrl+click or Enter key allowing to input text directly into the item.
		clamped (bool, optional): Applies the min and max limits to direct entry methods also such as double click and CTRL+Click.
		min_value (int, optional): Applies a limit only to sliding entry only.
		max_value (int, optional): Applies a limit only to sliding entry only.
		format (str, optional): Determines the format the int will be displayed as use python string formatting.
		id (Union[int, str], optional): (deprecated)
	Returns:
		Union[int, str]
	"""

	return internal_dpg.add_slider_intx(**kwargs)

def add_spacer(**kwargs):
	"""	 Adds a spacer item that can be used to help with layouts or can be used as a placeholder item.

	Args:
		label (str, optional): Overrides 'name' as label.
		user_data (Any, optional): User data for callbacks
		use_internal_label (bool, optional): Use generated internal label instead of user specified (appends ### uuid).
		tag (Union[int, str], optional): Unique id used to programmatically refer to the item.If label is unused this will be the label.
		width (int, optional): Width of the item.
		height (int, optional): Height of the item.
		indent (int, optional): Offsets the widget to the right the specified number multiplied by the indent style.
		parent (Union[int, str], optional): Parent to add this item to. (runtime adding)
		before (Union[int, str], optional): This item will be displayed before the specified item in the parent.
		show (bool, optional): Attempt to render widget.
		pos (Union[List[int], Tuple[int, ...]], optional): Places the item relative to window coordinates, [0,0] is top left.
		id (Union[int, str], optional): (deprecated)
	Returns:
		Union[int, str]
	"""

	return internal_dpg.add_spacer(**kwargs)

def add_stage(**kwargs):
	"""	 Adds a stage.

	Args:
		label (str, optional): Overrides 'name' as label.
		user_data (Any, optional): User data for callbacks
		use_internal_label (bool, optional): Use generated internal label instead of user specified (appends ### uuid).
		tag (Union[int, str], optional): Unique id used to programmatically refer to the item.If label is unused this will be the label.
		id (Union[int, str], optional): (deprecated)
	Returns:
		Union[int, str]
	"""

	return internal_dpg.add_stage(**kwargs)

def add_stair_series(x, y, **kwargs):
	"""	 Adds a stair series to a plot.

	Args:
		x (Any): 
		y (Any): 
		label (str, optional): Overrides 'name' as label.
		user_data (Any, optional): User data for callbacks
		use_internal_label (bool, optional): Use generated internal label instead of user specified (appends ### uuid).
		tag (Union[int, str], optional): Unique id used to programmatically refer to the item.If label is unused this will be the label.
		parent (Union[int, str], optional): Parent to add this item to. (runtime adding)
		before (Union[int, str], optional): This item will be displayed before the specified item in the parent.
		source (Union[int, str], optional): Overrides 'id' as value storage key.
		show (bool, optional): Attempt to render widget.
		id (Union[int, str], optional): (deprecated)
	Returns:
		Union[int, str]
	"""

	return internal_dpg.add_stair_series(x, y, **kwargs)

def add_static_texture(width, height, default_value, **kwargs):
	"""	 Adds a static texture.

	Args:
		width (int): 
		height (int): 
		default_value (Union[List[float], Tuple[float, ...]]): 
		label (str, optional): Overrides 'name' as label.
		user_data (Any, optional): User data for callbacks
		use_internal_label (bool, optional): Use generated internal label instead of user specified (appends ### uuid).
		tag (Union[int, str], optional): Unique id used to programmatically refer to the item.If label is unused this will be the label.
		parent (Union[int, str], optional): Parent to add this item to. (runtime adding)
		id (Union[int, str], optional): (deprecated)
	Returns:
		Union[int, str]
	"""

	return internal_dpg.add_static_texture(width, height, default_value, **kwargs)

def add_stem_series(x, y, **kwargs):
	"""	 Adds a stem series to a plot.

	Args:
		x (Any): 
		y (Any): 
		label (str, optional): Overrides 'name' as label.
		user_data (Any, optional): User data for callbacks
		use_internal_label (bool, optional): Use generated internal label instead of user specified (appends ### uuid).
		tag (Union[int, str], optional): Unique id used to programmatically refer to the item.If label is unused this will be the label.
		indent (int, optional): Offsets the widget to the right the specified number multiplied by the indent style.
		parent (Union[int, str], optional): Parent to add this item to. (runtime adding)
		before (Union[int, str], optional): This item will be displayed before the specified item in the parent.
		source (Union[int, str], optional): Overrides 'id' as value storage key.
		show (bool, optional): Attempt to render widget.
		id (Union[int, str], optional): (deprecated)
	Returns:
		Union[int, str]
	"""

	return internal_dpg.add_stem_series(x, y, **kwargs)

def add_string_value(**kwargs):
	"""	 Adds a string value.

	Args:
		label (str, optional): Overrides 'name' as label.
		user_data (Any, optional): User data for callbacks
		use_internal_label (bool, optional): Use generated internal label instead of user specified (appends ### uuid).
		tag (Union[int, str], optional): Unique id used to programmatically refer to the item.If label is unused this will be the label.
		source (Union[int, str], optional): Overrides 'id' as value storage key.
		default_value (str, optional): 
		parent (Union[int, str], optional): Parent to add this item to. (runtime adding)
		id (Union[int, str], optional): (deprecated)
	Returns:
		Union[int, str]
	"""

	return internal_dpg.add_string_value(**kwargs)

def add_subplots(rows, columns, **kwargs):
	"""	 Adds a collection of plots.

	Args:
		rows (int): 
		columns (int): 
		label (str, optional): Overrides 'name' as label.
		user_data (Any, optional): User data for callbacks
		use_internal_label (bool, optional): Use generated internal label instead of user specified (appends ### uuid).
		tag (Union[int, str], optional): Unique id used to programmatically refer to the item.If label is unused this will be the label.
		width (int, optional): Width of the item.
		height (int, optional): Height of the item.
		indent (int, optional): Offsets the widget to the right the specified number multiplied by the indent style.
		parent (Union[int, str], optional): Parent to add this item to. (runtime adding)
		before (Union[int, str], optional): This item will be displayed before the specified item in the parent.
		callback (Callable, optional): Registers a callback.
		show (bool, optional): Attempt to render widget.
		pos (Union[List[int], Tuple[int, ...]], optional): Places the item relative to window coordinates, [0,0] is top left.
		filter_key (str, optional): Used by filter widget.
		delay_search (bool, optional): Delays searching container for specified items until the end of the app. Possible optimization when a container has many children that are not accessed often.
		tracked (bool, optional): Scroll tracking
		track_offset (float, optional): 0.0f:top, 0.5f:center, 1.0f:bottom
		row_ratios (Union[List[float], Tuple[float, ...]], optional): 
		column_ratios (Union[List[float], Tuple[float, ...]], optional): 
		no_title (bool, optional): 
		no_menus (bool, optional): the user will not be able to open context menus with right-click
		no_resize (bool, optional): resize splitters between subplot cells will be not be provided
		no_align (bool, optional): subplot edges will not be aligned vertically or horizontally
		link_rows (bool, optional): link the y-axis limits of all plots in each row (does not apply auxiliary y-axes)
		link_columns (bool, optional): link the x-axis limits of all plots in each column
		link_all_x (bool, optional): link the x-axis limits in every plot in the subplot
		link_all_y (bool, optional): link the y-axis limits in every plot in the subplot (does not apply to auxiliary y-axes)
		column_major (bool, optional): subplots are added in column major order instead of the default row major order
		id (Union[int, str], optional): (deprecated)
	Returns:
		Union[int, str]
	"""

	return internal_dpg.add_subplots(rows, columns, **kwargs)

def add_tab(**kwargs):
	"""	 Adds a tab to a tab bar.

	Args:
		label (str, optional): Overrides 'name' as label.
		user_data (Any, optional): User data for callbacks
		use_internal_label (bool, optional): Use generated internal label instead of user specified (appends ### uuid).
		tag (Union[int, str], optional): Unique id used to programmatically refer to the item.If label is unused this will be the label.
		indent (int, optional): Offsets the widget to the right the specified number multiplied by the indent style.
		parent (Union[int, str], optional): Parent to add this item to. (runtime adding)
		before (Union[int, str], optional): This item will be displayed before the specified item in the parent.
		payload_type (str, optional): Sender string type must be the same as the target for the target to run the payload_callback.
		drop_callback (Callable, optional): Registers a drop callback for drag and drop.
		show (bool, optional): Attempt to render widget.
		filter_key (str, optional): Used by filter widget.
		delay_search (bool, optional): Delays searching container for specified items until the end of the app. Possible optimization when a container has many children that are not accessed often.
		tracked (bool, optional): Scroll tracking
		track_offset (float, optional): 0.0f:top, 0.5f:center, 1.0f:bottom
		closable (bool, optional): Creates a button on the tab that can hide the tab.
		no_tooltip (bool, optional): Disable tooltip for the given tab.
		order_mode (bool, optional): set using a constant: mvTabOrder_Reorderable: allows reordering, mvTabOrder_Fixed: fixed ordering, mvTabOrder_Leading: adds tab to front, mvTabOrder_Trailing: adds tab to back
		id (Union[int, str], optional): (deprecated)
	Returns:
		Union[int, str]
	"""

	return internal_dpg.add_tab(**kwargs)

def add_tab_bar(**kwargs):
	"""	 Adds a tab bar.

	Args:
		label (str, optional): Overrides 'name' as label.
		user_data (Any, optional): User data for callbacks
		use_internal_label (bool, optional): Use generated internal label instead of user specified (appends ### uuid).
		tag (Union[int, str], optional): Unique id used to programmatically refer to the item.If label is unused this will be the label.
		indent (int, optional): Offsets the widget to the right the specified number multiplied by the indent style.
		parent (Union[int, str], optional): Parent to add this item to. (runtime adding)
		before (Union[int, str], optional): This item will be displayed before the specified item in the parent.
		callback (Callable, optional): Registers a callback.
		show (bool, optional): Attempt to render widget.
		pos (Union[List[int], Tuple[int, ...]], optional): Places the item relative to window coordinates, [0,0] is top left.
		filter_key (str, optional): Used by filter widget.
		delay_search (bool, optional): Delays searching container for specified items until the end of the app. Possible optimization when a container has many children that are not accessed often.
		tracked (bool, optional): Scroll tracking
		track_offset (float, optional): 0.0f:top, 0.5f:center, 1.0f:bottom
		reorderable (bool, optional): Allows for the user to change the order of the tabs.
		id (Union[int, str], optional): (deprecated)
	Returns:
		Union[int, str]
	"""

	return internal_dpg.add_tab_bar(**kwargs)

def add_tab_button(**kwargs):
	"""	 Adds a tab button to a tab bar.

	Args:
		label (str, optional): Overrides 'name' as label.
		user_data (Any, optional): User data for callbacks
		use_internal_label (bool, optional): Use generated internal label instead of user specified (appends ### uuid).
		tag (Union[int, str], optional): Unique id used to programmatically refer to the item.If label is unused this will be the label.
		indent (int, optional): Offsets the widget to the right the specified number multiplied by the indent style.
		parent (Union[int, str], optional): Parent to add this item to. (runtime adding)
		before (Union[int, str], optional): This item will be displayed before the specified item in the parent.
		payload_type (str, optional): Sender string type must be the same as the target for the target to run the payload_callback.
		callback (Callable, optional): Registers a callback.
		drag_callback (Callable, optional): Registers a drag callback for drag and drop.
		drop_callback (Callable, optional): Registers a drop callback for drag and drop.
		show (bool, optional): Attempt to render widget.
		filter_key (str, optional): Used by filter widget.
		tracked (bool, optional): Scroll tracking
		track_offset (float, optional): 0.0f:top, 0.5f:center, 1.0f:bottom
		no_reorder (bool, optional): Disable reordering this tab or having another tab cross over this tab. Fixes the position of this tab in relation to the order of neighboring tabs at start. 
		leading (bool, optional): Enforce the tab position to the left of the tab bar (after the tab list popup button).
		trailing (bool, optional): Enforce the tab position to the right of the tab bar (before the scrolling buttons).
		no_tooltip (bool, optional): Disable tooltip for the given tab.
		id (Union[int, str], optional): (deprecated)
	Returns:
		Union[int, str]
	"""

	return internal_dpg.add_tab_button(**kwargs)

def add_table(**kwargs):
	"""	 Adds a table.

	Args:
		label (str, optional): Overrides 'name' as label.
		user_data (Any, optional): User data for callbacks
		use_internal_label (bool, optional): Use generated internal label instead of user specified (appends ### uuid).
		tag (Union[int, str], optional): Unique id used to programmatically refer to the item.If label is unused this will be the label.
		width (int, optional): Width of the item.
		height (int, optional): Height of the item.
		indent (int, optional): Offsets the widget to the right the specified number multiplied by the indent style.
		parent (Union[int, str], optional): Parent to add this item to. (runtime adding)
		before (Union[int, str], optional): This item will be displayed before the specified item in the parent.
		source (Union[int, str], optional): Overrides 'id' as value storage key.
		callback (Callable, optional): Registers a callback.
		show (bool, optional): Attempt to render widget.
		pos (Union[List[int], Tuple[int, ...]], optional): Places the item relative to window coordinates, [0,0] is top left.
		filter_key (str, optional): Used by filter widget.
		delay_search (bool, optional): Delays searching container for specified items until the end of the app. Possible optimization when a container has many children that are not accessed often.
		header_row (bool, optional): show headers at the top of the columns
		clipper (bool, optional): Use clipper (rows must be same height).
		inner_width (int, optional): 
		policy (int, optional): 
		freeze_rows (int, optional): 
		freeze_columns (int, optional): 
		sort_multi (bool, optional): Hold shift when clicking headers to sort on multiple column.
		sort_tristate (bool, optional): Allow no sorting, disable default sorting.
		resizable (bool, optional): Enable resizing columns
		reorderable (bool, optional): Enable reordering columns in header row (need calling TableSetupColumn() + TableHeadersRow() to display headers)
		hideable (bool, optional): Enable hiding/disabling columns in context menu.
		sortable (bool, optional): Enable sorting. Call TableGetSortSpecs() to obtain sort specs. Also see ImGuiTableFlags_SortMulti and ImGuiTableFlags_SortTristate.
		context_menu_in_body (bool, optional): Right-click on columns body/contents will display table context menu. By default it is available in TableHeadersRow().
		row_background (bool, optional): Set each RowBg color with ImGuiCol_TableRowBg or ImGuiCol_TableRowBgAlt (equivalent of calling TableSetBgColor with ImGuiTableBgFlags_RowBg0 on each row manually)
		borders_innerH (bool, optional): Draw horizontal borders between rows.
		borders_outerH (bool, optional): Draw horizontal borders at the top and bottom.
		borders_innerV (bool, optional): Draw vertical borders between columns.
		borders_outerV (bool, optional): Draw vertical borders on the left and right sides.
		no_host_extendX (bool, optional): Make outer width auto-fit to columns, overriding outer_size.x value. Only available when ScrollX/ScrollY are disabled and Stretch columns are not used.
		no_host_extendY (bool, optional): Make outer height stop exactly at outer_size.y (prevent auto-extending table past the limit). Only available when ScrollX/ScrollY are disabled. Data below the limit will be clipped and not visible.
		no_keep_columns_visible (bool, optional): Disable keeping column always minimally visible when ScrollX is off and table gets too small. Not recommended if columns are resizable.
		precise_widths (bool, optional): Disable distributing remainder width to stretched columns (width allocation on a 100-wide table with 3 columns: Without this flag: 33,33,34. With this flag: 33,33,33). With larger number of columns, resizing will appear to be less smooth.
		no_clip (bool, optional): Disable clipping rectangle for every individual columns.
		pad_outerX (bool, optional): Default if BordersOuterV is on. Enable outer-most padding. Generally desirable if you have headers.
		no_pad_outerX (bool, optional): Default if BordersOuterV is off. Disable outer-most padding.
		no_pad_innerX (bool, optional): Disable inner padding between columns (double inner padding if BordersOuterV is on, single inner padding if BordersOuterV is off).
		scrollX (bool, optional): Enable horizontal scrolling. Require 'outer_size' parameter of BeginTable() to specify the container size. Changes default sizing policy. Because this create a child window, ScrollY is currently generally recommended when using ScrollX.
		scrollY (bool, optional): Enable vertical scrolling.
		no_saved_settings (bool, optional): Never load/save settings in .ini file.
		id (Union[int, str], optional): (deprecated)
	Returns:
		Union[int, str]
	"""

	return internal_dpg.add_table(**kwargs)

def add_table_cell(**kwargs):
	"""	 Adds a table.

	Args:
		label (str, optional): Overrides 'name' as label.
		user_data (Any, optional): User data for callbacks
		use_internal_label (bool, optional): Use generated internal label instead of user specified (appends ### uuid).
		tag (Union[int, str], optional): Unique id used to programmatically refer to the item.If label is unused this will be the label.
		height (int, optional): Height of the item.
		parent (Union[int, str], optional): Parent to add this item to. (runtime adding)
		before (Union[int, str], optional): This item will be displayed before the specified item in the parent.
		show (bool, optional): Attempt to render widget.
		filter_key (str, optional): Used by filter widget.
		id (Union[int, str], optional): (deprecated)
	Returns:
		Union[int, str]
	"""

	return internal_dpg.add_table_cell(**kwargs)

def add_table_column(**kwargs):
	"""	 Adds a table column.

	Args:
		label (str, optional): Overrides 'name' as label.
		user_data (Any, optional): User data for callbacks
		use_internal_label (bool, optional): Use generated internal label instead of user specified (appends ### uuid).
		tag (Union[int, str], optional): Unique id used to programmatically refer to the item.If label is unused this will be the label.
		width (int, optional): Width of the item.
		parent (Union[int, str], optional): Parent to add this item to. (runtime adding)
		before (Union[int, str], optional): This item will be displayed before the specified item in the parent.
		show (bool, optional): Attempt to render widget.
		enabled (bool, optional): Turns off functionality of widget and applies the disabled theme.
		init_width_or_weight (float, optional): 
		default_hide (bool, optional): Default as a hidden/disabled column.
		default_sort (bool, optional): Default as a sorting column.
		width_stretch (bool, optional): Column will stretch. Preferable with horizontal scrolling disabled (default if table sizing policy is _SizingStretchSame or _SizingStretchProp).
		width_fixed (bool, optional): Column will not stretch. Preferable with horizontal scrolling enabled (default if table sizing policy is _SizingFixedFit and table is resizable).
		no_resize (bool, optional): Disable manual resizing.
		no_reorder (bool, optional): Disable manual reordering this column, this will also prevent other columns from crossing over this column.
		no_hide (bool, optional): Disable ability to hide/disable this column.
		no_clip (bool, optional): Disable clipping for this column (all NoClip columns will render in a same draw command).
		no_sort (bool, optional): Disable ability to sort on this field (even if ImGuiTableFlags_Sortable is set on the table).
		no_sort_ascending (bool, optional): Disable ability to sort in the ascending direction.
		no_sort_descending (bool, optional): Disable ability to sort in the descending direction.
		no_header_width (bool, optional): Disable header text width contribution to automatic column width.
		prefer_sort_ascending (bool, optional): Make the initial sort direction Ascending when first sorting on this column (default).
		prefer_sort_descending (bool, optional): Make the initial sort direction Descending when first sorting on this column.
		indent_enable (bool, optional): Use current Indent value when entering cell (default for column 0).
		indent_disable (bool, optional): Ignore current Indent value when entering cell (default for columns > 0). Indentation changes _within_ the cell will still be honored.
		id (Union[int, str], optional): (deprecated)
	Returns:
		Union[int, str]
	"""

	return internal_dpg.add_table_column(**kwargs)

def add_table_row(**kwargs):
	"""	 Adds a table row.

	Args:
		label (str, optional): Overrides 'name' as label.
		user_data (Any, optional): User data for callbacks
		use_internal_label (bool, optional): Use generated internal label instead of user specified (appends ### uuid).
		tag (Union[int, str], optional): Unique id used to programmatically refer to the item.If label is unused this will be the label.
		height (int, optional): Height of the item.
		parent (Union[int, str], optional): Parent to add this item to. (runtime adding)
		before (Union[int, str], optional): This item will be displayed before the specified item in the parent.
		show (bool, optional): Attempt to render widget.
		filter_key (str, optional): Used by filter widget.
		id (Union[int, str], optional): (deprecated)
	Returns:
		Union[int, str]
	"""

	return internal_dpg.add_table_row(**kwargs)

def add_template_registry(**kwargs):
	"""	 Adds a template registry.

	Args:
		label (str, optional): Overrides 'name' as label.
		user_data (Any, optional): User data for callbacks
		use_internal_label (bool, optional): Use generated internal label instead of user specified (appends ### uuid).
		tag (Union[int, str], optional): Unique id used to programmatically refer to the item.If label is unused this will be the label.
		id (Union[int, str], optional): (deprecated)
	Returns:
		Union[int, str]
	"""

	return internal_dpg.add_template_registry(**kwargs)

def add_text(default_value='', **kwargs):
	"""	 Adds text. Text can have an optional label that will display to the right of the text.

	Args:
		default_value (str, optional): 
		label (str, optional): Overrides 'name' as label.
		user_data (Any, optional): User data for callbacks
		use_internal_label (bool, optional): Use generated internal label instead of user specified (appends ### uuid).
		tag (Union[int, str], optional): Unique id used to programmatically refer to the item.If label is unused this will be the label.
		indent (int, optional): Offsets the widget to the right the specified number multiplied by the indent style.
		parent (Union[int, str], optional): Parent to add this item to. (runtime adding)
		before (Union[int, str], optional): This item will be displayed before the specified item in the parent.
		source (Union[int, str], optional): Overrides 'id' as value storage key.
		payload_type (str, optional): Sender string type must be the same as the target for the target to run the payload_callback.
		drag_callback (Callable, optional): Registers a drag callback for drag and drop.
		drop_callback (Callable, optional): Registers a drop callback for drag and drop.
		show (bool, optional): Attempt to render widget.
		pos (Union[List[int], Tuple[int, ...]], optional): Places the item relative to window coordinates, [0,0] is top left.
		filter_key (str, optional): Used by filter widget.
		tracked (bool, optional): Scroll tracking
		track_offset (float, optional): 0.0f:top, 0.5f:center, 1.0f:bottom
		wrap (int, optional): Number of pixels from the start of the item until wrapping starts.
		bullet (bool, optional): Places a bullet to the left of the text.
		color (Union[List[int], Tuple[int, ...]], optional): Color of the text (rgba).
		show_label (bool, optional): Displays the label to the right of the text.
		id (Union[int, str], optional): (deprecated)
	Returns:
		Union[int, str]
	"""

	return internal_dpg.add_text(default_value, **kwargs)

def add_text_point(x, y, **kwargs):
	"""	 Adds a label series to a plot.

	Args:
		x (float): 
		y (float): 
		label (str, optional): Overrides 'name' as label.
		user_data (Any, optional): User data for callbacks
		use_internal_label (bool, optional): Use generated internal label instead of user specified (appends ### uuid).
		tag (Union[int, str], optional): Unique id used to programmatically refer to the item.If label is unused this will be the label.
		parent (Union[int, str], optional): Parent to add this item to. (runtime adding)
		before (Union[int, str], optional): This item will be displayed before the specified item in the parent.
		source (Union[int, str], optional): Overrides 'id' as value storage key.
		show (bool, optional): Attempt to render widget.
		x_offset (int, optional): 
		y_offset (int, optional): 
		vertical (bool, optional): 
		id (Union[int, str], optional): (deprecated)
	Returns:
		Union[int, str]
	"""

	return internal_dpg.add_text_point(x, y, **kwargs)

def add_texture_registry(**kwargs):
	"""	 Adds a dynamic texture.

	Args:
		label (str, optional): Overrides 'name' as label.
		user_data (Any, optional): User data for callbacks
		use_internal_label (bool, optional): Use generated internal label instead of user specified (appends ### uuid).
		tag (Union[int, str], optional): Unique id used to programmatically refer to the item.If label is unused this will be the label.
		show (bool, optional): Attempt to render widget.
		id (Union[int, str], optional): (deprecated)
	Returns:
		Union[int, str]
	"""

	return internal_dpg.add_texture_registry(**kwargs)

def add_theme(**kwargs):
	"""	 Adds a theme.

	Args:
		label (str, optional): Overrides 'name' as label.
		user_data (Any, optional): User data for callbacks
		use_internal_label (bool, optional): Use generated internal label instead of user specified (appends ### uuid).
		tag (Union[int, str], optional): Unique id used to programmatically refer to the item.If label is unused this will be the label.
		id (Union[int, str], optional): (deprecated)
		default_theme (bool, optional): (deprecated)
	Returns:
		Union[int, str]
	"""

	return internal_dpg.add_theme(**kwargs)

def add_theme_color(target=0, value=(0, 0, 0, 255), **kwargs):
	"""	 Adds a theme color.

	Args:
		target (int, optional): 
		value (Union[List[int], Tuple[int, ...]], optional): 
		label (str, optional): Overrides 'name' as label.
		user_data (Any, optional): User data for callbacks
		use_internal_label (bool, optional): Use generated internal label instead of user specified (appends ### uuid).
		tag (Union[int, str], optional): Unique id used to programmatically refer to the item.If label is unused this will be the label.
		parent (Union[int, str], optional): Parent to add this item to. (runtime adding)
		category (int, optional): Options include mvThemeCat_Core, mvThemeCat_Plots, mvThemeCat_Nodes.
		id (Union[int, str], optional): (deprecated)
	Returns:
		Union[int, str]
	"""

	return internal_dpg.add_theme_color(target, value, **kwargs)

def add_theme_component(item_type=0, **kwargs):
	"""	 Adds a theme component.

	Args:
		item_type (int, optional): 
		label (str, optional): Overrides 'name' as label.
		user_data (Any, optional): User data for callbacks
		use_internal_label (bool, optional): Use generated internal label instead of user specified (appends ### uuid).
		tag (Union[int, str], optional): Unique id used to programmatically refer to the item.If label is unused this will be the label.
		parent (Union[int, str], optional): Parent to add this item to. (runtime adding)
		before (Union[int, str], optional): This item will be displayed before the specified item in the parent.
		enabled_state (bool, optional): 
		id (Union[int, str], optional): (deprecated)
	Returns:
		Union[int, str]
	"""

	return internal_dpg.add_theme_component(item_type, **kwargs)

def add_theme_style(target=0, x=1.0, y=-1.0, **kwargs):
	"""	 Adds a theme style.

	Args:
		target (int, optional): 
		x (float, optional): 
		y (float, optional): 
		label (str, optional): Overrides 'name' as label.
		user_data (Any, optional): User data for callbacks
		use_internal_label (bool, optional): Use generated internal label instead of user specified (appends ### uuid).
		tag (Union[int, str], optional): Unique id used to programmatically refer to the item.If label is unused this will be the label.
		parent (Union[int, str], optional): Parent to add this item to. (runtime adding)
		category (int, optional): Options include mvThemeCat_Core, mvThemeCat_Plots, mvThemeCat_Nodes.
		id (Union[int, str], optional): (deprecated)
	Returns:
		Union[int, str]
	"""

	return internal_dpg.add_theme_style(target, x, y, **kwargs)

def add_time_picker(**kwargs):
	"""	 Adds a time picker.

	Args:
		label (str, optional): Overrides 'name' as label.
		user_data (Any, optional): User data for callbacks
		use_internal_label (bool, optional): Use generated internal label instead of user specified (appends ### uuid).
		tag (Union[int, str], optional): Unique id used to programmatically refer to the item.If label is unused this will be the label.
		indent (int, optional): Offsets the widget to the right the specified number multiplied by the indent style.
		parent (Union[int, str], optional): Parent to add this item to. (runtime adding)
		before (Union[int, str], optional): This item will be displayed before the specified item in the parent.
		payload_type (str, optional): Sender string type must be the same as the target for the target to run the payload_callback.
		callback (Callable, optional): Registers a callback.
		drag_callback (Callable, optional): Registers a drag callback for drag and drop.
		drop_callback (Callable, optional): Registers a drop callback for drag and drop.
		show (bool, optional): Attempt to render widget.
		pos (Union[List[int], Tuple[int, ...]], optional): Places the item relative to window coordinates, [0,0] is top left.
		filter_key (str, optional): Used by filter widget.
		tracked (bool, optional): Scroll tracking
		track_offset (float, optional): 0.0f:top, 0.5f:center, 1.0f:bottom
		default_value (dict, optional): 
		hour24 (bool, optional): Show 24 hour clock instead of 12 hour.
		id (Union[int, str], optional): (deprecated)
	Returns:
		Union[int, str]
	"""

	return internal_dpg.add_time_picker(**kwargs)

def add_tooltip(parent, **kwargs):
	"""	 Adds a tooltip window.

	Args:
		parent (Union[int, str]): 
		label (str, optional): Overrides 'name' as label.
		user_data (Any, optional): User data for callbacks
		use_internal_label (bool, optional): Use generated internal label instead of user specified (appends ### uuid).
		tag (Union[int, str], optional): Unique id used to programmatically refer to the item.If label is unused this will be the label.
		show (bool, optional): Attempt to render widget.
		id (Union[int, str], optional): (deprecated)
	Returns:
		Union[int, str]
	"""

	return internal_dpg.add_tooltip(parent, **kwargs)

def add_tree_node(**kwargs):
	"""	 Adds a tree node to add items to.

	Args:
		label (str, optional): Overrides 'name' as label.
		user_data (Any, optional): User data for callbacks
		use_internal_label (bool, optional): Use generated internal label instead of user specified (appends ### uuid).
		tag (Union[int, str], optional): Unique id used to programmatically refer to the item.If label is unused this will be the label.
		indent (int, optional): Offsets the widget to the right the specified number multiplied by the indent style.
		parent (Union[int, str], optional): Parent to add this item to. (runtime adding)
		before (Union[int, str], optional): This item will be displayed before the specified item in the parent.
		payload_type (str, optional): Sender string type must be the same as the target for the target to run the payload_callback.
		drag_callback (Callable, optional): Registers a drag callback for drag and drop.
		drop_callback (Callable, optional): Registers a drop callback for drag and drop.
		show (bool, optional): Attempt to render widget.
		pos (Union[List[int], Tuple[int, ...]], optional): Places the item relative to window coordinates, [0,0] is top left.
		filter_key (str, optional): Used by filter widget.
		delay_search (bool, optional): Delays searching container for specified items until the end of the app. Possible optimization when a container has many children that are not accessed often.
		tracked (bool, optional): Scroll tracking
		track_offset (float, optional): 0.0f:top, 0.5f:center, 1.0f:bottom
		default_open (bool, optional): Sets the tree node open by default.
		open_on_double_click (bool, optional): Need double-click to open node.
		open_on_arrow (bool, optional): Only open when clicking on the arrow part.
		leaf (bool, optional): No collapsing, no arrow (use as a convenience for leaf nodes).
		bullet (bool, optional): Display a bullet instead of arrow.
		selectable (bool, optional): Makes the tree selectable.
		id (Union[int, str], optional): (deprecated)
	Returns:
		Union[int, str]
	"""

	return internal_dpg.add_tree_node(**kwargs)

def add_value_registry(**kwargs):
	"""	 Adds a value registry.

	Args:
		label (str, optional): Overrides 'name' as label.
		user_data (Any, optional): User data for callbacks
		use_internal_label (bool, optional): Use generated internal label instead of user specified (appends ### uuid).
		tag (Union[int, str], optional): Unique id used to programmatically refer to the item.If label is unused this will be the label.
		id (Union[int, str], optional): (deprecated)
	Returns:
		Union[int, str]
	"""

	return internal_dpg.add_value_registry(**kwargs)

def add_viewport_drawlist(**kwargs):
	"""	 A container that is used to present draw items or layers directly to the viewport. By default this will draw to the back of the viewport. Layers and draw items should be added to this widget as children.

	Args:
		label (str, optional): Overrides 'name' as label.
		user_data (Any, optional): User data for callbacks
		use_internal_label (bool, optional): Use generated internal label instead of user specified (appends ### uuid).
		tag (Union[int, str], optional): Unique id used to programmatically refer to the item.If label is unused this will be the label.
		show (bool, optional): Attempt to render widget.
		filter_key (str, optional): Used by filter widget.
		delay_search (bool, optional): Delays searching container for specified items until the end of the app. Possible optimization when a container has many children that are not accessed often.
		front (bool, optional): Draws to the front of the view port instead of the back.
		id (Union[int, str], optional): (deprecated)
	Returns:
		Union[int, str]
	"""

	return internal_dpg.add_viewport_drawlist(**kwargs)

def add_viewport_menu_bar(**kwargs):
	"""	 Adds a menubar to the viewport.

	Args:
		label (str, optional): Overrides 'name' as label.
		user_data (Any, optional): User data for callbacks
		use_internal_label (bool, optional): Use generated internal label instead of user specified (appends ### uuid).
		tag (Union[int, str], optional): Unique id used to programmatically refer to the item.If label is unused this will be the label.
		indent (int, optional): Offsets the widget to the right the specified number multiplied by the indent style.
		parent (Union[int, str], optional): Parent to add this item to. (runtime adding)
		show (bool, optional): Attempt to render widget.
		delay_search (bool, optional): Delays searching container for specified items until the end of the app. Possible optimization when a container has many children that are not accessed often.
		id (Union[int, str], optional): (deprecated)
	Returns:
		Union[int, str]
	"""

	return internal_dpg.add_viewport_menu_bar(**kwargs)

def add_vline_series(x, **kwargs):
	"""	 Adds an infinite vertical line series to a plot.

	Args:
		x (Any): 
		label (str, optional): Overrides 'name' as label.
		user_data (Any, optional): User data for callbacks
		use_internal_label (bool, optional): Use generated internal label instead of user specified (appends ### uuid).
		tag (Union[int, str], optional): Unique id used to programmatically refer to the item.If label is unused this will be the label.
		parent (Union[int, str], optional): Parent to add this item to. (runtime adding)
		before (Union[int, str], optional): This item will be displayed before the specified item in the parent.
		source (Union[int, str], optional): Overrides 'id' as value storage key.
		show (bool, optional): Attempt to render widget.
		id (Union[int, str], optional): (deprecated)
	Returns:
		Union[int, str]
	"""

	return internal_dpg.add_vline_series(x, **kwargs)

def add_window(**kwargs):
	"""	 Creates a new window for following items to be added to.

	Args:
		label (str, optional): Overrides 'name' as label.
		user_data (Any, optional): User data for callbacks
		use_internal_label (bool, optional): Use generated internal label instead of user specified (appends ### uuid).
		tag (Union[int, str], optional): Unique id used to programmatically refer to the item.If label is unused this will be the label.
		width (int, optional): Width of the item.
		height (int, optional): Height of the item.
		indent (int, optional): Offsets the widget to the right the specified number multiplied by the indent style.
		show (bool, optional): Attempt to render widget.
		pos (Union[List[int], Tuple[int, ...]], optional): Places the item relative to window coordinates, [0,0] is top left.
		delay_search (bool, optional): Delays searching container for specified items until the end of the app. Possible optimization when a container has many children that are not accessed often.
		min_size (Union[List[int], Tuple[int, ...]], optional): Minimum window size.
		max_size (Union[List[int], Tuple[int, ...]], optional): Maximum window size.
		menubar (bool, optional): Shows or hides the menubar.
		collapsed (bool, optional): Collapse the window.
		autosize (bool, optional): Autosized the window to fit it's items.
		no_resize (bool, optional): Allows for the window size to be changed or fixed.
		no_title_bar (bool, optional): Title name for the title bar of the window.
		no_move (bool, optional): Allows for the window's position to be changed or fixed.
		no_scrollbar (bool, optional):  Disable scrollbars. (window can still scroll with mouse or programmatically)
		no_collapse (bool, optional): Disable user collapsing window by double-clicking on it.
		horizontal_scrollbar (bool, optional): Allow horizontal scrollbar to appear. (off by default)
		no_focus_on_appearing (bool, optional): Disable taking focus when transitioning from hidden to visible state.
		no_bring_to_front_on_focus (bool, optional): Disable bringing window to front when taking focus. (e.g. clicking on it or programmatically giving it focus)
		no_close (bool, optional): Disable user closing the window by removing the close button.
		no_background (bool, optional): Sets Background and border alpha to transparent.
		modal (bool, optional): Fills area behind window according to the theme and disables user ability to interact with anything except the window.
		popup (bool, optional): Fills area behind window according to the theme, removes title bar, collapse and close. Window can be closed by selecting area in the background behind the window.
		no_saved_settings (bool, optional): Never load/save settings in .ini file.
		no_open_over_existing_popup (bool, optional): Don't open if there's already a popup
		on_close (Callable, optional): Callback ran when window is closed.
		id (Union[int, str], optional): (deprecated)
	Returns:
		Union[int, str]
	"""

	return internal_dpg.add_window(**kwargs)

def apply_transform(item, transform):
	"""	 New in 1.1. Applies a transformation matrix to a layer.

	Args:
		item (Union[int, str]): Drawing node to apply transform to.
		transform (Any): Transformation matrix.
	Returns:
		None
	"""

	return internal_dpg.apply_transform(item, transform)

def bind_colormap(item, source):
	"""	 Sets the color map for widgets that accept it.

	Args:
		item (Union[int, str]): item that the color map will be applied to
		source (Union[int, str]): The colormap tag. This should come from a colormap that was added to a colormap registry.  Built in color maps are accessible through their corresponding constants mvPlotColormap_Twilight, mvPlotColormap_***
	Returns:
		None
	"""

	return internal_dpg.bind_colormap(item, source)

def bind_font(font):
	"""	 Binds a global font.

	Args:
		font (Union[int, str]): 
	Returns:
		Union[int, str]
	"""

	return internal_dpg.bind_font(font)

def bind_item_font(item, font):
	"""	 Sets an item's font.

	Args:
		item (Union[int, str]): 
		font (Union[int, str]): 
	Returns:
		None
	"""

	return internal_dpg.bind_item_font(item, font)

def bind_item_handler_registry(item, handler_registry):
	"""	 Binds an item handler registry to an item.

	Args:
		item (Union[int, str]): 
		handler_registry (Union[int, str]): 
	Returns:
		None
	"""

	return internal_dpg.bind_item_handler_registry(item, handler_registry)

def bind_item_theme(item, theme):
	"""	 Binds a theme to an item.

	Args:
		item (Union[int, str]): 
		theme (Union[int, str]): 
	Returns:
		None
	"""

	return internal_dpg.bind_item_theme(item, theme)

def bind_template_registry(template_registry):
	"""	 Binds a global template registry.

	Args:
		template_registry (Union[int, str]): 
	Returns:
		None
	"""

	return internal_dpg.bind_template_registry(template_registry)

def bind_theme(theme):
	"""	 Binds a global theme.

	Args:
		theme (Union[int, str]): 
	Returns:
		None
	"""

	return internal_dpg.bind_theme(theme)

def capture_next_item(callback, **kwargs):
	"""	 Captures the next item.

	Args:
		callback (Callable): 
		user_data (Any, optional): New in 1.3. Optional user data to send to the callback
	Returns:
		None
	"""

	return internal_dpg.capture_next_item(callback, **kwargs)

def clear_selected_links(node_editor):
	"""	 Clears a node editor's selected links.

	Args:
		node_editor (Union[int, str]): 
	Returns:
		None
	"""

	return internal_dpg.clear_selected_links(node_editor)

def clear_selected_nodes(node_editor):
	"""	 Clears a node editor's selected nodes.

	Args:
		node_editor (Union[int, str]): 
	Returns:
		None
	"""

	return internal_dpg.clear_selected_nodes(node_editor)

def create_context():
	"""	 Creates the Dear PyGui context.

	Args:
	Returns:
		None
	"""

	return internal_dpg.create_context()

def create_fps_matrix(eye, pitch, yaw):
	"""	 New in 1.1. Create a 'first person shooter' matrix.

	Args:
		eye (Union[List[float], Tuple[float, ...]]): eye position
		pitch (float): pitch (in radians)
		yaw (float): yaw (in radians)
	Returns:
		Any
	"""

	return internal_dpg.create_fps_matrix(eye, pitch, yaw)

def create_lookat_matrix(eye, target, up):
	"""	 New in 1.1. Creates a 'Look at matrix'.

	Args:
		eye (Union[List[float], Tuple[float, ...]]): eye position
		target (Union[List[float], Tuple[float, ...]]): target position
		up (Union[List[float], Tuple[float, ...]]): up vector
	Returns:
		Any
	"""

	return internal_dpg.create_lookat_matrix(eye, target, up)

def create_orthographic_matrix(left, right, bottom, top, zNear, zFar):
	"""	 New in 1.1. Creates an orthographic matrix.

	Args:
		left (float): left plane
		right (float): right plane
		bottom (float): bottom plane
		top (float): top plane
		zNear (float): Near clipping plane.
		zFar (float): Far clipping plane.
	Returns:
		Any
	"""

	return internal_dpg.create_orthographic_matrix(left, right, bottom, top, zNear, zFar)

def create_perspective_matrix(fov, aspect, zNear, zFar):
	"""	 New in 1.1. Creates a perspective matrix.

	Args:
		fov (float): Field of view (in radians)
		aspect (float): Aspect ratio (width/height)
		zNear (float): Near clipping plane.
		zFar (float): Far clipping plane.
	Returns:
		Any
	"""

	return internal_dpg.create_perspective_matrix(fov, aspect, zNear, zFar)

def create_rotation_matrix(angle, axis):
	"""	 New in 1.1. Applies a transformation matrix to a layer.

	Args:
		angle (float): angle to rotate
		axis (Union[List[float], Tuple[float, ...]]): axis to rotate around
	Returns:
		Any
	"""

	return internal_dpg.create_rotation_matrix(angle, axis)

def create_scale_matrix(scales):
	"""	 New in 1.1. Applies a transformation matrix to a layer.

	Args:
		scales (Union[List[float], Tuple[float, ...]]): scale values per axis
	Returns:
		Any
	"""

	return internal_dpg.create_scale_matrix(scales)

def create_translation_matrix(translation):
	"""	 New in 1.1. Creates a translation matrix.

	Args:
		translation (Union[List[float], Tuple[float, ...]]): translation vector
	Returns:
		Any
	"""

	return internal_dpg.create_translation_matrix(translation)

def create_viewport(**kwargs):
	"""	 Creates a viewport. Viewports are required.

	Args:
		title (str, optional): Sets the title of the viewport.
		small_icon (str, optional): Sets the small icon that is found in the viewport's decorator bar. Must be ***.ico on windows and either ***.ico or ***.png on mac.
		large_icon (str, optional): Sets the large icon that is found in the task bar while the app is running. Must be ***.ico on windows and either ***.ico or ***.png on mac.
		width (int, optional): Sets the width of the drawable space on the viewport. Does not inclue the border.
		height (int, optional): Sets the height of the drawable space on the viewport. Does not inclue the border or decorator bar.
		x_pos (int, optional): Sets x position the viewport will be drawn in screen coordinates.
		y_pos (int, optional): Sets y position the viewport will be drawn in screen coordinates.
		min_width (int, optional): Applies a minimuim limit to the width of the viewport.
		max_width (int, optional): Applies a maximum limit to the width of the viewport.
		min_height (int, optional): Applies a minimuim limit to the height of the viewport.
		max_height (int, optional): Applies a maximum limit to the height of the viewport.
		resizable (bool, optional): Enables and Disables user ability to resize the viewport.
		vsync (bool, optional): Enables and Disables the renderloop vsync limit. vsync frame value is set by refresh rate of display.
		always_on_top (bool, optional): Forces the viewport to always be drawn ontop of all other viewports.
		decorated (bool, optional): Enabled and disabled the decorator bar at the top of the viewport.
		clear_color (Union[List[float], Tuple[float, ...]], optional): Sets the color of the back of the viewport.
	Returns:
		None
	"""

	return internal_dpg.create_viewport(**kwargs)

def delete_item(item, **kwargs):
	"""	 Deletes an item..

	Args:
		item (Union[int, str]): 
		children_only (bool, optional): 
		slot (int, optional): 
	Returns:
		None
	"""

	return internal_dpg.delete_item(item, **kwargs)

def destroy_context():
	"""	 Destroys the Dear PyGui context.

	Args:
	Returns:
		None
	"""

	return internal_dpg.destroy_context()

def does_alias_exist(alias):
	"""	 Checks if an alias exist.

	Args:
		alias (str): 
	Returns:
		bool
	"""

	return internal_dpg.does_alias_exist(alias)

def does_item_exist(item):
	"""	 Checks if an item exist..

	Args:
		item (Union[int, str]): 
	Returns:
		bool
	"""

	return internal_dpg.does_item_exist(item)

def draw_arrow(p1, p2, **kwargs):
	"""	 Adds an arrow.

	Args:
		p1 (Union[List[float], Tuple[float, ...]]): Arrow tip.
		p2 (Union[List[float], Tuple[float, ...]]): Arrow tail.
		label (str, optional): Overrides 'name' as label.
		user_data (Any, optional): User data for callbacks
		use_internal_label (bool, optional): Use generated internal label instead of user specified (appends ### uuid).
		tag (Union[int, str], optional): Unique id used to programmatically refer to the item.If label is unused this will be the label.
		parent (Union[int, str], optional): Parent to add this item to. (runtime adding)
		before (Union[int, str], optional): This item will be displayed before the specified item in the parent.
		show (bool, optional): Attempt to render widget.
		color (Union[List[int], Tuple[int, ...]], optional): 
		thickness (float, optional): 
		size (int, optional): 
		id (Union[int, str], optional): (deprecated)
	Returns:
		Union[int, str]
	"""

	return internal_dpg.draw_arrow(p1, p2, **kwargs)

def draw_bezier_cubic(p1, p2, p3, p4, **kwargs):
	"""	 Adds a cubic bezier curve.

	Args:
		p1 (Union[List[float], Tuple[float, ...]]): First point in curve.
		p2 (Union[List[float], Tuple[float, ...]]): Second point in curve.
		p3 (Union[List[float], Tuple[float, ...]]): Third point in curve.
		p4 (Union[List[float], Tuple[float, ...]]): Fourth point in curve.
		label (str, optional): Overrides 'name' as label.
		user_data (Any, optional): User data for callbacks
		use_internal_label (bool, optional): Use generated internal label instead of user specified (appends ### uuid).
		tag (Union[int, str], optional): Unique id used to programmatically refer to the item.If label is unused this will be the label.
		parent (Union[int, str], optional): Parent to add this item to. (runtime adding)
		before (Union[int, str], optional): This item will be displayed before the specified item in the parent.
		show (bool, optional): Attempt to render widget.
		color (Union[List[int], Tuple[int, ...]], optional): 
		thickness (float, optional): 
		segments (int, optional): Number of segments to approximate bezier curve.
		id (Union[int, str], optional): (deprecated)
	Returns:
		Union[int, str]
	"""

	return internal_dpg.draw_bezier_cubic(p1, p2, p3, p4, **kwargs)

def draw_bezier_quadratic(p1, p2, p3, **kwargs):
	"""	 Adds a quadratic bezier curve.

	Args:
		p1 (Union[List[float], Tuple[float, ...]]): First point in curve.
		p2 (Union[List[float], Tuple[float, ...]]): Second point in curve.
		p3 (Union[List[float], Tuple[float, ...]]): Third point in curve.
		label (str, optional): Overrides 'name' as label.
		user_data (Any, optional): User data for callbacks
		use_internal_label (bool, optional): Use generated internal label instead of user specified (appends ### uuid).
		tag (Union[int, str], optional): Unique id used to programmatically refer to the item.If label is unused this will be the label.
		parent (Union[int, str], optional): Parent to add this item to. (runtime adding)
		before (Union[int, str], optional): This item will be displayed before the specified item in the parent.
		show (bool, optional): Attempt to render widget.
		color (Union[List[int], Tuple[int, ...]], optional): 
		thickness (float, optional): 
		segments (int, optional): Number of segments to approximate bezier curve.
		id (Union[int, str], optional): (deprecated)
	Returns:
		Union[int, str]
	"""

	return internal_dpg.draw_bezier_quadratic(p1, p2, p3, **kwargs)

def draw_circle(center, radius, **kwargs):
	"""	 Adds a circle

	Args:
		center (Union[List[float], Tuple[float, ...]]): 
		radius (float): 
		label (str, optional): Overrides 'name' as label.
		user_data (Any, optional): User data for callbacks
		use_internal_label (bool, optional): Use generated internal label instead of user specified (appends ### uuid).
		tag (Union[int, str], optional): Unique id used to programmatically refer to the item.If label is unused this will be the label.
		parent (Union[int, str], optional): Parent to add this item to. (runtime adding)
		before (Union[int, str], optional): This item will be displayed before the specified item in the parent.
		show (bool, optional): Attempt to render widget.
		color (Union[List[int], Tuple[int, ...]], optional): 
		fill (Union[List[int], Tuple[int, ...]], optional): 
		thickness (float, optional): 
		segments (int, optional): Number of segments to approximate circle.
		id (Union[int, str], optional): (deprecated)
	Returns:
		Union[int, str]
	"""

	return internal_dpg.draw_circle(center, radius, **kwargs)

def draw_ellipse(pmin, pmax, **kwargs):
	"""	 Adds an ellipse.

	Args:
		pmin (Union[List[float], Tuple[float, ...]]): Min point of bounding rectangle.
		pmax (Union[List[float], Tuple[float, ...]]): Max point of bounding rectangle.
		label (str, optional): Overrides 'name' as label.
		user_data (Any, optional): User data for callbacks
		use_internal_label (bool, optional): Use generated internal label instead of user specified (appends ### uuid).
		tag (Union[int, str], optional): Unique id used to programmatically refer to the item.If label is unused this will be the label.
		parent (Union[int, str], optional): Parent to add this item to. (runtime adding)
		before (Union[int, str], optional): This item will be displayed before the specified item in the parent.
		show (bool, optional): Attempt to render widget.
		color (Union[List[int], Tuple[int, ...]], optional): 
		fill (Union[List[int], Tuple[int, ...]], optional): 
		thickness (float, optional): 
		segments (int, optional): Number of segments to approximate bezier curve.
		id (Union[int, str], optional): (deprecated)
	Returns:
		Union[int, str]
	"""

	return internal_dpg.draw_ellipse(pmin, pmax, **kwargs)

def draw_image(texture_tag, pmin, pmax, **kwargs):
	"""	 Adds an image (for a drawing).

	Args:
		texture_tag (Union[int, str]): 
		pmin (Union[List[float], Tuple[float, ...]]): Point of to start drawing texture.
		pmax (Union[List[float], Tuple[float, ...]]): Point to complete drawing texture.
		label (str, optional): Overrides 'name' as label.
		user_data (Any, optional): User data for callbacks
		use_internal_label (bool, optional): Use generated internal label instead of user specified (appends ### uuid).
		tag (Union[int, str], optional): Unique id used to programmatically refer to the item.If label is unused this will be the label.
		parent (Union[int, str], optional): Parent to add this item to. (runtime adding)
		before (Union[int, str], optional): This item will be displayed before the specified item in the parent.
		show (bool, optional): Attempt to render widget.
		uv_min (Union[List[float], Tuple[float, ...]], optional): Normalized coordinates on texture that will be drawn.
		uv_max (Union[List[float], Tuple[float, ...]], optional): Normalized coordinates on texture that will be drawn.
		color (Union[List[int], Tuple[int, ...]], optional): 
		id (Union[int, str], optional): (deprecated)
	Returns:
		Union[int, str]
	"""

	return internal_dpg.draw_image(texture_tag, pmin, pmax, **kwargs)

def draw_image_quad(texture_tag, p1, p2, p3, p4, **kwargs):
	"""	 Adds an image (for a drawing).

	Args:
		texture_tag (Union[int, str]): 
		p1 (Union[List[float], Tuple[float, ...]]): 
		p2 (Union[List[float], Tuple[float, ...]]): 
		p3 (Union[List[float], Tuple[float, ...]]): 
		p4 (Union[List[float], Tuple[float, ...]]): 
		label (str, optional): Overrides 'name' as label.
		user_data (Any, optional): User data for callbacks
		use_internal_label (bool, optional): Use generated internal label instead of user specified (appends ### uuid).
		tag (Union[int, str], optional): Unique id used to programmatically refer to the item.If label is unused this will be the label.
		parent (Union[int, str], optional): Parent to add this item to. (runtime adding)
		before (Union[int, str], optional): This item will be displayed before the specified item in the parent.
		show (bool, optional): Attempt to render widget.
		uv1 (Union[List[float], Tuple[float, ...]], optional): Normalized coordinates on texture that will be drawn.
		uv2 (Union[List[float], Tuple[float, ...]], optional): Normalized coordinates on texture that will be drawn.
		uv3 (Union[List[float], Tuple[float, ...]], optional): Normalized coordinates on texture that will be drawn.
		uv4 (Union[List[float], Tuple[float, ...]], optional): Normalized coordinates on texture that will be drawn.
		color (Union[List[int], Tuple[int, ...]], optional): 
		id (Union[int, str], optional): (deprecated)
	Returns:
		Union[int, str]
	"""

	return internal_dpg.draw_image_quad(texture_tag, p1, p2, p3, p4, **kwargs)

def draw_line(p1, p2, **kwargs):
	"""	 Adds a line.

	Args:
		p1 (Union[List[float], Tuple[float, ...]]): Start of line.
		p2 (Union[List[float], Tuple[float, ...]]): End of line.
		label (str, optional): Overrides 'name' as label.
		user_data (Any, optional): User data for callbacks
		use_internal_label (bool, optional): Use generated internal label instead of user specified (appends ### uuid).
		tag (Union[int, str], optional): Unique id used to programmatically refer to the item.If label is unused this will be the label.
		parent (Union[int, str], optional): Parent to add this item to. (runtime adding)
		before (Union[int, str], optional): This item will be displayed before the specified item in the parent.
		show (bool, optional): Attempt to render widget.
		color (Union[List[int], Tuple[int, ...]], optional): 
		thickness (float, optional): 
		id (Union[int, str], optional): (deprecated)
	Returns:
		Union[int, str]
	"""

	return internal_dpg.draw_line(p1, p2, **kwargs)

def draw_polygon(points, **kwargs):
	"""	 Adds a polygon.

	Args:
		points (List[List[float]]): 
		label (str, optional): Overrides 'name' as label.
		user_data (Any, optional): User data for callbacks
		use_internal_label (bool, optional): Use generated internal label instead of user specified (appends ### uuid).
		tag (Union[int, str], optional): Unique id used to programmatically refer to the item.If label is unused this will be the label.
		parent (Union[int, str], optional): Parent to add this item to. (runtime adding)
		before (Union[int, str], optional): This item will be displayed before the specified item in the parent.
		show (bool, optional): Attempt to render widget.
		color (Union[List[int], Tuple[int, ...]], optional): 
		fill (Union[List[int], Tuple[int, ...]], optional): 
		thickness (float, optional): 
		id (Union[int, str], optional): (deprecated)
	Returns:
		Union[int, str]
	"""

	return internal_dpg.draw_polygon(points, **kwargs)

def draw_polyline(points, **kwargs):
	"""	 Adds a polyline.

	Args:
		points (List[List[float]]): 
		label (str, optional): Overrides 'name' as label.
		user_data (Any, optional): User data for callbacks
		use_internal_label (bool, optional): Use generated internal label instead of user specified (appends ### uuid).
		tag (Union[int, str], optional): Unique id used to programmatically refer to the item.If label is unused this will be the label.
		parent (Union[int, str], optional): Parent to add this item to. (runtime adding)
		before (Union[int, str], optional): This item will be displayed before the specified item in the parent.
		show (bool, optional): Attempt to render widget.
		closed (bool, optional): Will close the polyline by returning to the first point.
		color (Union[List[int], Tuple[int, ...]], optional): 
		thickness (float, optional): 
		id (Union[int, str], optional): (deprecated)
	Returns:
		Union[int, str]
	"""

	return internal_dpg.draw_polyline(points, **kwargs)

def draw_quad(p1, p2, p3, p4, **kwargs):
	"""	 Adds a quad.

	Args:
		p1 (Union[List[float], Tuple[float, ...]]): 
		p2 (Union[List[float], Tuple[float, ...]]): 
		p3 (Union[List[float], Tuple[float, ...]]): 
		p4 (Union[List[float], Tuple[float, ...]]): 
		label (str, optional): Overrides 'name' as label.
		user_data (Any, optional): User data for callbacks
		use_internal_label (bool, optional): Use generated internal label instead of user specified (appends ### uuid).
		tag (Union[int, str], optional): Unique id used to programmatically refer to the item.If label is unused this will be the label.
		parent (Union[int, str], optional): Parent to add this item to. (runtime adding)
		before (Union[int, str], optional): This item will be displayed before the specified item in the parent.
		show (bool, optional): Attempt to render widget.
		color (Union[List[int], Tuple[int, ...]], optional): 
		fill (Union[List[int], Tuple[int, ...]], optional): 
		thickness (float, optional): 
		id (Union[int, str], optional): (deprecated)
	Returns:
		Union[int, str]
	"""

	return internal_dpg.draw_quad(p1, p2, p3, p4, **kwargs)

def draw_rectangle(pmin, pmax, **kwargs):
	"""	 Adds a rectangle.

	Args:
		pmin (Union[List[float], Tuple[float, ...]]): Min point of bounding rectangle.
		pmax (Union[List[float], Tuple[float, ...]]): Max point of bounding rectangle.
		label (str, optional): Overrides 'name' as label.
		user_data (Any, optional): User data for callbacks
		use_internal_label (bool, optional): Use generated internal label instead of user specified (appends ### uuid).
		tag (Union[int, str], optional): Unique id used to programmatically refer to the item.If label is unused this will be the label.
		parent (Union[int, str], optional): Parent to add this item to. (runtime adding)
		before (Union[int, str], optional): This item will be displayed before the specified item in the parent.
		show (bool, optional): Attempt to render widget.
		color (Union[List[int], Tuple[int, ...]], optional): 
		color_upper_left (Union[List[int], Tuple[int, ...]], optional): 'multicolor' must be set to 'True'
		color_upper_right (Union[List[int], Tuple[int, ...]], optional): 'multicolor' must be set to 'True'
		color_bottom_right (Union[List[int], Tuple[int, ...]], optional): 'multicolor' must be set to 'True'
		color_bottom_left (Union[List[int], Tuple[int, ...]], optional): 'multicolor' must be set to 'True'
		fill (Union[List[int], Tuple[int, ...]], optional): 
		multicolor (bool, optional): 
		rounding (float, optional): Number of pixels of the radius that will round the corners of the rectangle. Note: doesn't work with multicolor
		thickness (float, optional): 
		id (Union[int, str], optional): (deprecated)
	Returns:
		Union[int, str]
	"""

	return internal_dpg.draw_rectangle(pmin, pmax, **kwargs)

def draw_text(pos, text, **kwargs):
	"""	 Adds text (drawlist).

	Args:
		pos (Union[List[float], Tuple[float, ...]]): Top left point of bounding text rectangle.
		text (str): Text to draw.
		label (str, optional): Overrides 'name' as label.
		user_data (Any, optional): User data for callbacks
		use_internal_label (bool, optional): Use generated internal label instead of user specified (appends ### uuid).
		tag (Union[int, str], optional): Unique id used to programmatically refer to the item.If label is unused this will be the label.
		parent (Union[int, str], optional): Parent to add this item to. (runtime adding)
		before (Union[int, str], optional): This item will be displayed before the specified item in the parent.
		show (bool, optional): Attempt to render widget.
		color (Union[List[int], Tuple[int, ...]], optional): 
		size (float, optional): 
		id (Union[int, str], optional): (deprecated)
	Returns:
		Union[int, str]
	"""

	return internal_dpg.draw_text(pos, text, **kwargs)

def draw_triangle(p1, p2, p3, **kwargs):
	"""	 Adds a triangle.

	Args:
		p1 (Union[List[float], Tuple[float, ...]]): 
		p2 (Union[List[float], Tuple[float, ...]]): 
		p3 (Union[List[float], Tuple[float, ...]]): 
		label (str, optional): Overrides 'name' as label.
		user_data (Any, optional): User data for callbacks
		use_internal_label (bool, optional): Use generated internal label instead of user specified (appends ### uuid).
		tag (Union[int, str], optional): Unique id used to programmatically refer to the item.If label is unused this will be the label.
		parent (Union[int, str], optional): Parent to add this item to. (runtime adding)
		before (Union[int, str], optional): This item will be displayed before the specified item in the parent.
		show (bool, optional): Attempt to render widget.
		color (Union[List[int], Tuple[int, ...]], optional): 
		fill (Union[List[int], Tuple[int, ...]], optional): 
		thickness (float, optional): 
		id (Union[int, str], optional): (deprecated)
	Returns:
		Union[int, str]
	"""

	return internal_dpg.draw_triangle(p1, p2, p3, **kwargs)

def empty_container_stack():
	"""	 Emptyes the container stack.

	Args:
	Returns:
		None
	"""

	return internal_dpg.empty_container_stack()

def fit_axis_data(axis):
	"""	 Sets the axis boundaries max/min in the data series currently on the plot.

	Args:
		axis (Union[int, str]): 
	Returns:
		None
	"""

	return internal_dpg.fit_axis_data(axis)

def focus_item(item):
	"""	 Focuses an item.

	Args:
		item (Union[int, str]): 
	Returns:
		None
	"""

	return internal_dpg.focus_item(item)

def generate_uuid():
	"""	 Generate a new UUID.

	Args:
	Returns:
		Union[int, str]
	"""

	return internal_dpg.generate_uuid()

def get_active_window():
	"""	 Returns the active window.

	Args:
	Returns:
		Union[int, str]
	"""

	return internal_dpg.get_active_window()

def get_alias_id(alias):
	"""	 Returns the ID associated with an alias.

	Args:
		alias (str): 
	Returns:
		Union[int, str]
	"""

	return internal_dpg.get_alias_id(alias)

def get_aliases():
	"""	 Returns all aliases.

	Args:
	Returns:
		Union[List[str], Tuple[str, ...]]
	"""

	return internal_dpg.get_aliases()

def get_all_items():
	"""	 Returns all items.

	Args:
	Returns:
		Union[List[int], Tuple[int, ...]]
	"""

	return internal_dpg.get_all_items()

def get_app_configuration():
	"""	 Returns app configuration.

	Args:
	Returns:
		dict
	"""

	return internal_dpg.get_app_configuration()

def get_axis_limits(axis):
	"""	 Get the specified axis limits.

	Args:
		axis (Union[int, str]): 
	Returns:
		Union[List[float], Tuple[float, ...]]
	"""

	return internal_dpg.get_axis_limits(axis)

def get_callback_queue():
	"""	 New in 1.2. Returns and clears callback queue.

	Args:
	Returns:
		Any
	"""

	return internal_dpg.get_callback_queue()

def get_clipboard_text():
	"""	 New in 1.3. Gets the clipboard text.

	Args:
	Returns:
		str
	"""

	return internal_dpg.get_clipboard_text()

def get_colormap_color(colormap, index):
	"""	 Returns a color from a colormap given an index >= 0. (ex. 0 will be the first color in the color list of the color map) Modulo will be performed against the number of items in the color list.

	Args:
		colormap (Union[int, str]): The colormap tag. This should come from a colormap that was added to a colormap registry. Built in color maps are accessible through their corresponding constants mvPlotColormap_Twilight, mvPlotColormap_***
		index (int): Desired position of the color in the colors list value of the colormap being quiered 
	Returns:
		Union[List[int], Tuple[int, ...]]
	"""

	return internal_dpg.get_colormap_color(colormap, index)

def get_delta_time():
	"""	 Returns time since last frame.

	Args:
	Returns:
		float
	"""

	return internal_dpg.get_delta_time()

def get_drawing_mouse_pos():
	"""	 Returns mouse position in drawing.

	Args:
	Returns:
		Union[List[int], Tuple[int, ...]]
	"""

	return internal_dpg.get_drawing_mouse_pos()

def get_file_dialog_info(file_dialog):
	"""	 Returns information related to the file dialog. Typically used while the file dialog is in use to query data about the state or info related to the file dialog.

	Args:
		file_dialog (Union[int, str]): 
	Returns:
		dict
	"""

	return internal_dpg.get_file_dialog_info(file_dialog)

def get_frame_count():
	"""	 Returns frame count.

	Args:
	Returns:
		int
	"""

	return internal_dpg.get_frame_count()

def get_frame_rate():
	"""	 Returns the average frame rate across 120 frames.

	Args:
	Returns:
		float
	"""

	return internal_dpg.get_frame_rate()

def get_global_font_scale():
	"""	 Returns global font scale.

	Args:
	Returns:
		float
	"""

	return internal_dpg.get_global_font_scale()

def get_item_alias(item):
	"""	 Returns an item's alias.

	Args:
		item (Union[int, str]): 
	Returns:
		str
	"""

	return internal_dpg.get_item_alias(item)

def get_item_configuration(item):
	"""	 Returns an item's configuration.

	Args:
		item (Union[int, str]): 
	Returns:
		dict
	"""

	return internal_dpg.get_item_configuration(item)

def get_item_info(item):
	"""	 Returns an item's information.

	Args:
		item (Union[int, str]): 
	Returns:
		dict
	"""

	return internal_dpg.get_item_info(item)

def get_item_state(item):
	"""	 Returns an item's state.

	Args:
		item (Union[int, str]): 
	Returns:
		dict
	"""

	return internal_dpg.get_item_state(item)

def get_item_types():
	"""	 Returns an item types.

	Args:
	Returns:
		dict
	"""

	return internal_dpg.get_item_types()

def get_mouse_drag_delta():
	"""	 Returns mouse drag delta.

	Args:
	Returns:
		float
	"""

	return internal_dpg.get_mouse_drag_delta()

def get_mouse_pos(**kwargs):
	"""	 Returns mouse position.

	Args:
		local (bool, optional): 
	Returns:
		Union[List[int], Tuple[int, ...]]
	"""

	return internal_dpg.get_mouse_pos(**kwargs)

def get_plot_mouse_pos():
	"""	 Returns mouse position in plot.

	Args:
	Returns:
		Union[List[int], Tuple[int, ...]]
	"""

	return internal_dpg.get_plot_mouse_pos()

def get_plot_query_area(plot):
	"""	 Returns the last/current query area of the plot. (Requires plot 'query' kwarg to be enabled)

	Args:
		plot (Union[int, str]): 
	Returns:
		Union[List[float], Tuple[float, ...]]
	"""

	return internal_dpg.get_plot_query_area(plot)

def get_selected_links(node_editor):
	"""	 Returns a node editor's selected links.

	Args:
		node_editor (Union[int, str]): 
	Returns:
		List[List[str]]
	"""

	return internal_dpg.get_selected_links(node_editor)

def get_selected_nodes(node_editor):
	"""	 Returns a node editor's selected nodes.

	Args:
		node_editor (Union[int, str]): 
	Returns:
		Union[List[int], Tuple[int, ...]]
	"""

	return internal_dpg.get_selected_nodes(node_editor)

def get_text_size(text, **kwargs):
	"""	 Returns width/height of text with specified font (must occur after 1st frame).

	Args:
		text (str): 
		wrap_width (float, optional): Wrap width to use (-1.0 turns wrap off).
		font (Union[int, str], optional): Font to use.
	Returns:
		Union[List[float], Tuple[float, ...]]
	"""

	return internal_dpg.get_text_size(text, **kwargs)

def get_total_time():
	"""	 Returns total time since Dear PyGui has started.

	Args:
	Returns:
		float
	"""

	return internal_dpg.get_total_time()

def get_value(item):
	"""	 Returns an item's value.

	Args:
		item (Union[int, str]): 
	Returns:
		Any
	"""

	return internal_dpg.get_value(item)

def get_values(items):
	"""	 Returns values of a list of items.

	Args:
		items (Union[List[int], Tuple[int, ...]]): 
	Returns:
		Any
	"""

	return internal_dpg.get_values(items)

def get_viewport_configuration(item):
	"""	 Returns a viewport's configuration.

	Args:
		item (Union[int, str]): 
	Returns:
		dict
	"""

	return internal_dpg.get_viewport_configuration(item)

def get_windows():
	"""	 Returns all windows.

	Args:
	Returns:
		Union[List[int], Tuple[int, ...]]
	"""

	return internal_dpg.get_windows()

def get_x_scroll(item):
	"""	 Undocumented

	Args:
		item (Union[int, str]): 
	Returns:
		float
	"""

	return internal_dpg.get_x_scroll(item)

def get_x_scroll_max(item):
	"""	 Undocumented

	Args:
		item (Union[int, str]): 
	Returns:
		float
	"""

	return internal_dpg.get_x_scroll_max(item)

def get_y_scroll(item):
	"""	 Undocumented

	Args:
		item (Union[int, str]): 
	Returns:
		float
	"""

	return internal_dpg.get_y_scroll(item)

def get_y_scroll_max(item):
	"""	 Undocumented

	Args:
		item (Union[int, str]): 
	Returns:
		float
	"""

	return internal_dpg.get_y_scroll_max(item)

def highlight_table_cell(table, row, column, color):
	"""	 Highlight specified table cell.

	Args:
		table (Union[int, str]): 
		row (int): 
		column (int): 
		color (Union[List[int], Tuple[int, ...]]): 
	Returns:
		None
	"""

	return internal_dpg.highlight_table_cell(table, row, column, color)

def highlight_table_column(table, column, color):
	"""	 Highlight specified table column.

	Args:
		table (Union[int, str]): 
		column (int): 
		color (Union[List[int], Tuple[int, ...]]): 
	Returns:
		None
	"""

	return internal_dpg.highlight_table_column(table, column, color)

def highlight_table_row(table, row, color):
	"""	 Highlight specified table row.

	Args:
		table (Union[int, str]): 
		row (int): 
		color (Union[List[int], Tuple[int, ...]]): 
	Returns:
		None
	"""

	return internal_dpg.highlight_table_row(table, row, color)

def is_dearpygui_running():
	"""	 Checks if Dear PyGui is running

	Args:
	Returns:
		bool
	"""

	return internal_dpg.is_dearpygui_running()

def is_key_down(key):
	"""	 Checks if key is down.

	Args:
		key (int): 
	Returns:
		bool
	"""

	return internal_dpg.is_key_down(key)

def is_key_pressed(key):
	"""	 Checks if key is pressed.

	Args:
		key (int): 
	Returns:
		bool
	"""

	return internal_dpg.is_key_pressed(key)

def is_key_released(key):
	"""	 Checks if key is released.

	Args:
		key (int): 
	Returns:
		bool
	"""

	return internal_dpg.is_key_released(key)

def is_mouse_button_clicked(button):
	"""	 Checks if mouse button is clicked.

	Args:
		button (int): 
	Returns:
		bool
	"""

	return internal_dpg.is_mouse_button_clicked(button)

def is_mouse_button_double_clicked(button):
	"""	 Checks if mouse button is double clicked.

	Args:
		button (int): 
	Returns:
		bool
	"""

	return internal_dpg.is_mouse_button_double_clicked(button)

def is_mouse_button_down(button):
	"""	 Checks if mouse button is down.

	Args:
		button (int): 
	Returns:
		bool
	"""

	return internal_dpg.is_mouse_button_down(button)

def is_mouse_button_dragging(button, threshold):
	"""	 Checks if mouse button is down and dragging.

	Args:
		button (int): 
		threshold (float): 
	Returns:
		bool
	"""

	return internal_dpg.is_mouse_button_dragging(button, threshold)

def is_mouse_button_released(button):
	"""	 Checks if mouse button is released.

	Args:
		button (int): 
	Returns:
		bool
	"""

	return internal_dpg.is_mouse_button_released(button)

def is_plot_queried(plot):
	"""	 Returns true if the plot is currently being queried. (Requires plot 'query' kwarg to be enabled)

	Args:
		plot (Union[int, str]): 
	Returns:
		bool
	"""

	return internal_dpg.is_plot_queried(plot)

def is_table_cell_highlighted(table, row, column):
	"""	 Checks if a table cell is highlighted.

	Args:
		table (Union[int, str]): 
		row (int): 
		column (int): 
	Returns:
		bool
	"""

	return internal_dpg.is_table_cell_highlighted(table, row, column)

def is_table_column_highlighted(table, column):
	"""	 Checks if a table column is highlighted.

	Args:
		table (Union[int, str]): 
		column (int): 
	Returns:
		bool
	"""

	return internal_dpg.is_table_column_highlighted(table, column)

def is_table_row_highlighted(table, row):
	"""	 Checks if a table row is highlighted.

	Args:
		table (Union[int, str]): 
		row (int): 
	Returns:
		bool
	"""

	return internal_dpg.is_table_row_highlighted(table, row)

def is_viewport_ok():
	"""	 Checks if a viewport has been created and shown.

	Args:
	Returns:
		bool
	"""

	return internal_dpg.is_viewport_ok()

def last_container():
	"""	 Returns the last container item added.

	Args:
	Returns:
		Union[int, str]
	"""

	return internal_dpg.last_container()

def last_item():
	"""	 Returns the last item added.

	Args:
	Returns:
		Union[int, str]
	"""

	return internal_dpg.last_item()

def last_root():
	"""	 Returns the last root added (registry or window).

	Args:
	Returns:
		Union[int, str]
	"""

	return internal_dpg.last_root()

def load_image(file, **kwargs):
	"""	 Loads an image. Returns width, height, channels, mvBuffer

	Args:
		file (str): 
		gamma (float, optional): Gamma correction factor. (default is 1.0 to avoid automatic gamma correction on loading.
		gamma_scale_factor (float, optional): Gamma scale factor.
	Returns:
		Any
	"""

	return internal_dpg.load_image(file, **kwargs)

def lock_mutex():
	"""	 Locks render thread mutex.

	Args:
	Returns:
		None
	"""

	return internal_dpg.lock_mutex()

def maximize_viewport():
	"""	 Maximizes the viewport.

	Args:
	Returns:
		None
	"""

	return internal_dpg.maximize_viewport()

def minimize_viewport():
	"""	 Minimizes a viewport.

	Args:
	Returns:
		None
	"""

	return internal_dpg.minimize_viewport()

def move_item(item, **kwargs):
	"""	 Moves an item to a new location.

	Args:
		item (Union[int, str]): 
		parent (Union[int, str], optional): 
		before (Union[int, str], optional): 
	Returns:
		None
	"""

	return internal_dpg.move_item(item, **kwargs)

def move_item_down(item):
	"""	 Moves an item down.

	Args:
		item (Union[int, str]): 
	Returns:
		None
	"""

	return internal_dpg.move_item_down(item)

def move_item_up(item):
	"""	 Moves an item up.

	Args:
		item (Union[int, str]): 
	Returns:
		None
	"""

	return internal_dpg.move_item_up(item)

def pop_container_stack():
	"""	 Pops the top item off the parent stack and return its ID.

	Args:
	Returns:
		Union[int, str]
	"""

	return internal_dpg.pop_container_stack()

def push_container_stack(item):
	"""	 Pushes an item onto the container stack.

	Args:
		item (Union[int, str]): 
	Returns:
		bool
	"""

	return internal_dpg.push_container_stack(item)

def remove_alias(alias):
	"""	 Removes an alias.

	Args:
		alias (str): 
	Returns:
		None
	"""

	return internal_dpg.remove_alias(alias)

def render_dearpygui_frame():
	"""	 Render a single Dear PyGui frame.

	Args:
	Returns:
		None
	"""

	return internal_dpg.render_dearpygui_frame()

def reorder_items(container, slot, new_order):
	"""	 Reorders an item's children.

	Args:
		container (Union[int, str]): 
		slot (int): 
		new_order (Union[List[int], Tuple[int, ...]]): 
	Returns:
		None
	"""

	return internal_dpg.reorder_items(container, slot, new_order)

def reset_axis_ticks(axis):
	"""	 Removes the manually set axis ticks and applies the default axis ticks

	Args:
		axis (Union[int, str]): 
	Returns:
		None
	"""

	return internal_dpg.reset_axis_ticks(axis)

def reset_pos(item):
	"""	 Resets an item's position after using 'set_item_pos'.

	Args:
		item (Union[int, str]): 
	Returns:
		None
	"""

	return internal_dpg.reset_pos(item)

def sample_colormap(colormap, t):
	"""	 Returns a color from a colormap given t between 0.0-1.0.

	Args:
		colormap (Union[int, str]): The colormap tag. This should come from a colormap that was added to a colormap registry. Built in color maps are accessible through their corresponding constants mvPlotColormap_Twilight, mvPlotColormap_***
		t (float): Value of the colormap to sample between 0.0-1.0
	Returns:
		Union[List[int], Tuple[int, ...]]
	"""

	return internal_dpg.sample_colormap(colormap, t)

def save_init_file(file):
	"""	 Save dpg.ini file.

	Args:
		file (str): 
	Returns:
		None
	"""

	return internal_dpg.save_init_file(file)

def set_axis_limits(axis, ymin, ymax):
	"""	 Sets limits on the axis for pan and zoom.

	Args:
		axis (Union[int, str]): 
		ymin (float): 
		ymax (float): 
	Returns:
		None
	"""

	return internal_dpg.set_axis_limits(axis, ymin, ymax)

def set_axis_limits_auto(axis):
	"""	 Removes all limits on specified axis.

	Args:
		axis (Union[int, str]): 
	Returns:
		None
	"""

	return internal_dpg.set_axis_limits_auto(axis)

def set_axis_ticks(axis, label_pairs):
	"""	 Replaces axis ticks with 'label_pairs' argument.

	Args:
		axis (Union[int, str]): 
		label_pairs (Any): Tuples of label and value in the form '((label, axis_value), (label, axis_value), ...)'
	Returns:
		None
	"""

	return internal_dpg.set_axis_ticks(axis, label_pairs)

def set_clip_space(item, top_left_x, top_left_y, width, height, min_depth, max_depth):
	"""	 New in 1.1. Set the clip space for depth clipping and 'viewport' transformation.

	Args:
		item (Union[int, str]): draw layer to set clip space
		top_left_x (float): angle to rotate
		top_left_y (float): angle to rotate
		width (float): angle to rotate
		height (float): angle to rotate
		min_depth (float): angle to rotate
		max_depth (float): angle to rotate
	Returns:
		None
	"""

	return internal_dpg.set_clip_space(item, top_left_x, top_left_y, width, height, min_depth, max_depth)

def set_clipboard_text(text):
	"""	 New in 1.3. Sets the clipboard text.

	Args:
		text (str): 
	Returns:
		None
	"""

	return internal_dpg.set_clipboard_text(text)

def set_exit_callback(callback, **kwargs):
	"""	 Sets a callback to run on last frame.

	Args:
		callback (Callable): 
		user_data (Any, optional): New in 1.3. Optional user data to send to the callback
	Returns:
		str
	"""

	return internal_dpg.set_exit_callback(callback, **kwargs)

def set_frame_callback(frame, callback, **kwargs):
	"""	 Sets a callback to run on first frame.

	Args:
		frame (int): 
		callback (Callable): 
		user_data (Any, optional): New in 1.3. Optional user data to send to the callback
	Returns:
		str
	"""

	return internal_dpg.set_frame_callback(frame, callback, **kwargs)

def set_global_font_scale(scale):
	"""	 Sets global font scale.

	Args:
		scale (float): 
	Returns:
		None
	"""

	return internal_dpg.set_global_font_scale(scale)

def set_item_alias(item, alias):
	"""	 Sets an item's alias.

	Args:
		item (Union[int, str]): 
		alias (str): 
	Returns:
		None
	"""

	return internal_dpg.set_item_alias(item, alias)

def set_item_children(item, source, slot):
	"""	 Sets an item's children.

	Args:
		item (Union[int, str]): 
		source (Union[int, str]): 
		slot (int): 
	Returns:
		None
	"""

	return internal_dpg.set_item_children(item, source, slot)

def set_primary_window(window, value):
	"""	 Sets the primary window.

	Args:
		window (Union[int, str]): 
		value (bool): 
	Returns:
		None
	"""

	return internal_dpg.set_primary_window(window, value)

def set_table_row_color(table, row, color):
	"""	 Set table row color.

	Args:
		table (Union[int, str]): 
		row (int): 
		color (Union[List[int], Tuple[int, ...]]): 
	Returns:
		None
	"""

	return internal_dpg.set_table_row_color(table, row, color)

def set_value(item, value):
	"""	 Set's an item's value.

	Args:
		item (Union[int, str]): 
		value (Any): 
	Returns:
		None
	"""

	return internal_dpg.set_value(item, value)

def set_viewport_resize_callback(callback, **kwargs):
	"""	 Sets a callback to run on viewport resize.

	Args:
		callback (Callable): 
		user_data (Any, optional): New in 1.3. Optional user data to send to the callback
	Returns:
		str
	"""

	return internal_dpg.set_viewport_resize_callback(callback, **kwargs)

def set_x_scroll(item, value):
	"""	 Undocumented

	Args:
		item (Union[int, str]): 
		value (float): 
	Returns:
		None
	"""

	return internal_dpg.set_x_scroll(item, value)

def set_y_scroll(item, value):
	"""	 Undocumented

	Args:
		item (Union[int, str]): 
		value (float): 
	Returns:
		None
	"""

	return internal_dpg.set_y_scroll(item, value)

def setup_dearpygui():
	"""	 Sets up Dear PyGui

	Args:
		viewport (Union[int, str], optional): (deprecated)
	Returns:
		None
	"""

	return internal_dpg.setup_dearpygui()

def show_imgui_demo():
	"""	 Shows the imgui demo.

	Args:
	Returns:
		None
	"""

	return internal_dpg.show_imgui_demo()

def show_implot_demo():
	"""	 Shows the implot demo.

	Args:
	Returns:
		None
	"""

	return internal_dpg.show_implot_demo()

def show_item_debug(item):
	"""	 Shows an item's debug window

	Args:
		item (Union[int, str]): 
	Returns:
		None
	"""

	return internal_dpg.show_item_debug(item)

def show_tool(tool):
	"""	 Shows a built in tool.

	Args:
		tool (Union[int, str]): 
	Returns:
		str
	"""

	return internal_dpg.show_tool(tool)

def show_viewport(**kwargs):
	"""	 Shows the main viewport.

	Args:
		minimized (bool, optional): Sets the state of the viewport to minimized
		maximized (bool, optional): Sets the state of the viewport to maximized
		viewport (Union[int, str], optional): (deprecated)
	Returns:
		None
	"""

	return internal_dpg.show_viewport(**kwargs)

def split_frame(**kwargs):
	"""	 Waits one frame.

	Args:
		delay (int, optional): Minimal delay in in milliseconds
	Returns:
		None
	"""

	return internal_dpg.split_frame(**kwargs)

def stop_dearpygui():
	"""	 Stops Dear PyGui

	Args:
	Returns:
		None
	"""

	return internal_dpg.stop_dearpygui()

def toggle_viewport_fullscreen():
	"""	 Toggle viewport fullscreen mode..

	Args:
	Returns:
		None
	"""

	return internal_dpg.toggle_viewport_fullscreen()

def top_container_stack():
	"""	 Returns the item on the top of the container stack.

	Args:
	Returns:
		Union[int, str]
	"""

	return internal_dpg.top_container_stack()

def unhighlight_table_cell(table, row, column):
	"""	 Unhighlight specified table cell.

	Args:
		table (Union[int, str]): 
		row (int): 
		column (int): 
	Returns:
		None
	"""

	return internal_dpg.unhighlight_table_cell(table, row, column)

def unhighlight_table_column(table, column):
	"""	 Unhighlight specified table column.

	Args:
		table (Union[int, str]): 
		column (int): 
	Returns:
		None
	"""

	return internal_dpg.unhighlight_table_column(table, column)

def unhighlight_table_row(table, row):
	"""	 Unhighlight specified table row.

	Args:
		table (Union[int, str]): 
		row (int): 
	Returns:
		None
	"""

	return internal_dpg.unhighlight_table_row(table, row)

def unlock_mutex():
	"""	 Unlocks render thread mutex

	Args:
	Returns:
		None
	"""

	return internal_dpg.unlock_mutex()

def unset_table_row_color(table, row):
	"""	 Remove user set table row color.

	Args:
		table (Union[int, str]): 
		row (int): 
	Returns:
		None
	"""

	return internal_dpg.unset_table_row_color(table, row)

def unstage(item):
	"""	 Unstages an item.

	Args:
		item (Union[int, str]): 
	Returns:
		None
	"""

	return internal_dpg.unstage(item)


##########################################################
# Constants #
##########################################################

mvMouseButton_Left=internal_dpg.mvMouseButton_Left
mvMouseButton_Right=internal_dpg.mvMouseButton_Right
mvMouseButton_Middle=internal_dpg.mvMouseButton_Middle
mvMouseButton_X1=internal_dpg.mvMouseButton_X1
mvMouseButton_X2=internal_dpg.mvMouseButton_X2
mvKey_0=internal_dpg.mvKey_0
mvKey_1=internal_dpg.mvKey_1
mvKey_2=internal_dpg.mvKey_2
mvKey_3=internal_dpg.mvKey_3
mvKey_4=internal_dpg.mvKey_4
mvKey_5=internal_dpg.mvKey_5
mvKey_6=internal_dpg.mvKey_6
mvKey_7=internal_dpg.mvKey_7
mvKey_8=internal_dpg.mvKey_8
mvKey_9=internal_dpg.mvKey_9
mvKey_A=internal_dpg.mvKey_A
mvKey_B=internal_dpg.mvKey_B
mvKey_C=internal_dpg.mvKey_C
mvKey_D=internal_dpg.mvKey_D
mvKey_E=internal_dpg.mvKey_E
mvKey_F=internal_dpg.mvKey_F
mvKey_G=internal_dpg.mvKey_G
mvKey_H=internal_dpg.mvKey_H
mvKey_I=internal_dpg.mvKey_I
mvKey_J=internal_dpg.mvKey_J
mvKey_K=internal_dpg.mvKey_K
mvKey_L=internal_dpg.mvKey_L
mvKey_M=internal_dpg.mvKey_M
mvKey_N=internal_dpg.mvKey_N
mvKey_O=internal_dpg.mvKey_O
mvKey_P=internal_dpg.mvKey_P
mvKey_Q=internal_dpg.mvKey_Q
mvKey_R=internal_dpg.mvKey_R
mvKey_S=internal_dpg.mvKey_S
mvKey_T=internal_dpg.mvKey_T
mvKey_U=internal_dpg.mvKey_U
mvKey_V=internal_dpg.mvKey_V
mvKey_W=internal_dpg.mvKey_W
mvKey_X=internal_dpg.mvKey_X
mvKey_Y=internal_dpg.mvKey_Y
mvKey_Z=internal_dpg.mvKey_Z
mvKey_Back=internal_dpg.mvKey_Back
mvKey_Tab=internal_dpg.mvKey_Tab
mvKey_Clear=internal_dpg.mvKey_Clear
mvKey_Return=internal_dpg.mvKey_Return
mvKey_Shift=internal_dpg.mvKey_Shift
mvKey_Control=internal_dpg.mvKey_Control
mvKey_Alt=internal_dpg.mvKey_Alt
mvKey_Pause=internal_dpg.mvKey_Pause
mvKey_Capital=internal_dpg.mvKey_Capital
mvKey_Escape=internal_dpg.mvKey_Escape
mvKey_Spacebar=internal_dpg.mvKey_Spacebar
mvKey_Prior=internal_dpg.mvKey_Prior
mvKey_Next=internal_dpg.mvKey_Next
mvKey_End=internal_dpg.mvKey_End
mvKey_Home=internal_dpg.mvKey_Home
mvKey_Left=internal_dpg.mvKey_Left
mvKey_Up=internal_dpg.mvKey_Up
mvKey_Right=internal_dpg.mvKey_Right
mvKey_Down=internal_dpg.mvKey_Down
mvKey_Select=internal_dpg.mvKey_Select
mvKey_Print=internal_dpg.mvKey_Print
mvKey_Execute=internal_dpg.mvKey_Execute
mvKey_PrintScreen=internal_dpg.mvKey_PrintScreen
mvKey_Insert=internal_dpg.mvKey_Insert
mvKey_Delete=internal_dpg.mvKey_Delete
mvKey_Help=internal_dpg.mvKey_Help
mvKey_LWin=internal_dpg.mvKey_LWin
mvKey_RWin=internal_dpg.mvKey_RWin
mvKey_Apps=internal_dpg.mvKey_Apps
mvKey_Sleep=internal_dpg.mvKey_Sleep
mvKey_NumPad0=internal_dpg.mvKey_NumPad0
mvKey_NumPad1=internal_dpg.mvKey_NumPad1
mvKey_NumPad2=internal_dpg.mvKey_NumPad2
mvKey_NumPad3=internal_dpg.mvKey_NumPad3
mvKey_NumPad4=internal_dpg.mvKey_NumPad4
mvKey_NumPad5=internal_dpg.mvKey_NumPad5
mvKey_NumPad6=internal_dpg.mvKey_NumPad6
mvKey_NumPad7=internal_dpg.mvKey_NumPad7
mvKey_NumPad8=internal_dpg.mvKey_NumPad8
mvKey_NumPad9=internal_dpg.mvKey_NumPad9
mvKey_Multiply=internal_dpg.mvKey_Multiply
mvKey_Add=internal_dpg.mvKey_Add
mvKey_Separator=internal_dpg.mvKey_Separator
mvKey_Subtract=internal_dpg.mvKey_Subtract
mvKey_Decimal=internal_dpg.mvKey_Decimal
mvKey_Divide=internal_dpg.mvKey_Divide
mvKey_F1=internal_dpg.mvKey_F1
mvKey_F2=internal_dpg.mvKey_F2
mvKey_F3=internal_dpg.mvKey_F3
mvKey_F4=internal_dpg.mvKey_F4
mvKey_F5=internal_dpg.mvKey_F5
mvKey_F6=internal_dpg.mvKey_F6
mvKey_F7=internal_dpg.mvKey_F7
mvKey_F8=internal_dpg.mvKey_F8
mvKey_F9=internal_dpg.mvKey_F9
mvKey_F10=internal_dpg.mvKey_F10
mvKey_F11=internal_dpg.mvKey_F11
mvKey_F12=internal_dpg.mvKey_F12
mvKey_F13=internal_dpg.mvKey_F13
mvKey_F14=internal_dpg.mvKey_F14
mvKey_F15=internal_dpg.mvKey_F15
mvKey_F16=internal_dpg.mvKey_F16
mvKey_F17=internal_dpg.mvKey_F17
mvKey_F18=internal_dpg.mvKey_F18
mvKey_F19=internal_dpg.mvKey_F19
mvKey_F20=internal_dpg.mvKey_F20
mvKey_F21=internal_dpg.mvKey_F21
mvKey_F22=internal_dpg.mvKey_F22
mvKey_F23=internal_dpg.mvKey_F23
mvKey_F24=internal_dpg.mvKey_F24
mvKey_NumLock=internal_dpg.mvKey_NumLock
mvKey_ScrollLock=internal_dpg.mvKey_ScrollLock
mvKey_LShift=internal_dpg.mvKey_LShift
mvKey_RShift=internal_dpg.mvKey_RShift
mvKey_LControl=internal_dpg.mvKey_LControl
mvKey_RControl=internal_dpg.mvKey_RControl
mvKey_LMenu=internal_dpg.mvKey_LMenu
mvKey_RMenu=internal_dpg.mvKey_RMenu
mvKey_Browser_Back=internal_dpg.mvKey_Browser_Back
mvKey_Browser_Forward=internal_dpg.mvKey_Browser_Forward
mvKey_Browser_Refresh=internal_dpg.mvKey_Browser_Refresh
mvKey_Browser_Stop=internal_dpg.mvKey_Browser_Stop
mvKey_Browser_Search=internal_dpg.mvKey_Browser_Search
mvKey_Browser_Favorites=internal_dpg.mvKey_Browser_Favorites
mvKey_Browser_Home=internal_dpg.mvKey_Browser_Home
mvKey_Volume_Mute=internal_dpg.mvKey_Volume_Mute
mvKey_Volume_Down=internal_dpg.mvKey_Volume_Down
mvKey_Volume_Up=internal_dpg.mvKey_Volume_Up
mvKey_Media_Next_Track=internal_dpg.mvKey_Media_Next_Track
mvKey_Media_Prev_Track=internal_dpg.mvKey_Media_Prev_Track
mvKey_Media_Stop=internal_dpg.mvKey_Media_Stop
mvKey_Media_Play_Pause=internal_dpg.mvKey_Media_Play_Pause
mvKey_Launch_Mail=internal_dpg.mvKey_Launch_Mail
mvKey_Launch_Media_Select=internal_dpg.mvKey_Launch_Media_Select
mvKey_Launch_App1=internal_dpg.mvKey_Launch_App1
mvKey_Launch_App2=internal_dpg.mvKey_Launch_App2
mvKey_Colon=internal_dpg.mvKey_Colon
mvKey_Plus=internal_dpg.mvKey_Plus
mvKey_Comma=internal_dpg.mvKey_Comma
mvKey_Minus=internal_dpg.mvKey_Minus
mvKey_Period=internal_dpg.mvKey_Period
mvKey_Slash=internal_dpg.mvKey_Slash
mvKey_Tilde=internal_dpg.mvKey_Tilde
mvKey_Open_Brace=internal_dpg.mvKey_Open_Brace
mvKey_Backslash=internal_dpg.mvKey_Backslash
mvKey_Close_Brace=internal_dpg.mvKey_Close_Brace
mvKey_Quote=internal_dpg.mvKey_Quote
mvAll=internal_dpg.mvAll
mvTool_About=internal_dpg.mvTool_About
mvTool_Debug=internal_dpg.mvTool_Debug
mvTool_Doc=internal_dpg.mvTool_Doc
mvTool_ItemRegistry=internal_dpg.mvTool_ItemRegistry
mvTool_Metrics=internal_dpg.mvTool_Metrics
mvTool_Style=internal_dpg.mvTool_Style
mvTool_Font=internal_dpg.mvTool_Font
mvFontAtlas=internal_dpg.mvFontAtlas
mvAppUUID=internal_dpg.mvAppUUID
mvInvalidUUID=internal_dpg.mvInvalidUUID
mvDir_None=internal_dpg.mvDir_None
mvDir_Left=internal_dpg.mvDir_Left
mvDir_Right=internal_dpg.mvDir_Right
mvDir_Up=internal_dpg.mvDir_Up
mvDir_Down=internal_dpg.mvDir_Down
mvComboHeight_Small=internal_dpg.mvComboHeight_Small
mvComboHeight_Regular=internal_dpg.mvComboHeight_Regular
mvComboHeight_Large=internal_dpg.mvComboHeight_Large
mvComboHeight_Largest=internal_dpg.mvComboHeight_Largest
mvColorEdit_AlphaPreviewNone=internal_dpg.mvColorEdit_AlphaPreviewNone
mvColorEdit_AlphaPreview=internal_dpg.mvColorEdit_AlphaPreview
mvColorEdit_AlphaPreviewHalf=internal_dpg.mvColorEdit_AlphaPreviewHalf
mvColorEdit_uint8=internal_dpg.mvColorEdit_uint8
mvColorEdit_float=internal_dpg.mvColorEdit_float
mvColorEdit_rgb=internal_dpg.mvColorEdit_rgb
mvColorEdit_hsv=internal_dpg.mvColorEdit_hsv
mvColorEdit_hex=internal_dpg.mvColorEdit_hex
mvColorEdit_input_rgb=internal_dpg.mvColorEdit_input_rgb
mvColorEdit_input_hsv=internal_dpg.mvColorEdit_input_hsv
mvPlotColormap_Default=internal_dpg.mvPlotColormap_Default
mvPlotColormap_Deep=internal_dpg.mvPlotColormap_Deep
mvPlotColormap_Dark=internal_dpg.mvPlotColormap_Dark
mvPlotColormap_Pastel=internal_dpg.mvPlotColormap_Pastel
mvPlotColormap_Paired=internal_dpg.mvPlotColormap_Paired
mvPlotColormap_Viridis=internal_dpg.mvPlotColormap_Viridis
mvPlotColormap_Plasma=internal_dpg.mvPlotColormap_Plasma
mvPlotColormap_Hot=internal_dpg.mvPlotColormap_Hot
mvPlotColormap_Cool=internal_dpg.mvPlotColormap_Cool
mvPlotColormap_Pink=internal_dpg.mvPlotColormap_Pink
mvPlotColormap_Jet=internal_dpg.mvPlotColormap_Jet
mvPlotColormap_Twilight=internal_dpg.mvPlotColormap_Twilight
mvPlotColormap_RdBu=internal_dpg.mvPlotColormap_RdBu
mvPlotColormap_BrBG=internal_dpg.mvPlotColormap_BrBG
mvPlotColormap_PiYG=internal_dpg.mvPlotColormap_PiYG
mvPlotColormap_Spectral=internal_dpg.mvPlotColormap_Spectral
mvPlotColormap_Greys=internal_dpg.mvPlotColormap_Greys
mvColorPicker_bar=internal_dpg.mvColorPicker_bar
mvColorPicker_wheel=internal_dpg.mvColorPicker_wheel
mvTabOrder_Reorderable=internal_dpg.mvTabOrder_Reorderable
mvTabOrder_Fixed=internal_dpg.mvTabOrder_Fixed
mvTabOrder_Leading=internal_dpg.mvTabOrder_Leading
mvTabOrder_Trailing=internal_dpg.mvTabOrder_Trailing
mvDatePickerLevel_Day=internal_dpg.mvDatePickerLevel_Day
mvDatePickerLevel_Month=internal_dpg.mvDatePickerLevel_Month
mvDatePickerLevel_Year=internal_dpg.mvDatePickerLevel_Year
mvCullMode_None=internal_dpg.mvCullMode_None
mvCullMode_Back=internal_dpg.mvCullMode_Back
mvCullMode_Front=internal_dpg.mvCullMode_Front
mvFontRangeHint_Default=internal_dpg.mvFontRangeHint_Default
mvFontRangeHint_Japanese=internal_dpg.mvFontRangeHint_Japanese
mvFontRangeHint_Korean=internal_dpg.mvFontRangeHint_Korean
mvFontRangeHint_Chinese_Full=internal_dpg.mvFontRangeHint_Chinese_Full
mvFontRangeHint_Chinese_Simplified_Common=internal_dpg.mvFontRangeHint_Chinese_Simplified_Common
mvFontRangeHint_Cyrillic=internal_dpg.mvFontRangeHint_Cyrillic
mvFontRangeHint_Thai=internal_dpg.mvFontRangeHint_Thai
mvFontRangeHint_Vietnamese=internal_dpg.mvFontRangeHint_Vietnamese
mvNode_PinShape_Circle=internal_dpg.mvNode_PinShape_Circle
mvNode_PinShape_CircleFilled=internal_dpg.mvNode_PinShape_CircleFilled
mvNode_PinShape_Triangle=internal_dpg.mvNode_PinShape_Triangle
mvNode_PinShape_TriangleFilled=internal_dpg.mvNode_PinShape_TriangleFilled
mvNode_PinShape_Quad=internal_dpg.mvNode_PinShape_Quad
mvNode_PinShape_QuadFilled=internal_dpg.mvNode_PinShape_QuadFilled
mvNode_Attr_Input=internal_dpg.mvNode_Attr_Input
mvNode_Attr_Output=internal_dpg.mvNode_Attr_Output
mvNode_Attr_Static=internal_dpg.mvNode_Attr_Static
mvPlotBin_Sqrt=internal_dpg.mvPlotBin_Sqrt
mvPlotBin_Sturges=internal_dpg.mvPlotBin_Sturges
mvPlotBin_Rice=internal_dpg.mvPlotBin_Rice
mvPlotBin_Scott=internal_dpg.mvPlotBin_Scott
mvXAxis=internal_dpg.mvXAxis
mvYAxis=internal_dpg.mvYAxis
mvPlotMarker_None=internal_dpg.mvPlotMarker_None
mvPlotMarker_Circle=internal_dpg.mvPlotMarker_Circle
mvPlotMarker_Square=internal_dpg.mvPlotMarker_Square
mvPlotMarker_Diamond=internal_dpg.mvPlotMarker_Diamond
mvPlotMarker_Up=internal_dpg.mvPlotMarker_Up
mvPlotMarker_Down=internal_dpg.mvPlotMarker_Down
mvPlotMarker_Left=internal_dpg.mvPlotMarker_Left
mvPlotMarker_Right=internal_dpg.mvPlotMarker_Right
mvPlotMarker_Cross=internal_dpg.mvPlotMarker_Cross
mvPlotMarker_Plus=internal_dpg.mvPlotMarker_Plus
mvPlotMarker_Asterisk=internal_dpg.mvPlotMarker_Asterisk
mvPlot_Location_Center=internal_dpg.mvPlot_Location_Center
mvPlot_Location_North=internal_dpg.mvPlot_Location_North
mvPlot_Location_South=internal_dpg.mvPlot_Location_South
mvPlot_Location_West=internal_dpg.mvPlot_Location_West
mvPlot_Location_East=internal_dpg.mvPlot_Location_East
mvPlot_Location_NorthWest=internal_dpg.mvPlot_Location_NorthWest
mvPlot_Location_NorthEast=internal_dpg.mvPlot_Location_NorthEast
mvPlot_Location_SouthWest=internal_dpg.mvPlot_Location_SouthWest
mvPlot_Location_SouthEast=internal_dpg.mvPlot_Location_SouthEast
mvTable_SizingFixedFit=internal_dpg.mvTable_SizingFixedFit
mvTable_SizingFixedSame=internal_dpg.mvTable_SizingFixedSame
mvTable_SizingStretchProp=internal_dpg.mvTable_SizingStretchProp
mvTable_SizingStretchSame=internal_dpg.mvTable_SizingStretchSame
mvFormat_Float_rgba=internal_dpg.mvFormat_Float_rgba
mvFormat_Float_rgb=internal_dpg.mvFormat_Float_rgb
mvThemeCat_Core=internal_dpg.mvThemeCat_Core
mvThemeCat_Plots=internal_dpg.mvThemeCat_Plots
mvThemeCat_Nodes=internal_dpg.mvThemeCat_Nodes
mvThemeCol_Text=internal_dpg.mvThemeCol_Text
mvThemeCol_TextDisabled=internal_dpg.mvThemeCol_TextDisabled
mvThemeCol_WindowBg=internal_dpg.mvThemeCol_WindowBg
mvThemeCol_ChildBg=internal_dpg.mvThemeCol_ChildBg
mvThemeCol_Border=internal_dpg.mvThemeCol_Border
mvThemeCol_PopupBg=internal_dpg.mvThemeCol_PopupBg
mvThemeCol_BorderShadow=internal_dpg.mvThemeCol_BorderShadow
mvThemeCol_FrameBg=internal_dpg.mvThemeCol_FrameBg
mvThemeCol_FrameBgHovered=internal_dpg.mvThemeCol_FrameBgHovered
mvThemeCol_FrameBgActive=internal_dpg.mvThemeCol_FrameBgActive
mvThemeCol_TitleBg=internal_dpg.mvThemeCol_TitleBg
mvThemeCol_TitleBgActive=internal_dpg.mvThemeCol_TitleBgActive
mvThemeCol_TitleBgCollapsed=internal_dpg.mvThemeCol_TitleBgCollapsed
mvThemeCol_MenuBarBg=internal_dpg.mvThemeCol_MenuBarBg
mvThemeCol_ScrollbarBg=internal_dpg.mvThemeCol_ScrollbarBg
mvThemeCol_ScrollbarGrab=internal_dpg.mvThemeCol_ScrollbarGrab
mvThemeCol_ScrollbarGrabHovered=internal_dpg.mvThemeCol_ScrollbarGrabHovered
mvThemeCol_ScrollbarGrabActive=internal_dpg.mvThemeCol_ScrollbarGrabActive
mvThemeCol_CheckMark=internal_dpg.mvThemeCol_CheckMark
mvThemeCol_SliderGrab=internal_dpg.mvThemeCol_SliderGrab
mvThemeCol_SliderGrabActive=internal_dpg.mvThemeCol_SliderGrabActive
mvThemeCol_Button=internal_dpg.mvThemeCol_Button
mvThemeCol_ButtonHovered=internal_dpg.mvThemeCol_ButtonHovered
mvThemeCol_ButtonActive=internal_dpg.mvThemeCol_ButtonActive
mvThemeCol_Header=internal_dpg.mvThemeCol_Header
mvThemeCol_HeaderHovered=internal_dpg.mvThemeCol_HeaderHovered
mvThemeCol_HeaderActive=internal_dpg.mvThemeCol_HeaderActive
mvThemeCol_Separator=internal_dpg.mvThemeCol_Separator
mvThemeCol_SeparatorHovered=internal_dpg.mvThemeCol_SeparatorHovered
mvThemeCol_SeparatorActive=internal_dpg.mvThemeCol_SeparatorActive
mvThemeCol_ResizeGrip=internal_dpg.mvThemeCol_ResizeGrip
mvThemeCol_ResizeGripHovered=internal_dpg.mvThemeCol_ResizeGripHovered
mvThemeCol_ResizeGripActive=internal_dpg.mvThemeCol_ResizeGripActive
mvThemeCol_Tab=internal_dpg.mvThemeCol_Tab
mvThemeCol_TabHovered=internal_dpg.mvThemeCol_TabHovered
mvThemeCol_TabActive=internal_dpg.mvThemeCol_TabActive
mvThemeCol_TabUnfocused=internal_dpg.mvThemeCol_TabUnfocused
mvThemeCol_TabUnfocusedActive=internal_dpg.mvThemeCol_TabUnfocusedActive
mvThemeCol_DockingPreview=internal_dpg.mvThemeCol_DockingPreview
mvThemeCol_DockingEmptyBg=internal_dpg.mvThemeCol_DockingEmptyBg
mvThemeCol_PlotLines=internal_dpg.mvThemeCol_PlotLines
mvThemeCol_PlotLinesHovered=internal_dpg.mvThemeCol_PlotLinesHovered
mvThemeCol_PlotHistogram=internal_dpg.mvThemeCol_PlotHistogram
mvThemeCol_PlotHistogramHovered=internal_dpg.mvThemeCol_PlotHistogramHovered
mvThemeCol_TableHeaderBg=internal_dpg.mvThemeCol_TableHeaderBg
mvThemeCol_TableBorderStrong=internal_dpg.mvThemeCol_TableBorderStrong
mvThemeCol_TableBorderLight=internal_dpg.mvThemeCol_TableBorderLight
mvThemeCol_TableRowBg=internal_dpg.mvThemeCol_TableRowBg
mvThemeCol_TableRowBgAlt=internal_dpg.mvThemeCol_TableRowBgAlt
mvThemeCol_TextSelectedBg=internal_dpg.mvThemeCol_TextSelectedBg
mvThemeCol_DragDropTarget=internal_dpg.mvThemeCol_DragDropTarget
mvThemeCol_NavHighlight=internal_dpg.mvThemeCol_NavHighlight
mvThemeCol_NavWindowingHighlight=internal_dpg.mvThemeCol_NavWindowingHighlight
mvThemeCol_NavWindowingDimBg=internal_dpg.mvThemeCol_NavWindowingDimBg
mvThemeCol_ModalWindowDimBg=internal_dpg.mvThemeCol_ModalWindowDimBg
mvPlotCol_Line=internal_dpg.mvPlotCol_Line
mvPlotCol_Fill=internal_dpg.mvPlotCol_Fill
mvPlotCol_MarkerOutline=internal_dpg.mvPlotCol_MarkerOutline
mvPlotCol_MarkerFill=internal_dpg.mvPlotCol_MarkerFill
mvPlotCol_ErrorBar=internal_dpg.mvPlotCol_ErrorBar
mvPlotCol_FrameBg=internal_dpg.mvPlotCol_FrameBg
mvPlotCol_PlotBg=internal_dpg.mvPlotCol_PlotBg
mvPlotCol_PlotBorder=internal_dpg.mvPlotCol_PlotBorder
mvPlotCol_LegendBg=internal_dpg.mvPlotCol_LegendBg
mvPlotCol_LegendBorder=internal_dpg.mvPlotCol_LegendBorder
mvPlotCol_LegendText=internal_dpg.mvPlotCol_LegendText
mvPlotCol_TitleText=internal_dpg.mvPlotCol_TitleText
mvPlotCol_InlayText=internal_dpg.mvPlotCol_InlayText
mvPlotCol_XAxis=internal_dpg.mvPlotCol_XAxis
mvPlotCol_XAxisGrid=internal_dpg.mvPlotCol_XAxisGrid
mvPlotCol_YAxis=internal_dpg.mvPlotCol_YAxis
mvPlotCol_YAxisGrid=internal_dpg.mvPlotCol_YAxisGrid
mvPlotCol_YAxis2=internal_dpg.mvPlotCol_YAxis2
mvPlotCol_YAxisGrid2=internal_dpg.mvPlotCol_YAxisGrid2
mvPlotCol_YAxis3=internal_dpg.mvPlotCol_YAxis3
mvPlotCol_YAxisGrid3=internal_dpg.mvPlotCol_YAxisGrid3
mvPlotCol_Selection=internal_dpg.mvPlotCol_Selection
mvPlotCol_Query=internal_dpg.mvPlotCol_Query
mvPlotCol_Crosshairs=internal_dpg.mvPlotCol_Crosshairs
mvNodeCol_NodeBackground=internal_dpg.mvNodeCol_NodeBackground
mvNodeCol_NodeBackgroundHovered=internal_dpg.mvNodeCol_NodeBackgroundHovered
mvNodeCol_NodeBackgroundSelected=internal_dpg.mvNodeCol_NodeBackgroundSelected
mvNodeCol_NodeOutline=internal_dpg.mvNodeCol_NodeOutline
mvNodeCol_TitleBar=internal_dpg.mvNodeCol_TitleBar
mvNodeCol_TitleBarHovered=internal_dpg.mvNodeCol_TitleBarHovered
mvNodeCol_TitleBarSelected=internal_dpg.mvNodeCol_TitleBarSelected
mvNodeCol_Link=internal_dpg.mvNodeCol_Link
mvNodeCol_LinkHovered=internal_dpg.mvNodeCol_LinkHovered
mvNodeCol_LinkSelected=internal_dpg.mvNodeCol_LinkSelected
mvNodeCol_Pin=internal_dpg.mvNodeCol_Pin
mvNodeCol_PinHovered=internal_dpg.mvNodeCol_PinHovered
mvNodeCol_BoxSelector=internal_dpg.mvNodeCol_BoxSelector
mvNodeCol_BoxSelectorOutline=internal_dpg.mvNodeCol_BoxSelectorOutline
mvNodeCol_GridBackground=internal_dpg.mvNodeCol_GridBackground
mvNodeCol_GridLine=internal_dpg.mvNodeCol_GridLine
mvStyleVar_Alpha=internal_dpg.mvStyleVar_Alpha
mvStyleVar_WindowPadding=internal_dpg.mvStyleVar_WindowPadding
mvStyleVar_WindowRounding=internal_dpg.mvStyleVar_WindowRounding
mvStyleVar_WindowBorderSize=internal_dpg.mvStyleVar_WindowBorderSize
mvStyleVar_WindowMinSize=internal_dpg.mvStyleVar_WindowMinSize
mvStyleVar_WindowTitleAlign=internal_dpg.mvStyleVar_WindowTitleAlign
mvStyleVar_ChildRounding=internal_dpg.mvStyleVar_ChildRounding
mvStyleVar_ChildBorderSize=internal_dpg.mvStyleVar_ChildBorderSize
mvStyleVar_PopupRounding=internal_dpg.mvStyleVar_PopupRounding
mvStyleVar_PopupBorderSize=internal_dpg.mvStyleVar_PopupBorderSize
mvStyleVar_FramePadding=internal_dpg.mvStyleVar_FramePadding
mvStyleVar_FrameRounding=internal_dpg.mvStyleVar_FrameRounding
mvStyleVar_FrameBorderSize=internal_dpg.mvStyleVar_FrameBorderSize
mvStyleVar_ItemSpacing=internal_dpg.mvStyleVar_ItemSpacing
mvStyleVar_ItemInnerSpacing=internal_dpg.mvStyleVar_ItemInnerSpacing
mvStyleVar_IndentSpacing=internal_dpg.mvStyleVar_IndentSpacing
mvStyleVar_CellPadding=internal_dpg.mvStyleVar_CellPadding
mvStyleVar_ScrollbarSize=internal_dpg.mvStyleVar_ScrollbarSize
mvStyleVar_ScrollbarRounding=internal_dpg.mvStyleVar_ScrollbarRounding
mvStyleVar_GrabMinSize=internal_dpg.mvStyleVar_GrabMinSize
mvStyleVar_GrabRounding=internal_dpg.mvStyleVar_GrabRounding
mvStyleVar_TabRounding=internal_dpg.mvStyleVar_TabRounding
mvStyleVar_ButtonTextAlign=internal_dpg.mvStyleVar_ButtonTextAlign
mvStyleVar_SelectableTextAlign=internal_dpg.mvStyleVar_SelectableTextAlign
mvPlotStyleVar_LineWeight=internal_dpg.mvPlotStyleVar_LineWeight
mvPlotStyleVar_Marker=internal_dpg.mvPlotStyleVar_Marker
mvPlotStyleVar_MarkerSize=internal_dpg.mvPlotStyleVar_MarkerSize
mvPlotStyleVar_MarkerWeight=internal_dpg.mvPlotStyleVar_MarkerWeight
mvPlotStyleVar_FillAlpha=internal_dpg.mvPlotStyleVar_FillAlpha
mvPlotStyleVar_ErrorBarSize=internal_dpg.mvPlotStyleVar_ErrorBarSize
mvPlotStyleVar_ErrorBarWeight=internal_dpg.mvPlotStyleVar_ErrorBarWeight
mvPlotStyleVar_DigitalBitHeight=internal_dpg.mvPlotStyleVar_DigitalBitHeight
mvPlotStyleVar_DigitalBitGap=internal_dpg.mvPlotStyleVar_DigitalBitGap
mvPlotStyleVar_PlotBorderSize=internal_dpg.mvPlotStyleVar_PlotBorderSize
mvPlotStyleVar_MinorAlpha=internal_dpg.mvPlotStyleVar_MinorAlpha
mvPlotStyleVar_MajorTickLen=internal_dpg.mvPlotStyleVar_MajorTickLen
mvPlotStyleVar_MinorTickLen=internal_dpg.mvPlotStyleVar_MinorTickLen
mvPlotStyleVar_MajorTickSize=internal_dpg.mvPlotStyleVar_MajorTickSize
mvPlotStyleVar_MinorTickSize=internal_dpg.mvPlotStyleVar_MinorTickSize
mvPlotStyleVar_MajorGridSize=internal_dpg.mvPlotStyleVar_MajorGridSize
mvPlotStyleVar_MinorGridSize=internal_dpg.mvPlotStyleVar_MinorGridSize
mvPlotStyleVar_PlotPadding=internal_dpg.mvPlotStyleVar_PlotPadding
mvPlotStyleVar_LabelPadding=internal_dpg.mvPlotStyleVar_LabelPadding
mvPlotStyleVar_LegendPadding=internal_dpg.mvPlotStyleVar_LegendPadding
mvPlotStyleVar_LegendInnerPadding=internal_dpg.mvPlotStyleVar_LegendInnerPadding
mvPlotStyleVar_LegendSpacing=internal_dpg.mvPlotStyleVar_LegendSpacing
mvPlotStyleVar_MousePosPadding=internal_dpg.mvPlotStyleVar_MousePosPadding
mvPlotStyleVar_AnnotationPadding=internal_dpg.mvPlotStyleVar_AnnotationPadding
mvPlotStyleVar_FitPadding=internal_dpg.mvPlotStyleVar_FitPadding
mvPlotStyleVar_PlotDefaultSize=internal_dpg.mvPlotStyleVar_PlotDefaultSize
mvPlotStyleVar_PlotMinSize=internal_dpg.mvPlotStyleVar_PlotMinSize
mvNodeStyleVar_GridSpacing=internal_dpg.mvNodeStyleVar_GridSpacing
mvNodeStyleVar_NodeCornerRounding=internal_dpg.mvNodeStyleVar_NodeCornerRounding
mvNodeStyleVar_NodePaddingHorizontal=internal_dpg.mvNodeStyleVar_NodePaddingHorizontal
mvNodeStyleVar_NodePaddingVertical=internal_dpg.mvNodeStyleVar_NodePaddingVertical
mvNodeStyleVar_NodeBorderThickness=internal_dpg.mvNodeStyleVar_NodeBorderThickness
mvNodeStyleVar_LinkThickness=internal_dpg.mvNodeStyleVar_LinkThickness
mvNodeStyleVar_LinkLineSegmentsPerLength=internal_dpg.mvNodeStyleVar_LinkLineSegmentsPerLength
mvNodeStyleVar_LinkHoverDistance=internal_dpg.mvNodeStyleVar_LinkHoverDistance
mvNodeStyleVar_PinCircleRadius=internal_dpg.mvNodeStyleVar_PinCircleRadius
mvNodeStyleVar_PinQuadSideLength=internal_dpg.mvNodeStyleVar_PinQuadSideLength
mvNodeStyleVar_PinTriangleSideLength=internal_dpg.mvNodeStyleVar_PinTriangleSideLength
mvNodeStyleVar_PinLineThickness=internal_dpg.mvNodeStyleVar_PinLineThickness
mvNodeStyleVar_PinHoverRadius=internal_dpg.mvNodeStyleVar_PinHoverRadius
mvNodeStyleVar_PinOffset=internal_dpg.mvNodeStyleVar_PinOffset
mvInputText=internal_dpg.mvInputText
mvButton=internal_dpg.mvButton
mvRadioButton=internal_dpg.mvRadioButton
mvTabBar=internal_dpg.mvTabBar
mvTab=internal_dpg.mvTab
mvImage=internal_dpg.mvImage
mvMenuBar=internal_dpg.mvMenuBar
mvViewportMenuBar=internal_dpg.mvViewportMenuBar
mvMenu=internal_dpg.mvMenu
mvMenuItem=internal_dpg.mvMenuItem
mvChildWindow=internal_dpg.mvChildWindow
mvGroup=internal_dpg.mvGroup
mvSliderFloat=internal_dpg.mvSliderFloat
mvSliderInt=internal_dpg.mvSliderInt
mvFilterSet=internal_dpg.mvFilterSet
mvDragFloat=internal_dpg.mvDragFloat
mvDragInt=internal_dpg.mvDragInt
mvInputFloat=internal_dpg.mvInputFloat
mvInputInt=internal_dpg.mvInputInt
mvColorEdit=internal_dpg.mvColorEdit
mvClipper=internal_dpg.mvClipper
mvColorPicker=internal_dpg.mvColorPicker
mvTooltip=internal_dpg.mvTooltip
mvCollapsingHeader=internal_dpg.mvCollapsingHeader
mvSeparator=internal_dpg.mvSeparator
mvCheckbox=internal_dpg.mvCheckbox
mvListbox=internal_dpg.mvListbox
mvText=internal_dpg.mvText
mvCombo=internal_dpg.mvCombo
mvPlot=internal_dpg.mvPlot
mvSimplePlot=internal_dpg.mvSimplePlot
mvDrawlist=internal_dpg.mvDrawlist
mvWindowAppItem=internal_dpg.mvWindowAppItem
mvSelectable=internal_dpg.mvSelectable
mvTreeNode=internal_dpg.mvTreeNode
mvProgressBar=internal_dpg.mvProgressBar
mvSpacer=internal_dpg.mvSpacer
mvImageButton=internal_dpg.mvImageButton
mvTimePicker=internal_dpg.mvTimePicker
mvDatePicker=internal_dpg.mvDatePicker
mvColorButton=internal_dpg.mvColorButton
mvFileDialog=internal_dpg.mvFileDialog
mvTabButton=internal_dpg.mvTabButton
mvDrawNode=internal_dpg.mvDrawNode
mvNodeEditor=internal_dpg.mvNodeEditor
mvNode=internal_dpg.mvNode
mvNodeAttribute=internal_dpg.mvNodeAttribute
mvTable=internal_dpg.mvTable
mvTableColumn=internal_dpg.mvTableColumn
mvTableRow=internal_dpg.mvTableRow
mvDrawLine=internal_dpg.mvDrawLine
mvDrawArrow=internal_dpg.mvDrawArrow
mvDrawTriangle=internal_dpg.mvDrawTriangle
mvDrawImageQuad=internal_dpg.mvDrawImageQuad
mvDrawCircle=internal_dpg.mvDrawCircle
mvDrawEllipse=internal_dpg.mvDrawEllipse
mvDrawBezierCubic=internal_dpg.mvDrawBezierCubic
mvDrawBezierQuadratic=internal_dpg.mvDrawBezierQuadratic
mvDrawQuad=internal_dpg.mvDrawQuad
mvDrawRect=internal_dpg.mvDrawRect
mvDrawText=internal_dpg.mvDrawText
mvDrawPolygon=internal_dpg.mvDrawPolygon
mvDrawPolyline=internal_dpg.mvDrawPolyline
mvDrawImage=internal_dpg.mvDrawImage
mvDragFloatMulti=internal_dpg.mvDragFloatMulti
mvDragIntMulti=internal_dpg.mvDragIntMulti
mvSliderFloatMulti=internal_dpg.mvSliderFloatMulti
mvSliderIntMulti=internal_dpg.mvSliderIntMulti
mvInputIntMulti=internal_dpg.mvInputIntMulti
mvInputFloatMulti=internal_dpg.mvInputFloatMulti
mvDragPoint=internal_dpg.mvDragPoint
mvDragLine=internal_dpg.mvDragLine
mvAnnotation=internal_dpg.mvAnnotation
mvLineSeries=internal_dpg.mvLineSeries
mvScatterSeries=internal_dpg.mvScatterSeries
mvStemSeries=internal_dpg.mvStemSeries
mvStairSeries=internal_dpg.mvStairSeries
mvBarSeries=internal_dpg.mvBarSeries
mvErrorSeries=internal_dpg.mvErrorSeries
mvVLineSeries=internal_dpg.mvVLineSeries
mvHLineSeries=internal_dpg.mvHLineSeries
mvHeatSeries=internal_dpg.mvHeatSeries
mvImageSeries=internal_dpg.mvImageSeries
mvPieSeries=internal_dpg.mvPieSeries
mvShadeSeries=internal_dpg.mvShadeSeries
mvLabelSeries=internal_dpg.mvLabelSeries
mvHistogramSeries=internal_dpg.mvHistogramSeries
mv2dHistogramSeries=internal_dpg.mv2dHistogramSeries
mvCandleSeries=internal_dpg.mvCandleSeries
mvAreaSeries=internal_dpg.mvAreaSeries
mvColorMapScale=internal_dpg.mvColorMapScale
mvSlider3D=internal_dpg.mvSlider3D
mvKnobFloat=internal_dpg.mvKnobFloat
mvLoadingIndicator=internal_dpg.mvLoadingIndicator
mvNodeLink=internal_dpg.mvNodeLink
mvTextureRegistry=internal_dpg.mvTextureRegistry
mvStaticTexture=internal_dpg.mvStaticTexture
mvDynamicTexture=internal_dpg.mvDynamicTexture
mvStage=internal_dpg.mvStage
mvDrawLayer=internal_dpg.mvDrawLayer
mvViewportDrawlist=internal_dpg.mvViewportDrawlist
mvFileExtension=internal_dpg.mvFileExtension
mvPlotLegend=internal_dpg.mvPlotLegend
mvPlotAxis=internal_dpg.mvPlotAxis
mvHandlerRegistry=internal_dpg.mvHandlerRegistry
mvKeyDownHandler=internal_dpg.mvKeyDownHandler
mvKeyPressHandler=internal_dpg.mvKeyPressHandler
mvKeyReleaseHandler=internal_dpg.mvKeyReleaseHandler
mvMouseMoveHandler=internal_dpg.mvMouseMoveHandler
mvMouseWheelHandler=internal_dpg.mvMouseWheelHandler
mvMouseClickHandler=internal_dpg.mvMouseClickHandler
mvMouseDoubleClickHandler=internal_dpg.mvMouseDoubleClickHandler
mvMouseDownHandler=internal_dpg.mvMouseDownHandler
mvMouseReleaseHandler=internal_dpg.mvMouseReleaseHandler
mvMouseDragHandler=internal_dpg.mvMouseDragHandler
mvHoverHandler=internal_dpg.mvHoverHandler
mvActiveHandler=internal_dpg.mvActiveHandler
mvFocusHandler=internal_dpg.mvFocusHandler
mvVisibleHandler=internal_dpg.mvVisibleHandler
mvEditedHandler=internal_dpg.mvEditedHandler
mvActivatedHandler=internal_dpg.mvActivatedHandler
mvDeactivatedHandler=internal_dpg.mvDeactivatedHandler
mvDeactivatedAfterEditHandler=internal_dpg.mvDeactivatedAfterEditHandler
mvToggledOpenHandler=internal_dpg.mvToggledOpenHandler
mvClickedHandler=internal_dpg.mvClickedHandler
mvDragPayload=internal_dpg.mvDragPayload
mvResizeHandler=internal_dpg.mvResizeHandler
mvFont=internal_dpg.mvFont
mvFontRegistry=internal_dpg.mvFontRegistry
mvTheme=internal_dpg.mvTheme
mvThemeColor=internal_dpg.mvThemeColor
mvThemeStyle=internal_dpg.mvThemeStyle
mvThemeComponent=internal_dpg.mvThemeComponent
mvFontRangeHint=internal_dpg.mvFontRangeHint
mvFontRange=internal_dpg.mvFontRange
mvFontChars=internal_dpg.mvFontChars
mvCharRemap=internal_dpg.mvCharRemap
mvValueRegistry=internal_dpg.mvValueRegistry
mvIntValue=internal_dpg.mvIntValue
mvFloatValue=internal_dpg.mvFloatValue
mvFloat4Value=internal_dpg.mvFloat4Value
mvInt4Value=internal_dpg.mvInt4Value
mvBoolValue=internal_dpg.mvBoolValue
mvStringValue=internal_dpg.mvStringValue
mvDoubleValue=internal_dpg.mvDoubleValue
mvDouble4Value=internal_dpg.mvDouble4Value
mvColorValue=internal_dpg.mvColorValue
mvFloatVectValue=internal_dpg.mvFloatVectValue
mvSeriesValue=internal_dpg.mvSeriesValue
mvRawTexture=internal_dpg.mvRawTexture
mvSubPlots=internal_dpg.mvSubPlots
mvColorMap=internal_dpg.mvColorMap
mvColorMapRegistry=internal_dpg.mvColorMapRegistry
mvColorMapButton=internal_dpg.mvColorMapButton
mvColorMapSlider=internal_dpg.mvColorMapSlider
mvTemplateRegistry=internal_dpg.mvTemplateRegistry
mvTableCell=internal_dpg.mvTableCell
mvItemHandlerRegistry=internal_dpg.mvItemHandlerRegistry
mvReservedUUID_0=internal_dpg.mvReservedUUID_0
mvReservedUUID_1=internal_dpg.mvReservedUUID_1
mvReservedUUID_2=internal_dpg.mvReservedUUID_2
mvReservedUUID_3=internal_dpg.mvReservedUUID_3
mvReservedUUID_4=internal_dpg.mvReservedUUID_4
mvReservedUUID_5=internal_dpg.mvReservedUUID_5
mvReservedUUID_6=internal_dpg.mvReservedUUID_6
mvReservedUUID_7=internal_dpg.mvReservedUUID_7
mvReservedUUID_8=internal_dpg.mvReservedUUID_8
mvReservedUUID_9=internal_dpg.mvReservedUUID_9
